#include <math.h>
#include <stdio.h>
#include <stdlib.h>


/* a fortran file obtained from Shaobeng Zeng that was run through f2c and
   then slightly modified by Jeff Thorne */

/* returns pseudo-random number from beta distribution with parameters
   alpha and beta, should work if parameters are both bigger than
   roughly 0.01 .  Mean of the distribution should be alpha/ (alpha+beta) 
   and variance should be alpha*beta/((alpha + beta + 1.0)*(alpha + beta)^2) */
double betarnd(alpha,beta)
double alpha,beta;
{
   double gammarnd();
   double asamp,bsamp;

   asamp = gammarnd(alpha,1.0);
   bsamp = gammarnd(beta,1.0);

   return(asamp/(asamp+bsamp));
}

/* routine returns random number from gamma distribution with
   shape parameter "shapepar" and scale parameter "scalepar" -
   mean of the distribution should be shapepar*scalepar,
   variance should be shapepar*scalepar^2 
   It seems to work as long as the shapepar is bigger than roughly 0.01
*/
double gammarnd(shapepar,scalepar)
double shapepar,scalepar; 
{
   double gamnl1_();
   double gamgbh_();
   double theanswer;
   if(shapepar < 0.75) theanswer = gamnl1_(&shapepar); 
                               /* gamnl1_ works for 0 < shapepar <1 */
   else theanswer = gamgbh_(&shapepar); /* gamgbh_ works for 0.25 < shapepar */
   return(scalepar*theanswer);
} 


/****************************************************************************
***/

double gamgbh_(beta)
double *beta;
{
    /* System generated locals */
    double ret_val, d_1;

    /* Builtin functions */
    double sqrt(), log();

    /* Local variables */
    static double a, b, c, d, t, u, w, h1, h2, u1, u2;
    double rnd();

/************************************************************************
*******/
/*        This is a general Gamma variate generator with the shape */
/*    parameter beta > 0.25. Coded from the Algorithm GBH of Cheng and 
Feast*/
/*    (1980 Communications of the ACM 23:389-394) by Shaobang Zeng on 
Jan. 30,*/
/*     1992. */
/************************************************************************
*******/
    if (*beta <= .25) {
	printf("error in gamgbh_()!!\n");
    }
    a = *beta - .25;
    b = *beta / a;
    c = 2. / a;
    d = c + 2.;
    t = 1. / sqrt(*beta);
    h1 = (t * .0245 / *beta + .4417) * t;
    h2 = (.222 - t * .043) * t;
L1:
    u = rnd();
    u1 = rnd();
    u2 = u1 + h1 * u - h2;
    if (u2 <= 0. || u2 >= 1.) {
	goto L1;
    }
/* Computing 4th power */
    d_1 = u1 / u2, d_1 *= d_1;
    w = b * (d_1 * d_1);
    if (c * u2 - d + w + 1. / w <= 0.) {
	goto L4;
    }
    if (c * log(u2) - log(w) + w - 1. >= 0.) {
	goto L1;
    }
L4:
    ret_val = a * w;
    return ret_val;
} /* gamgbh_ */



/****************************************************************************
**/

double gamnl1_(beta)
double *beta;
{
    double gammp_();
    /* System generated locals */
    double ret_val, d_1;

    /* Builtin functions */
    double pow(), log();

    /* Local variables */
    double a, b, p;
    static double u0, u1, u2;
    double rnd();

/************************************************************************
*******/
/*     This gamma generator generates gamma variates by composition- */
/*  rejection from the Weibull density for the shape parameter beta */
/*  smaller than 1 (0 < beta < 1). Coded from I. Vaduva (1977 Math. */
/*  Operationsforsch. Statist., Ser. Statistics, Vol. 8:545-576) by */
/*  Shaobang Zeng on Feb. 5 1992.   (Best one of this kind) */
/************************************************************************
*******/

    a = 1.0/( *beta);
    b = -1.0/(1.0-( *beta));
    p = gammp_(*beta,1.0);
    if (*beta >= 1. || *beta <= 0.) {
	printf("Error in gamnl1()!\n");
    }
    if (rnd() > p) {
	goto L3;
    }
L1:
    d_1 = rnd();
    u0 = pow(d_1, a);
    ret_val = u0;
L2:
    u1 = rnd();
    if (u0 < u1) {
	return ret_val;
    }
    u2 = rnd();
    if (u1 < u2) {
	goto L1;
    }
    u0 = u2;
    goto L2;
L3:
    d_1 = rnd();
    u0 = pow(d_1, b);
    ret_val = 1. - log((rnd()));
    if (ret_val >= u0) {
	goto L3;
    }
    return ret_val;
} /* gamnl1_ */


/****************************************************************************
**/

double gammp_(a, x)
double a, x;
{
    /* System generated locals */
    double ret_val;

    /* Local variables */
    extern /* Subroutine */ int gser_();
    static double gammcf, gamser;
    extern /* Subroutine */ int gcf_();
    static double gln;

    if (x < 0. || a <= 0.) {
	printf("Error in gammp_() !!\n");
    }
    if (x < a + 1.) {
	gser_(&gamser, a, x, &gln);
	ret_val = gamser;
    } else {
	gcf_(&gammcf, a, x, &gln);
	ret_val = 1. - gammcf;
    }
    return ret_val;
} /* gammp_ */


/****************************************************************************
**/

/* Subroutine */ int gser_(gamser, a, x, gln)
double *gamser, a, x, *gln;
{
    double log(), exp();

    /* Local variables */
    static int n;
    static double ap;
    double gammln_();
    static double del, sum;

    *gln = gammln_(&a);
    if (x <= 0.) {
	if (x < 0.) {
	    printf("Error in gser_() ! ! \n");
	}
	*gamser = 0.;
	return 0;
    }
    ap = a;
    sum = 1. / a;
    del = sum;
    for (n = 1; n <= 100; ++n) {
	ap += 1.;
	del = del * x / ap;
	sum += del;
	if (del*del < sum * sum * 3e-7) {
	    goto L1;
	}
/* L11: */
    }
    printf("gser_ ERROR: a TOO LARGE, itmax TOO SMALL\n");
L1:
    *gamser = sum * exp(-x + a * log(x) - *gln);
    return 0;
} /* gser_ */


/****************************************************************************
**/

/* Subroutine */ int gcf_(gammcf, a, x, gln)
double *gammcf, a, x, *gln;
{
    /* System generated locals */
    double d_1;

    /* Builtin functions */
    double log(), exp();

    /* Local variables */
    static double gold, g;
    static int n;
    static double a0, a1, b0, b1, an;
    extern double gammln_();
    double fac, ana, anf;

    *gln = gammln_(&a);
    gold = 0.;
    a0 = 1.;
    a1 = x;
    b0 = 0.;
    b1 = 1.;
    fac = 1.;
    for (n = 1; n <= 100; ++n) {
	an = (double) n;
	ana = an - a;
	a0 = (a1 + a0 * ana) * fac;
	b0 = (b1 + b0 * ana) * fac;
	anf = an * fac;
	a1 = x * a0 + anf * a1;
	b1 = x * b0 + anf * b1;
	if (a1 != 0.) {
	    fac = 1. / a1;
	    g = b1 * fac;
	    if ((d_1 = (g - gold) / g, sqrt(d_1*d_1)) < 3e-7) {
		goto L1;
	    }
	    gold = g;
	}
/* L11: */
    }
    printf("gcf_ ERROR: a TOO LARGE, itmax TOO SMALL", 28L);
L1:
    *gammcf = exp(-x + a * log(x) - *gln) * g;
    return 0;
} /* gcf_ */


/****************************************************************************
**/

double gammln_(xx)
double *xx;
{
    /* Initialized data */

    static double cof[6] = { 76.18009173,-86.50532033,24.01409822,
	    -1.231739516,.00120858003,-5.36382e-6 };
    static double stp = 2.50662827465;
    static double half = .5;
    static double one = 1.;
    static double fpf = 5.5;

    /* System generated locals */
    double ret_val;

    /* Builtin functions */
    double log();

    /* Local variables */
    static int j;
    static double x, ser, tmp;

    x = *xx - one;
    tmp = x + fpf;
    tmp = (x + half) * log(tmp) - tmp;
    ser = one;
    for (j = 1; j <= 6; ++j) {
	x += one;
	ser += cof[j - 1] / x;
/* L11: */
    }
    ret_val = tmp + log(stp * ser);
    return ret_val;
} /* gammln_ */

