/* (c) Copyright 2002 by Jeff Thorne

VERSION 8/5/02 

As this program is very crude, please do not distribute it. Instead,
people can obtain hopefully improved versions by contacting me (Jeff Thorne)
(Also, please don't delete this message)
 
This version has the expected rate at a descendant node equal to the rate
at the ancestral node - this is different than the previous version which
had expected log rate at descendant equal to log rate of ancestor.

However, one arbitrarily selected branch that is attached to the
root has the rate of the node ending the branch being forced
to be identical to the rate at the root node.

To compile to analyze a data set, type:

cl -Ot -O2 -W3 multidivtime.c dlinpack.c arrayutl.c

cc -o multidivtime -O4 -ieee_with_no_inexact multidivtime.c dlinpack.c arrayutl.c -lm

cc -o multidivtime -march=pentiumpro -mcpu=pentiumpro -O4 multidivtime.c dlinpack.c arrayutl.c -lm

cc -o multidivtime -march=athlon -mcpu=athlon -O4 -funroll-loops -fomit-frame-pointer -finline-functions multidivtime.c dlinpack.c arrayutl.c -lm


gcc multidivtime.c dlinpack.c arrayutl.c -lm -O  -o multidivtime 

(use "cc" above instead of "gcc" if "cc" is the compiler to which you have
 access)

To run, type:

multidivtime suffix 

"suffix" is the suffix with which output files will be named ...

*/

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "arrayutl.h"

#define min2(A,B) ((A) < (B) ? (A) : (B))
#define min3(A,B,C) ((A) < (B) ? min2(A,C) : min2(B,C)) 
#define max2(A,B) ((A) > (B) ? (A) : (B))
#define max3(A,B,C) ((A) > (B) ? max2(A,C) : max2(B,C))

#define PI 3.1415927
#define BIGNEGLOG -1000 /* a big negative logarithm */
#define MAXGENE 66 /* maximum number of gene data sets MINUS 1 */
#define MAXNODE 200 /* maximum number of nodes on rooted tree */
#define MAXOFF 10 /* Maximum number of offspring nodes of a node */
#define SMALLNUM 1.0e-10 /* a number slightly greater than 0 */
#define NEGSMALLNUM -1.0e-10 /* a number slightly less than 0 */
#define NAMELEN 52 
#define TINYLEN 0.0001
#define TINYTIME 0.00001
#define SMALLRATE 0.0001
#define NUMCORREPS 1000
#define LOWFRAC 0.025 /* lowest 0.025 of samples cut from 
                         posterior interval */
#define HIGHFRAC 0.975 /* highest 0.025 of samples cut from
                          posterior interval */
/* needed by rnd() routine */
typedef long longer[6];
static longer seed;
 
short fulldata; /* fulldata=0 if any taxa from any genes are missing
                   fulldata=1 if all taxa represented for all genes */
int commonbrown; /* commonbrown = 1 if all genes have same tendency to
                      change rates ...
                      commonbrown = 0 if each gene has individual (separate)
                      tendency to change rates ...
                   */
int **tonode;
int nodata; /* value 0 to approximate posterior, value 1 to approx. prior */ 
int viral; /* value 0 means all tip sequences isolated at same time,
              value 1 means not all tip sequences isolated at same time */ 
int consnode, ntipcons;
int currentnode, nbranches, nnodes, ntips;
int starttype;
int nintnodes; /* number internal nodes on tree that are NOT root */
char **name;
double a[MAXNODE];
double b[MAXNODE];
double **mean, **varcovar, *hightime, *lowtime, *tipcons;
double timegamr, timegamlam, rategamr, rategamlam, browngamr, browngamlam;
double rttm, rttmsd,rtrate,rtratesd;
double brownmean; /* value 0.0 means constant rates (i.e., "perfect" clock)
                     whereas value > 0.0 means allow rates to evolve */
double brownsd, minab,newk, othk;
double thek, bigtime;
double *curmvn;
long failrate[MAXGENE][MAXNODE], failtime[MAXNODE],failcons;
long sucrate[MAXGENE][MAXNODE];
long suctime[MAXNODE],succons, sucboth, failboth, sucrateflate[MAXGENE];
long failrateflate[MAXGENE];
long suctimeflate, failtimeflate;
long sucpartcons, failpartcons, sucpartboth, failpartboth;
long sucparttime, failparttime, sucpartrate[MAXGENE], failpartrate[MAXGENE];
long numsamps, sampfreq, burnin;

int ngenes; /* set to number of genes minus 1 */
int **mbr2gbr; /* mbr2gbr[i][j] equals the branch of gene i to which branch j
                  of the master tree maps */
int ***ttonode,*tnnodes,*tntips,*tnbranches;
char **genefile;

FILE *gfopen(char *filename, char *mode);

main(argc, argv)
int argc;
char *argv[];
{

   void rddata();
   void initialize();
   void dothemcmc();
   void summarize();
   double rnd();
   extern void freethreeimat();

   double *time, **rate, ***samprate, **samptime;
/* needed by rnd() routine */
   FILE *inseed; /* contains the seed */
   long i,initseed;
   double lastrnd;
   int ktrack, ktrack2, sizemem=0;

   if(argc!=2) {
      puts("\nusage: multidivtime <file suffix>");
      exit(-1);
   }

   printf("Suffix for output files from this run is %s\n", argv[1]);
/* needed by rnd() routine */
   inseed=gfopen("inseed","r");
   fscanf(inseed,"%ld",&initseed);
   fclose(inseed);
   printf("initseed = %ld\n", initseed);
   for (i = 0; i <= 5; i++) seed[i] = 0;
   i = 0;
   do {
     seed[i] = initseed & 63;
     initseed /= 64;
     i++;
   } while (initseed != 0);
   for(i=1; i<= 100; i++)
   {
      lastrnd = rnd(seed);
   }

   rddata();
   if(brownmean > SMALLNUM)
   {
      browngamr = brownmean*brownmean/(brownsd*brownsd);
      browngamlam = brownmean/(brownsd*brownsd);
   }

   timegamr = rttm*rttm/(rttmsd*rttmsd);
   timegamlam = rttm/(rttmsd*rttmsd);
   rategamr = rtrate*rtrate/(rtratesd*rtratesd);
   rategamlam = rtrate/(rtratesd*rtratesd);

   if(nodata==1)
   {
      printf("The NO DATA - EXPLORE YOUR PRIOR VERSION\n");
   }
   else
   {
      printf("Approximate the posterior version\n");
   }
   if(commonbrown ==0 && brownmean > SMALLNUM)
   {
      printf("Genes differ in tendencies to change rates over time\n");
   }
   if(commonbrown == 1 && brownmean >  SMALLNUM)
   {
      printf("Genes all have common tendency to change rate over time\n");
   }
   printf("sampfreq %6ld numsamps %6ld burnin %6ld\n",sampfreq,numsamps,burnin);
   printf("newk %6.3lf othk %6.3lf thek %6.3lf\n", newk, othk, thek);
   printf("brownmean %6.3lf brownsd %6.3lf\n", brownmean, brownsd);
   if(brownmean > SMALLNUM)
   {
      printf("browngamr %6.3lf browngamlam %6.3lf\n", browngamr, browngamlam);
   }
   else
   {
      printf("THIS ANALYSIS ASSUMES CONSTANT RATE OF EVOLUTION ");
      printf("(i.e., so-called perfect clock)\n");
   }
   printf("rttm %6.3lf rttmsd %6.3lf minab %6.3lf\n", rttm, rttmsd,
      minab);
   printf("timegamr %6.3lf timegamlam %6.3lf\n", timegamr, timegamlam);
   printf("rtrate %8.4lf rtratesd %8.4lf\n", rtrate, rtratesd);
   printf("rategamr %6.3lf rategamlam %6.3lf\n", rategamr, rategamlam);

   printf("nbranches= %2d nnodes= %2d\n", nbranches, nnodes);
   nintnodes = nnodes - ntips -1;
   printf("ntips = %2d nintnodes = %2d\n", ntips, nintnodes);
   printf("bigtime = %8.3lf\n", bigtime);
   printf("NUMCORREPS = %5d\n", NUMCORREPS);

   time = dvec(nnodes);
   rate = dmat(ngenes,nnodes);
   for(ktrack=0; ktrack <= nnodes; ktrack++)
   {
      if(brownmean > SMALLNUM)
      {
         for(ktrack2 = 0; ktrack2 <= ngenes; ktrack2++)
         {
            sucrate[ktrack2][ktrack] = 0;
            failrate[ktrack2][ktrack] = 0;
         }
      }
      suctime[ktrack] = 0;
      failtime[ktrack] = 0;
   }
   if(brownmean > SMALLNUM)
   {
      succons = 0;
      failcons = 0;
      sucpartcons = 0;
      failpartcons = 0;
      sucpartboth = 0;
      failpartboth = 0;
      for(ktrack=0; ktrack <= ngenes; ktrack++)
      {
         sucpartrate[ktrack] = 0;
         failpartrate[ktrack] = 0;
      }
   }
   sucboth = 0;
   failboth = 0;
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      sucrateflate[ktrack] = 0;
      failrateflate[ktrack] = 0;
   }
   suctimeflate = 0;
   failtimeflate = 0;
   sucparttime = 0;
   failparttime = 0;


/* Added by Ziheng to give a rough idea of the mem required */
sizemem = ((ngenes+1)*(nnodes+5)*numsamps + ngenes*nnodes*nnodes)*sizeof(double);
fprintf(stderr, "\nNeeds >%d bytes of memory.  Enter to continue.\n", sizemem);
if(sizemem>10000000) getchar();

   samprate = threedmat(ngenes,nnodes+5,numsamps);
   samptime = dmat(nnodes+5,numsamps);
   initialize(time, rate, starttype); /* 3rd parameter=1 means 
      rndstart, other values mean start at hopefully good guess */
   if(viral == 0)
   {
      printf("All sequences (tips) isolated at same time\n");
   }
   else
   {
      printf("Sequences (tips) are not all isolated at same time\n");
   }
   dothemcmc(time,rate,samptime, samprate,argv[1]);
   summarize(samptime, samprate, argv[1]);

/* needed by rnd() routine */
   initseed = ((long) 1000000000*rnd(seed));
   if(initseed % 2 ==0)
   {
      initseed++;
   }
   inseed=gfopen("inseed","w");
   fprintf(inseed,"%ld",initseed);  /* put in a new random seed */
   fclose(inseed);

   freethreedmat(samprate,ngenes,nnodes+5,numsamps);
   freedmat(samptime,nnodes+5,numsamps);
   freedvec(time);
   freedvec(tipcons);
   freedvec(hightime);
   freedvec(lowtime);
   freedmat(rate,ngenes,nbranches);
   freedmat(mean,ngenes,nbranches);
   freecmat(name,MAXNODE, NAMELEN);
   freedmat(varcovar,ngenes, (nbranches+1)*(nbranches+1));
   freeimat(tonode,nnodes,MAXOFF);
   freeimat(mbr2gbr, ngenes,nnodes);
   freecmat(genefile, ngenes,20);
   freethreeimat(ttonode,ngenes,nnodes,MAXOFF);
   freeivec(tnnodes, ngenes);
   freeivec(tntips, ngenes);
   freeivec(tnbranches, ngenes);
}


/* this routine assumes that the log-likelihood surface is being approximated
by a multivariate normal distribution with mean equal to the
maximum likelihood values of the branch lengths and the variance
derived from the curvature of the log-likelihood surface.  
*/
/* nodata=1 forces this routine to return
the same value no matter what the data are -- this option
is for investigating the prior distribution */
double logmvn(mvalues, whichgene)
const double *mvalues;
const int whichgene;
{
   static int holdtnb; 
   static int hereyet = 0;
   static int ktrack, ktrack2, ktrack3;
   static int *holdgb;
   static double tempsum;
   static double **mathold, *holdmean;
   static double thesum;
   static double tr2hold;
   static double *tr2values;
   static double *tempvec;
   static double ***invvcv;
   static double *values;
   static double *lmth;

   void partialinvert();
   extern double **dmat();
   extern double *dvec();
   extern double ***threedmat();
   extern void freedmat();

   if(nodata != 1)
   {

      if(hereyet==0)
      {
         values = dvec(nnodes);
         invvcv = threedmat(ngenes,nbranches-1, nbranches-1);
         tempvec = dvec(nbranches-1);
         tr2values = dvec(nbranches-1);
         for(ktrack3 = 0; ktrack3 <= ngenes; ktrack3++)
         {
            mathold = invvcv[ktrack3];
            for(ktrack=0; ktrack < tnbranches[ktrack3]; ktrack++)
            {  
               for(ktrack2= 0 ; ktrack2 < tnbranches[ktrack3]; ktrack2++)
               {
                  mathold[ktrack][ktrack2] = 
                     varcovar[ktrack3][ktrack*tnbranches[ktrack3]+ktrack2];
               }
            }
            partialinvert(mathold, 0,tnbranches[ktrack3]-1);
         }   
         hereyet = 1;
      }
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         values[ktrack] = 0.0;
      }
      
      holdgb = mbr2gbr[whichgene];
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         if( *holdgb >=0)
         {
            values[ *holdgb]  += mvalues[ktrack];
         }
         holdgb++; 
      }    
      holdtnb = tnbranches[whichgene];
      holdmean = mean[whichgene];
      for(ktrack = 0; ktrack < holdtnb; ktrack++)
      {
         tr2values[ktrack] = values[ktrack] - holdmean[ktrack];
      }
      mathold = invvcv[whichgene];
      thesum = 0.0;
      for(ktrack = 0; ktrack < holdtnb; ktrack++)
      {
         tr2hold = tr2values[ktrack];
         thesum -= tr2hold*tr2hold *mathold[ktrack][ktrack];
      }
      thesum *= 0.5;
       
      for(ktrack = 1; ktrack < holdtnb; ktrack++)
      {
         lmth = mathold[ktrack];
         tempsum = 0.0;
         for(ktrack2 = 0; ktrack2 < ktrack; ktrack2++)
         {
            tempsum -= *lmth * tr2values[ktrack2];
            lmth++;
         }   
         thesum += tempsum*tr2values[ktrack];
      }
   }
   else
   {
      thesum= 0.0; /* NO  DATA VERSION */
   }
   return(thesum); 
}

/* this routine assumes that the log-likelihood surface is being approximated
by a multivariate normal distribution with mean equal to the
maximum likelihood values of the branch lengths and the variance
derived from the curvature of the log-likelihood surface.  the routine
assumes that branch lengths surrounding a particular node have changed
but all others are constant.  So, it calculates the density (up to a 
proportionality constant) of the branch lengths surrounding the node
conditional on all other branch lengths.  It returns the logarithm
of this conditional density. nodata = 1 causes this routine to return
the same value no matter what the data are -- this option
is for investigating the prior distribution */
double logcondmvn(mvalues, whichnode, whichgene)
double *mvalues;
int whichnode, whichgene; 
{
   static int hereyet = 0;
   static int **pos, **invpos;
   static int *maxbrpos;
   int holdindex,pivotnode, ktrack, ktrack2, ktrack3, ktrack4, brpos1, brpos2;
   double thesum;
   double **vcv, **dummy;
   static int maxmaxbrpos;
   static double ****pp;
   static double ****qq;
   static double *tr2values;
   static double *tr3;
   static double *values;
   static double ***holdpp;
   static double ***holdqq;
   void partialinvert();
   extern int *ivec();
   extern int **imat();
   extern double **dmat();
   extern double *dvec();
   extern double ***threedmat();
   extern double ****fourdmat();
   extern void freedmat();

   if(nodata !=1)
   {
      if(hereyet==0)
      {
         values = dvec(nnodes);
         pos = imat(nnodes-1, nnodes-1);
         invpos = imat(nnodes-1, nnodes-1);
         maxbrpos = ivec(nnodes-1);
         maxmaxbrpos = 0;
         for(pivotnode=0; pivotnode < nnodes; pivotnode++)
         {
            if(pivotnode == consnode)
            {
               maxbrpos[pivotnode] =tonode[consnode][0] +
                  tonode[nnodes-1][0];
            }
            else
            {
               if(pivotnode == nnodes-1)
               {
                  maxbrpos[pivotnode] = tonode[nnodes-1][0];
               }
               else
               {
                  maxbrpos[pivotnode] = 1+tonode[pivotnode][0];
               }
            }        
            maxmaxbrpos = max2(maxmaxbrpos,maxbrpos[pivotnode]);
         }
         maxmaxbrpos -= 1;
         pp = fourdmat(ngenes,nnodes-1, maxmaxbrpos, nnodes-1);
         qq = fourdmat(ngenes, nnodes-1, maxmaxbrpos, maxmaxbrpos);
/* COMMENTED OUT CHANGE
         holdpp = threedmat(nnodes-1, maxmaxbrpos, nnodes-1);
         holdqq = threedmat(nnodes-1, maxmaxbrpos, maxmaxbrpos);
*/
         tr2values = dvec(nnodes-1);
         tr3 = dvec(nnodes-1);
         dummy = dmat(nnodes-1,nnodes-1);
         for(pivotnode=0; pivotnode < nnodes; pivotnode++)
         {
            brpos1 = 0;
            brpos2 = 0;
            for(ktrack = 0; ktrack < nbranches; ktrack++)
            {
               if((ktrack== pivotnode) || (pivotnode == consnode 
                  && (nnodes-1) == tonode[ktrack][MAXOFF]) || (
                  tonode[ktrack][MAXOFF] == pivotnode))
               {      
                  invpos[pivotnode][brpos1] = ktrack;
                  pos[pivotnode][ktrack] = brpos1;
                  brpos1++;
               }
               else
               {
                  invpos[pivotnode][brpos2+maxbrpos[pivotnode]] = ktrack;
                  pos[pivotnode][ktrack] = brpos2+maxbrpos[pivotnode];
                  brpos2++;
               }
            }
            vcv = dmat(nbranches-1, nbranches-1);
            for(ktrack4 = 0; ktrack4 <= ngenes; ktrack4++)
            {
               holdpp = pp[ktrack4];
               holdqq = qq[ktrack4];
               for(ktrack=0; ktrack < nbranches; ktrack++)
               {
                  for(ktrack2= 0 ; ktrack2 < nbranches; ktrack2++)
                  {
                     vcv[ktrack][ktrack2] = 
                        varcovar[ktrack4][ktrack*nbranches+ktrack2];
                  }
               }
               for(ktrack = 0; ktrack < nbranches; ktrack++)
               {
                  for(ktrack2=0; ktrack2 < nbranches; ktrack2++)
                  {
                     dummy[ktrack][ktrack2] = vcv[invpos[pivotnode][ktrack]][
                        invpos[ pivotnode][ktrack2]];
                  }
               }
               partialinvert(dummy, maxbrpos[pivotnode],nbranches-1); 
               for(ktrack=0; ktrack< maxbrpos[pivotnode]; ktrack++)
               {
                  for(ktrack2=maxbrpos[pivotnode]; ktrack2 < nbranches; 
                     ktrack2++)
                  {
                     holdpp[pivotnode][ktrack][ktrack2-maxbrpos[pivotnode]] = 
                        0.0;
                     for(ktrack3=maxbrpos[pivotnode];ktrack3 < nbranches;
                        ktrack3++)
                     {
                        holdpp[pivotnode][ktrack][ktrack2-maxbrpos[pivotnode]] 
                           += dummy[ktrack][ktrack3]*dummy[ktrack3][ktrack2];
                     }
                  }
               }
               for(ktrack=0; ktrack < maxbrpos[pivotnode]; ktrack++)
               {
                  for(ktrack2=0; ktrack2 < maxbrpos[pivotnode]; ktrack2++)
                  {
                     holdqq[pivotnode][ktrack][ktrack2] = vcv[
                        invpos[pivotnode][ktrack]][invpos[pivotnode][ktrack2]];
                     for(ktrack3=0; ktrack3 < nbranches-maxbrpos[pivotnode]; 
                        ktrack3++)
                     {
                        holdqq[pivotnode][ktrack][ktrack2] -= 
                           holdpp[pivotnode][ktrack][ ktrack3] * 
                           dummy[ktrack3+maxbrpos[pivotnode]][ktrack2];
                     }
                  }
               }
               partialinvert(holdqq[pivotnode],0,maxbrpos[pivotnode]-1);
            }
         }
         freedmat(dummy,nnodes-1,nnodes-1);
         freedmat(vcv,nbranches-1, nbranches-1);
         hereyet = 1;
      }
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         values[ktrack] = 0.0;
      }
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         if(mbr2gbr[whichgene][ktrack] >=0)
         {
            values[mbr2gbr[whichgene][ktrack]]  += mvalues[ktrack];
         }
      }
      holdpp = pp[whichgene];
      holdqq = qq[whichgene];
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         holdindex = invpos[whichnode][ktrack];
         tr2values[ktrack] = values[holdindex] - 
            mean[whichgene][holdindex];
      }
      for(ktrack  =0; ktrack < maxbrpos[whichnode]; ktrack++)
      {
         for(ktrack2 = 0; ktrack2 < nbranches-maxbrpos[whichnode]; ktrack2++)
         {
            tr2values[ktrack] -= holdpp[whichnode][ktrack][ktrack2] * 
            tr2values[ktrack2+maxbrpos[whichnode]];   
         }
      } 
      for(ktrack=0; ktrack < maxbrpos[whichnode]; ktrack++)
      {
         tr3[ktrack] = 0.0;
         for(ktrack2=0; ktrack2 < maxbrpos[whichnode]; ktrack2++)
         {
            tr3[ktrack] +=tr2values[ktrack2]*holdqq[whichnode][ktrack2][ktrack];
         }
      }
      thesum = 0.0;
      for(ktrack=0; ktrack < maxbrpos[whichnode]; ktrack++)
      {
         thesum += tr3[ktrack]*tr2values[ktrack];
      }
      thesum *= -0.5; /* logarithm of the conditional density */
   }
   else
   {
      thesum = 0.0; /* NO  DATA VERSION */
   }
   return(thesum); 
}


void partialinvert(a,lower, upper)
int lower,upper;
double **a;
{
  int i, j, k ;
  double temp;

  /*Gauss-Jordan reduction -- invert matrix a in place,
         overwriting previous contents of a.  On exit, matrix a
         contains the inverse.*/
  for (i = lower; i <= upper; i++) {
    temp = 1.0 / a[i][i];
    a[i][i] = 1.0;
    for (j = lower; j <= upper; j++) {
      a[i][j] *= temp;
    }
    for (j = lower; j <= upper; j++) {
      if (j != i) {
	temp = a[j][i];
	a[j][i] = 0.0;
	for (k = lower; k <= upper; k++) {
	  a[j][k] -= temp * a[i][k];
	}
      }
    }
  }
}  /*invert*/

void rddata()
{
   void rdtopology();
   void rdmulticntrl();
   void printname();
   extern int ***threeimat();
   extern char ***threecmat();
   extern void freethreecmat();

   int lastgbr, themasterpar, thegenepar;
   int stopflag, thenode, ktrack, ktrack2, ktrack3, ktrack4, ktrack5;
   int numchildren, checkthem, keepgoing, child[MAXNODE], *order, place;
   int **masternodehere, **mastertogene, **genetomaster;
   FILE *datfile, *treefile;
   char thischar, *cstring, ***tname,*tempname;
 
   rdmulticntrl();

   for(ktrack =0; ktrack < ntips; ktrack++)
   {
      lowtime[ktrack] = tipcons[ktrack];
      hightime[ktrack] = tipcons[ktrack];
   }
   for(ktrack= 0; ktrack < nnodes-1; ktrack++)
   {
      lowtime[tonode[ktrack][MAXOFF]] = max2( 
      lowtime[tonode[ktrack][MAXOFF]], lowtime[ktrack]);
   }
   for(ktrack= nnodes-2; ktrack  >= 0; ktrack--)
   {
      hightime[ktrack] = min2(hightime[ktrack],
      hightime[tonode[ktrack][MAXOFF]]);
   }
     
   mean = dmat(ngenes,nnodes);
   ttonode = threeimat(ngenes,nnodes,MAXOFF);
   tname = threecmat(ngenes,nnodes,NAMELEN);
   tnnodes = ivec(ngenes);
   tntips = ivec(ngenes);
   tnbranches = ivec(ngenes);
   fulldata =1;
   printf("Number of genes= %2d\n", ngenes+1);
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      printf("Name of gene %2d is %s\n", ktrack, genefile[ktrack]);
   }
   for(ktrack5=0; ktrack5 <= ngenes; ktrack5++)
   {

      treefile = gfopen(genefile[ktrack5],"r");

      /* read rooted topology */
      rdtopology(treefile,ttonode[ktrack5], &(tnnodes[ktrack5]),
         &(tntips[ktrack5]), mean[ktrack5],tname[ktrack5]);
      tnbranches[ktrack5] = tnnodes[ktrack5]-1;

      while(getc(treefile) != '\n');
      while(getc(treefile) != '\n');

      if(ktrack5==0)
      {
         order = ivec(nnodes);
         varcovar = dmat(ngenes, (nbranches+1) * (nbranches+1));
         tempname = cvec(NAMELEN);
         masternodehere = imat(ngenes,nnodes);
         mastertogene = imat(ngenes,nnodes);
         genetomaster = imat(ngenes,nnodes);
         mbr2gbr = imat(ngenes,nnodes);
      }

      for(ktrack=0; ktrack < nnodes; ktrack++)
      {
         mastertogene[ktrack5][ktrack] = -1;
         genetomaster[ktrack5][ktrack] = -1;
         masternodehere[ktrack5][ktrack] = 0;
      }

      place = 0;
      for(ktrack=0; ktrack < tntips[ktrack5]; ktrack++)
      {
         thischar = getc(treefile);
         while(thischar != ' ')
         {
            tempname[place] = thischar;
            thischar = getc(treefile);
            place++;
         } 
         tempname[place] = '\n';
         place = 0;
         fscanf(treefile,"%d\n", &thenode);
         stopflag = 0;
         for(ktrack2=0; ktrack2 < ntips; ktrack2++)
         {
            for(ktrack3=0; ktrack3 <= NAMELEN; ktrack3++)
            {
               if( tempname[ktrack3] != tname[ktrack5][ktrack2][ktrack3])
               {
                  break;
               }
               else
               {
                  if(tempname[ktrack3] == '\n')
                  {
                     order[thenode] = ktrack2;
                     stopflag = 1;
                     break;
                  }
               }
            }
            if( stopflag ==1)
            {
               break;
            }
         }
         stopflag = 0;
         for(ktrack2=0;ktrack2 < ntips; ktrack2++)
         {
            for(ktrack3=0;ktrack3 <= NAMELEN; ktrack3++)
            {
               if(tempname[ktrack3] != name[ktrack2][ktrack3])
               {
                  break;
               }
               else
               {
                  if(tempname[ktrack3] == '\n')
                  {
                     masternodehere[ktrack5][ktrack2] = 1;
                     mastertogene[ktrack5][ktrack2]=thenode;
                     genetomaster[ktrack5][thenode]= ktrack2;
                     stopflag = 1;
                     break;
                  }
               }
            }  
            if(stopflag==1)
            {
               break;
            }
         }
      }
      
      while(getc(treefile) != '\n');
      for(ktrack = tntips[ktrack5]; ktrack < tnnodes[ktrack5]; ktrack++)
      {
         numchildren = 2;
         fscanf(treefile,"%d%d%d", &child[1], &child[2], &thenode);
         if(getc(treefile) != '\n')
         {
            keepgoing = 1;
            numchildren++;
            child[numchildren] = thenode;
            while(keepgoing ==1)
            {
               fscanf(treefile, "%d", &thenode);
               if(getc(treefile) != '\n' )
               {
                  numchildren++;
                  child[numchildren] = thenode;
               }
               else
               {
                  keepgoing = 0;
               }
            }
         }
         for(ktrack2= tntips[ktrack5]; ktrack2 <= tnnodes[ktrack5]; ktrack2++)
         { 
            if(numchildren == ttonode[ktrack5][ktrack2][0])
            {
               for(ktrack4 = 1; ktrack4 <= numchildren; ktrack4++)
               {
                  for(ktrack3=1; ktrack3 <= numchildren; ktrack3++)
                  {
                     if(order[child[ktrack3]] != 
                        ttonode[ktrack5][ktrack2][ktrack4])
                     {
                        checkthem =1;
                     }
                     else
                     {
                        checkthem = 0;
                        break;
                     }
                  }   
                  if(checkthem == 1)
                  {
                     break;
                  }
               }
            }
            else
            {
               checkthem =1;
            }
            if(checkthem ==0)
            {
               order[thenode] = ktrack2;
               break;
            }
            if(ktrack2 == tnnodes[ktrack5])
            {
               printf("trouble mapping names and nodes to tree\n");
               break;
            }
         }
      }


      while(getc(treefile) != '\n');
   
      /* read covariances */
      for(ktrack=0; ktrack < tnbranches[ktrack5]; ktrack++)
      {
         for(ktrack2=0; ktrack2 < tnbranches[ktrack5]; ktrack2++)
         {
            fscanf(treefile, "%lf", &varcovar[ktrack5][order[ktrack]*
               tnbranches[ktrack5] + order[ktrack2]]);
         }
      }
      fclose(treefile);

      for(ktrack2=ntips; ktrack2 < nnodes; ktrack2++)
      {
         for(ktrack3=1; ktrack3 <= tonode[ktrack2][0]; ktrack3++)
         {
            if(masternodehere[ktrack5][tonode[ktrack2][ktrack3]] >=1)
            {
               masternodehere[ktrack5][ktrack2]++;
            }
         }
      } 
      for(ktrack2=0; ktrack2 < ntips; ktrack2++)
      {
/*
         printf("Master Sequence %2d (",ktrack2);
         printname(ktrack2);
*/
         if(masternodehere[ktrack5][ktrack2] ==1)
         {
/*
            printf(") is present in gene data set %2d\n", ktrack5); 
*/
            lastgbr = mastertogene[ktrack5][ktrack2];
            mbr2gbr[ktrack5][ktrack2] = lastgbr;
            thegenepar = ttonode[ktrack5][mastertogene[ktrack5][ktrack2]][
               MAXOFF];
            themasterpar = tonode[ktrack2][MAXOFF];
            while(thegenepar != -1)
            {
               while(masternodehere[ktrack5][themasterpar] <= 1)
               {
                  mbr2gbr[ktrack5][themasterpar] = lastgbr;
                  themasterpar = tonode[themasterpar][MAXOFF];
               }
               mastertogene[ktrack5][themasterpar] = thegenepar;
               genetomaster[ktrack5][thegenepar] = themasterpar;
               lastgbr = thegenepar;
               mbr2gbr[ktrack5][themasterpar] = lastgbr;
               themasterpar = tonode[themasterpar][MAXOFF];
               thegenepar = ttonode[ktrack5][thegenepar][MAXOFF];
            }
         }
         else
         {
            mbr2gbr[ktrack5][ktrack2] = -1;
            fulldata=0;
            printf("Master Sequence %2d (",ktrack2);
            printname(ktrack2);
            printf(") is absent in gene data set %2d\n", ktrack5);
         }
      }
   }
   for(ktrack = 0; ktrack <= ngenes; ktrack++)
   {
      if(masternodehere[ktrack][nnodes-1] < 2)
      {
         mbr2gbr[ktrack][nnodes-1] = -1;
         for(ktrack2 = nnodes-2; ktrack2 >= ntips; ktrack2--)
         {
            if(masternodehere[ktrack][ktrack2] < 2 && mbr2gbr[ktrack][
               tonode[ktrack2][MAXOFF]] == -1)
            {
               mbr2gbr[ktrack][ktrack2] = -1;
            }
         }
      }
   }  
   for(ktrack = ntips; ktrack < nbranches; ktrack++)
   {
      for(ktrack2 = 0; ktrack2 <= ngenes; ktrack2++)
      {
         stopflag = 0;
         for(ktrack3=1; ktrack3 <= tonode[ktrack][0]; ktrack3++)
         {
            if(mbr2gbr[ktrack2][tonode[ktrack][ktrack3]] != -1)
            {
               stopflag =1;
               break;
            }
         }
         if(stopflag == 0 || masternodehere[ktrack2][tonode[ktrack][
            MAXOFF]] <= 0)
         {
            mbr2gbr[ktrack2][ktrack] = -1;
         }
         if(mbr2gbr[ktrack2][ktrack] >= tnbranches[ktrack2])
         {
            mbr2gbr[ktrack2][ktrack] = -1;
         }
      }
   }
                  
#ifdef VERBOSE
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      for(ktrack2=0; ktrack2 < nnodes; ktrack2++)
      {
         printf("masternodehere[gene=%2d][mbr=%2d]= %2d\n", ktrack, ktrack2,
            masternodehere[ktrack][ktrack2]);
      }
   }            
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      for(ktrack2=0; ktrack2 < nnodes; ktrack2++)
      {
         printf("mastertogene[gene=%2d][mbr=%2d]= %2d\n", ktrack, ktrack2,
            mastertogene[ktrack][ktrack2]);
      }
   }            
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      for(ktrack2=0; ktrack2 < nnodes; ktrack2++)
      {
         printf("genetomaster[gene=%2d][gbr=%2d]= %2d\n", ktrack, ktrack2,
            genetomaster[ktrack][ktrack2]);
      }
   }            
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      for(ktrack2=0; ktrack2 < nbranches; ktrack2++)
      {
         printf("mbr2gbr[gene=%2d][mbr=%2d] = %2d\n",  ktrack, ktrack2,
            mbr2gbr[ktrack][ktrack2]); 
      }
   }            
#endif
   freeimat(mastertogene,ngenes,nnodes);
   freeimat(genetomaster,ngenes,nnodes);
   freeimat(masternodehere,ngenes,nnodes);
   freeivec(order);
   freethreecmat(tname,ngenes,nnodes,NAMELEN);
   freecvec(tempname);
}

void rdmulticntrl()
{
   int **ttonode, ktrack, ktrack2,ncons, whichnode;
   char constype;
   double *tmean,*junkvec,thecons;
   FILE *datfile;
   char *cstring;

   void rdtopology();
   void fprinttree();

   cstring = cvec(96);
   datfile = gfopen("multicntrl.dat", "r");

   while ( getc(datfile) != '\n');
   while ( getc(datfile) != '\n');
   while ( getc(datfile) != '\n');
/* GET MASTER TREE - START CHANGE */

   tmean = dvec(MAXNODE);
   ttonode = imat(MAXNODE,MAXOFF);
   name = cmat(MAXNODE,NAMELEN);
   rdtopology(datfile,ttonode,&nnodes,&ntips,tmean,name);
   while(getc(datfile) != '\n');
   freedvec(tmean);
   nbranches = nnodes-1;
   tonode = imat(nnodes,MAXOFF);
   for(ktrack = 0; ktrack < nnodes; ktrack++)
   {
      tonode[ktrack][MAXOFF] = ttonode[ktrack][MAXOFF];
      for(ktrack2 = 0; ktrack2 <= ttonode[ktrack][0]; ktrack2++)
      {
         tonode[ktrack][ktrack2] = ttonode[ktrack][ktrack2];
      }
   }
   freedmat(ttonode,MAXNODE,MAXOFF);
   junkvec = dvec(nnodes);
   for(ktrack=0;ktrack <= nnodes; ktrack++)
   {
      junkvec[ktrack] = (double) ktrack;
   }
   printf("Root node number of master tree is %3d\n", nnodes-1);
   printf("Here are node numbers of other internal nodes on master tree:\n");
   fprinttree(stdout,nnodes-1,2,nnodes-1,junkvec);
   freedvec(junkvec);


/* GET MASTER TREE - END CHANGE */

   fscanf(datfile,"%d",&ngenes);  
   while ( getc(datfile) != '\n');
   ngenes--;  /* ngenes = NUMBER OF GENES minus 1 */
   genefile = cmat(ngenes, 20);
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      fscanf(datfile,"%s\n", cstring);
      strcpy(genefile[ktrack], cstring);
   } 

   fscanf(datfile, "%ld", &numsamps);
   while ( getc(datfile) != '\n'); 
   fscanf(datfile, "%ld", &sampfreq);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%ld", &burnin);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%d", &starttype);
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &rttm);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &rttmsd);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &rtrate);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &rtratesd);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &brownmean);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &brownsd);
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &minab);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &newk);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &othk);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &thek);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &bigtime);
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   fscanf(datfile,"%d", &ncons);
   while ( getc(datfile) != '\n');  

   hightime = dvec(nnodes);
   lowtime = dvec(nnodes);
   tipcons = dvec(nnodes);

   for(ktrack = 0; ktrack < nnodes; ktrack++)
   {
      lowtime[ktrack] = 0.0;
      hightime[ktrack] = bigtime;
   }

   for(ktrack=1; ktrack <= ncons; ktrack++)
   {
      fscanf(datfile, "%1c%d%lf", &constype, &whichnode, &thecons);
      if(constype == 'L')
      {
         lowtime[whichnode] = thecons;
      }
      else
      {
         if(constype == 'U')
         {
            hightime[whichnode] = thecons;
         }
         else
         {
            printf("Problem: constype is %1c ... What does this mean?\n", 
               constype);
         }
      }
      while ( getc(datfile) != '\n');  
   }
   for(ktrack=0; ktrack < ntips; ktrack++)
   {
      tipcons[ktrack] = 0.0;
   }
   fscanf(datfile, "%d", &ntipcons);
   while ( getc(datfile) != '\n');  
   for(ktrack=1; ktrack <= ntipcons; ktrack++)
   {
      fscanf(datfile, "%1c%d%lf", &constype, &whichnode, &thecons);
      if(constype == 'T')
      {
         tipcons[whichnode] = thecons;
      }
      else
      {
         printf("Problem tipcons: constype is %1c ... What does this mean?\n",
            constype);
      }
      while ( getc(datfile) != '\n');
   }
   fscanf(datfile,"%d", &nodata);
   while ( getc(datfile) != '\n');
   fscanf(datfile, "%d", &commonbrown);
   while ( getc(datfile) != '\n');

   fclose(datfile);

   freecvec(cstring,20);
}


void initialize(time,rate,rndstart)
int rndstart;
double *time, **rate;
{

   double rnd();
   double exptial();
   double logmvn();
 
   int ktrack, ktrack2, highnode;
   int maxdepth, thisparent,depth[MAXNODE];
   double *mfac, *tblen, holdptime;
   double temptime,maxdis, *dist;


   mfac = dvec(nnodes);
   tblen = dvec(nbranches);
   curmvn = dvec(ngenes);


   for(ktrack=0; ktrack < ntips; ktrack++)
   {
      mfac[ktrack] = 1.0;
      depth[ktrack] = 0;
   }
   for(ktrack=ntips; ktrack < nnodes; ktrack++)
   {
      mfac[ktrack] = 1.0;
      maxdepth = depth[tonode[ktrack][1]];
      for(ktrack2 = 2; ktrack2 <= tonode[ktrack][0]; ktrack2++)
      {
         maxdepth = max2(maxdepth, depth[tonode[ktrack][ktrack2]]);
      }
      depth[ktrack] = 1+ maxdepth;
   }
   for(ktrack = ntips; ktrack < nnodes-1; ktrack++)
   {
      a[ktrack] = minab*depth[ktrack];
      b[ktrack] = minab;
      a[ktrack] -= 1.0;
      b[ktrack] -= 1.0;
   }        
   viral = 0;
   for(ktrack = 0; ktrack < ntips; ktrack++)
   {
      time[ktrack] = tipcons[ktrack];
      if(time[ktrack] > SMALLNUM)
      {
         viral = 1;
      }
   }
   for(ktrack=1; ktrack <= tonode[nnodes-1][0]; ktrack++)
   {
      if(tonode[nnodes-1][ktrack] >= ntips || 
         time[tonode[nnodes-1][ktrack]] == 0.0)
      {
         consnode = tonode[nnodes-1][ktrack];
         break;
      }
   }
   printf("starttype = %1d ... ", rndstart);
   if( rndstart != 1)
   {
      printf("starting from a hopefully good guess\n");
      maxdis = 0.0;
      dist = dvec(ntips-1);
      for(ktrack = 0; ktrack < ntips; ktrack++)
      {
         highnode = tonode[ktrack][MAXOFF];
         dist[ktrack] = max2(0.01,mean[0][ktrack]);
         while(tonode[highnode][MAXOFF] != -1)
         {
            dist[ktrack] += max2(0.01,mean[0][highnode]);
            highnode = tonode[highnode][MAXOFF];
         }
         if(maxdis < dist[ktrack])
         {
            maxdis = dist[ktrack];
         }
      }

      rate[0][nnodes-1] = rtrate;
      rate[0][consnode] = rtrate;
      time[nnodes-1] = maxdis/rtrate;
      for(ktrack = nnodes-2; ktrack >= ntips; ktrack--)
      {
         time[ktrack] = time[tonode[ktrack][MAXOFF]] - 
            max2(mean[0][ktrack]/rtrate, 0.01);
      }
      if(brownmean > SMALLNUM)
      {
         for(ktrack = nnodes-2; ktrack >= 0; ktrack--)
         {
            if(ktrack != consnode)
            {
               rate[0][ktrack] = 2.0*max2(0.01,mean[0][ktrack])/(
                  time[tonode[ktrack][MAXOFF]]- time[ktrack]) - 
                  rate[0][tonode[ktrack][MAXOFF]];
               rate[0][ktrack] = max2(rate[0][ktrack], SMALLRATE);
            }
         }
      }
      else
      {
          for(ktrack=0; ktrack < nnodes-1; ktrack++)
          {
             rate[0][ktrack] = rate[0][nnodes-1];
          }
      }
      rate[0][nnodes] = brownmean; 
         /* initialize BROWNIAN motion constant */
   }
   else
   {
      printf("starting at semi-random guess\n");
      time[nnodes-1] = exptial(1.0/rttm);
      if(brownmean > SMALLNUM)
      {
         rate[0][nnodes] = exptial(1.0/brownmean);
      }
      else
      {
         rate[0][nnodes] = 0.0;
      }
      if(brownmean > SMALLNUM)
      {
         for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
         {
            rate[0][ktrack] = 0.5+rnd(seed)*1.5;
         }
         rate[0][consnode] = rate[0][nnodes-1];
      }
      else
      {
         rate[0][nnodes-1] = 0.1*rtrate + 1.9*rtrate*rnd(seed);
         for(ktrack=0; ktrack < nnodes-1; ktrack++)
         {
            rate[0][ktrack] = rate[0][nnodes-1];
         }
      }

      for(ktrack=nnodes-2; ktrack >= ntips; ktrack--)
      {
         if(mean[0][ktrack] > TINYLEN)
         {
            time[ktrack] = time[tonode[ktrack][MAXOFF]] - 
               (mean[0][ktrack]/(0.5*(rate[0][ktrack]+
               rate[0][tonode[ktrack][MAXOFF]])));
         }
         else 
         {
            time[ktrack] = (1.0-TINYLEN)*time[tonode[ktrack][MAXOFF]];
         }
         if(time[ktrack] < TINYTIME)
         {
            time[ktrack] = (1.0-TINYTIME)*time[tonode[ktrack][MAXOFF]];
         }
      }  

   }

   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      tblen[ktrack] = (time[tonode[ktrack][MAXOFF]] - time[ktrack])*
         (rate[0][tonode[ktrack][MAXOFF]] + rate[0][ktrack])* 0.5;
   }

   for(ktrack= nnodes-1; ktrack >= ntips; ktrack--)
   {
      thisparent = tonode[ktrack][MAXOFF];
      if(ktrack != nnodes-1)
      {
         mfac[ktrack] = mfac[thisparent];
         time[ktrack] *= mfac[ktrack];
         holdptime = time[thisparent];
      }
      else
      {
         holdptime = hightime[ktrack];
      }
      temptime = time[ktrack];
      if(temptime > min2(holdptime, hightime[ktrack])
         || time[ktrack] < lowtime[ktrack])
      {
         if(rndstart == 1)
         {
            temptime = lowtime[ktrack] + rnd(seed)*(min2(holdptime,
               hightime[ktrack])- lowtime[ktrack]);
         }
         else
         {
            temptime = lowtime[ktrack] + 0.5*(min2(holdptime
               ,hightime[ktrack])- lowtime[ktrack]);
         }
         mfac[ktrack] *= temptime/time[ktrack];
      }
      time[ktrack] = temptime;
   }      
   rate[0][nnodes-1] = tblen[consnode]/(time[nnodes-1] - time[
      consnode]);
   if(brownmean > SMALLNUM)
   {
      rate[0][consnode] = rate[0][nnodes-1];
      for(ktrack = nnodes -2; ktrack >=0; ktrack--)
      {
         if(ktrack != consnode)
         {
            thisparent = tonode[ktrack][MAXOFF];
            rate[0][ktrack] = 2*tblen[ktrack]/(time[thisparent] -
               time[ktrack]) - rate[0][thisparent];
            if(rate[0][ktrack] < SMALLRATE)
            {
               rate[0][ktrack] = SMALLRATE;
            }
         }
      }
   }
   else
   {
      for(ktrack = 0; ktrack < nnodes-1; ktrack++)
      {
         rate[0][ktrack] = rate[0][nnodes-1];
      }
   }

   for(ktrack2=0; ktrack2 < nnodes-1; ktrack2++)
   {
      thisparent = tonode[ktrack2][MAXOFF];
      tblen[ktrack2] = 0.5* (rate[0][ktrack2]+rate[0][thisparent])*
         (time[thisparent] - time[ktrack2]);   
   }
   curmvn[0] = logmvn(tblen,0);
   for(ktrack=1; ktrack <=ngenes; ktrack++)
   {
      for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
      {
         rate[ktrack][ktrack2] = rate[0][ktrack2];
      }
      for(ktrack2=0; ktrack2 < nnodes-1; ktrack2++)
      {
         thisparent = tonode[ktrack2][MAXOFF];
         tblen[ktrack2] = 0.5* (rate[ktrack][ktrack2]+rate[ktrack][thisparent])*
            (time[thisparent] - time[ktrack2]);
      }
      curmvn[ktrack] = logmvn(tblen,ktrack);
   }

   for(ktrack = 0 ; ktrack < nnodes; ktrack++)
   {
      printf("%3d:t=%9.4lf hight=%9.4lf lowt=%9.4lf rt=%9.4lf pnode=%3d\n",
         ktrack,time[ktrack],hightime[ktrack], lowtime[ktrack], rate[0][ktrack],
         tonode[ktrack][MAXOFF]);
   }

   printf("Each gene initialized to have same rate at node 0!\n");

   freedvec(mfac);
   freedvec(tblen);
}

void dothemcmc(time,rate,samptime, samprate, sufname)
char *sufname;
double *time, **rate, **samptime, ***samprate;
{
   int whichsamp, ktrack, ktrack2, ktrack3;
   FILE *sampfile;
   double **logr; /* save time by precalculating log of rates */
   char sampfilename[100];

   void timechange();
   void ratechange();
   void changebrown();
   void consflate();
   void rateflate();
   void timeflate();
   void bothflate();
   void partrateflate();
   void parttimeflate();
   void partbothflate();
   void partconsflate();

   logr = dmat(ngenes,nnodes);
   strcpy(sampfilename,"samp.");
   strcat(sampfilename,sufname); 

   /* Ziheng moved the fopen statement out of the loop, and used fflush() */
   sampfile = gfopen(sampfilename,"w");

   for(ktrack = 0; ktrack < nnodes; ktrack++)
   {
      for(ktrack2=0; ktrack2 <=ngenes; ktrack2++)
      {
         logr[ktrack2][ktrack] = log(rate[ktrack2][ktrack]);
      }
   }

   for(ktrack = 0; ktrack <= burnin+ (numsamps-1)*sampfreq; ktrack++)
   {

      if((ktrack+1)%500==0) 
            fprintf(stderr, "\rMCMC %4d / %d ", ktrack+1, burnin+ (numsamps-1)*sampfreq);

      if( (ktrack >= burnin) && ((ktrack - burnin)%sampfreq== 0 ))
      {
         whichsamp = (ktrack-burnin)/sampfreq;
         for(ktrack2=0; ktrack2<nnodes; ktrack2++)
         {
            fprintf(sampfile, "%10.5lf ", time[ktrack2]);
            samptime[ktrack2][whichsamp] = time[ktrack2];
            for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
            {
               samprate[ktrack3][ktrack2][whichsamp] = rate[ktrack3][ktrack2];
            }
         }
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            samprate[ktrack2][nnodes][whichsamp] = rate[ktrack2][nnodes];
         }
         for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
         {
            for(ktrack2=0; ktrack2<=nnodes; ktrack2++)
            {
               fprintf(sampfile, "%10.5lf ", rate[ktrack3][ktrack2]);
            }
         }
         fprintf(sampfile,"\n");
         fflush(sampfile);

      }
      timechange(time,rate,logr);
      if(brownmean > SMALLNUM)
      {
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            ratechange(time,rate,logr, ktrack2);
         }
         changebrown(time,rate,logr);
         consflate(time,rate,logr);
      }
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         rateflate(time,rate,logr,ktrack2);
      }
      timeflate(time,rate,logr);
      bothflate(time,rate,logr);
      /* below is necessary to insure that there is a subtree below
         root on which to inflate/deflate -- if nintnodes =1 then there
         is only 1 internal non-root node and this might be constrained
         to have same rate as root node */
      if(nintnodes > 1)
      {
         if(brownmean > SMALLNUM)
         {
            partconsflate(time,rate,logr);
            for(ktrack2=0;ktrack2 <= ngenes; ktrack2++)
            {
               partrateflate(time,rate,logr, ktrack2);
            }
            partbothflate(time,rate,logr);
         }
         parttimeflate(time,rate,logr);
      }
   }
   freedmat(logr, ngenes, nnodes);
   fclose(sampfile);
}



void parttimeflate(time,rate,logr)
double *time, **rate, **logr;
{
   double multiratio();
   double logmvn();
   double rnd();

   double ratio, randnum, invfac, flatefac, holdthis;
   double jterm, maxcurratio, maxnewratio;
   double lbound,ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   double lmin, lmax;
   static double *newmvn;
   static double summvn;
   static int **belownode;
   static double halfk, neghalfk, **brlen, **propbrlen;
   static double invothk;
   static double *newtime, **newrate, **newlogr;
   int ktrack, ktrack2;
   int thisnode, parentnode, thisparent;
   int chooseparent, choosenode;
   static int hereyet=0;
 
   if(hereyet ==0)
   {
      brlen = dmat(ngenes,nbranches);
      propbrlen = dmat(ngenes,nbranches);
 
      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      belownode = imat(nnodes,nnodes);
      for(ktrack = 0; ktrack <= nnodes; ktrack++)
      {
         for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
         {
            belownode[ktrack][ktrack2] = 0;
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         belownode[ktrack][ktrack] = 1;
/* belownode[k][nnodes] counts the number of nodes that are tips below node k
   belownode[nnodes][k] counts total number of node below and including node k
*/
         for(ktrack2= ktrack+1; ktrack2 < nnodes; ktrack2++)
         {
            if(belownode[ktrack2][tonode[ktrack][MAXOFF]] == 1)
            {
               belownode[ktrack2][ktrack] = 1;
            }
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         for(ktrack2 = 0; ktrack2 <= ktrack ; ktrack2++)
         {
            if(belownode[ktrack][ktrack2] == 1 )
            {
               if(ktrack2 < ntips)
               {
                  belownode[ktrack][nnodes]++;
               }
               belownode[nnodes][ktrack]++;
            }
         }
      }   


      newtime = dvec(nnodes);
      newrate = dmat(ngenes,nnodes);
      newlogr = dmat(ngenes,nnodes);
      newmvn = dvec(ngenes);
   }
   choosenode = consnode;
   while(choosenode == consnode)
   {
      choosenode = (int) ((nnodes-ntips-1)*rnd(seed)+ntips);
   }
   chooseparent = tonode[choosenode][MAXOFF];


   mincurratio = min2(time[chooseparent]/time[choosenode],
      hightime[choosenode]/time[choosenode]);
   maxcurratio = lowtime[choosenode]/time[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {  
         maxcurratio = max2(maxcurratio, lowtime[ktrack]/time[ktrack]);
         mincurratio = min2(mincurratio, hightime[ktrack]/time[ktrack]);
      }  
   }   

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(maxcurratio == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {   
      lmax = log(maxcurratio);
   }
   if(othk <  -2.0*lmax)
   {
      lownum = 0.0;
   }
   else
   {   
      lownum = 0.5+ lmax*invothk;
   }
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         newrate[ktrack2][ktrack] = rate[ktrack2][ktrack];
         newlogr[ktrack2][ktrack] = logr[ktrack2][ktrack];
      }
      if(ktrack >= ntips && belownode[choosenode][ktrack] == 1)
      {  
         newtime[ktrack] = flatefac*time[ktrack];
      }   
      else
      {   
         newtime[ktrack] = time[ktrack];
      }  
   }   
   for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
   {

      newrate[ktrack2][nnodes] = rate[ktrack2][nnodes];
   }
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         holdthis = 0.5*(rate[ktrack2][thisparent]+rate[ktrack2][ktrack]);
         brlen[ktrack2][ktrack] = holdthis* (time[thisparent] - time[ktrack]);
         propbrlen[ktrack2][ktrack] = holdthis* 
            (newtime[thisparent] - newtime[ktrack]);
      }
   }

   minnewratio = min2(newtime[chooseparent]/newtime[choosenode],
      hightime[choosenode]/newtime[choosenode]);
   maxnewratio = lowtime[choosenode]/newtime[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {  
         maxnewratio = max2(maxnewratio, lowtime[ktrack]/newtime[ktrack]);
         minnewratio = min2(minnewratio, hightime[ktrack]/newtime[ktrack]);
      }  
   }   

   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(lmax == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {   
      lmax = log(maxnewratio);
   }
   if(othk > -2.0*lmax)
   {
      lownum = 0.5+ lmax*invothk;
   }
   else
   {   
      lownum = 0.0;
   }
   ratio = (jterm /(highnum-lownum))*
      exp((belownode[ktrack][nnodes] - belownode[nnodes][ktrack])*loginvfac)
      *multiratio(newtime,newrate,newlogr,time, rate, logr);
   summvn = 0.0;
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      newmvn[ktrack] = logmvn(propbrlen[ktrack],ktrack);
      summvn += newmvn[ktrack] - curmvn[ktrack];
   }
   ratio *= exp( summvn); 
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack <= choosenode; ktrack++)
      {  
         time[ktrack] = newtime[ktrack];
      }  
      for(ktrack=0; ktrack <= ngenes; ktrack++)
      {
         curmvn[ktrack] = newmvn[ktrack];
      }   
      sucparttime++;
   }
   else
   {   
      failparttime++;
   }
}

void partbothflate(time,rate,logr)
double *time, **rate, **logr;
{
   double multiratio();
   double logmvn();
   double rnd();

   double ratio, randnum, invfac, flatefac ;
   double jterm, maxcurratio, maxnewratio;
   double lbound,ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   double lmin, lmax;
   static double *newmvn;
   static double summvn;

   static int **belownode;
   static double halfk, neghalfk, **brlen, **propbrlen;
   static double invothk;
   static double *newtime, **newrate, **newlogr;
   int ktrack, ktrack2;
   int thisnode, parentnode, thisparent;
   int chooseparent, choosenode;
   static int hereyet=0;
 
   if(hereyet ==0)
   {
      brlen = dmat(ngenes,nbranches);
      propbrlen = dmat(ngenes,nbranches);
      newmvn = dvec(ngenes);
      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      belownode = imat(nnodes,nnodes);
      for(ktrack = 0; ktrack <= nnodes; ktrack++)
      {
         for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
         {
            belownode[ktrack][ktrack2] = 0;
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         belownode[ktrack][ktrack] = 1;
/* belownode[k][nnodes] counts the number of nodes that are tips below node k
   belownode[nnodes][k] counts total number of node below and including node k
*/
         for(ktrack2= ktrack+1; ktrack2 < nnodes; ktrack2++)
         {
            if(belownode[ktrack2][tonode[ktrack][MAXOFF]] == 1)
            {
               belownode[ktrack2][ktrack] = 1;
            }
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         for(ktrack2 = 0; ktrack2 <= ktrack ; ktrack2++)
         {
            if(belownode[ktrack][ktrack2] == 1 )
            {
               if(ktrack2 < ntips)
               {
                  belownode[ktrack][nnodes]++;
               }
               belownode[nnodes][ktrack]++;
            }
         }
      }   

      newtime = dvec(nnodes);
      newrate = dmat(ngenes,nnodes);
      newlogr = dmat(ngenes,nnodes);

   }
   choosenode = consnode;
   while(choosenode == consnode)
   {
      choosenode = (int) ((nnodes-ntips-1)*rnd(seed)+ntips);
   }
   chooseparent = tonode[choosenode][MAXOFF];


   mincurratio = min2(time[chooseparent]/time[choosenode],
      hightime[choosenode]/time[choosenode]);
   maxcurratio = lowtime[choosenode]/time[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {  
         maxcurratio = max2(maxcurratio, lowtime[ktrack]/time[ktrack]);
         mincurratio = min2(mincurratio, hightime[ktrack]/time[ktrack]);
      }  
   }   

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(maxcurratio == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {   
      lmax = log(maxcurratio);
   }
   if(othk <  -2.0*lmax)
   {
      lownum = 0.0;
   }
   else
   {   
      lownum = 0.5+ lmax*invothk;
   }
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(belownode[choosenode][ktrack] == 1)
      {  
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            newrate[ktrack2][ktrack] = invfac*rate[ktrack2][ktrack];
            newlogr[ktrack2][ktrack] = logr[ktrack2][ktrack] + loginvfac;
         }
         if(ktrack >= ntips)
         {
            newtime[ktrack] = flatefac*time[ktrack];
         }
         else
         {
            newtime[ktrack] = time[ktrack];
         }
      }   
      else
      {   
         newtime[ktrack] = time[ktrack];
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            newrate[ktrack2][ktrack] = rate[ktrack2][ktrack];
            newlogr[ktrack2][ktrack] = logr[ktrack2][ktrack];
         }
      }  
   }   
   for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
   {
      newrate[ktrack2][nnodes] = rate[ktrack2][nnodes];
   }
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         brlen[ktrack2][ktrack] = 0.5*(rate[ktrack2][thisparent]+
            rate[ktrack2][ktrack])* (time[thisparent] - time[ktrack]);
         propbrlen[ktrack2][ktrack] = 0.5*(newrate[ktrack2][thisparent] + 
            newrate[ktrack2][ktrack])* (newtime[thisparent] - newtime[ktrack]);
      }
   }

   minnewratio = min2(newtime[chooseparent]/newtime[choosenode],
      hightime[choosenode]/newtime[choosenode]);
   maxnewratio = lowtime[choosenode]/newtime[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {  
         maxnewratio = max2(maxnewratio, lowtime[ktrack]/newtime[ktrack]);
         minnewratio = min2(minnewratio, hightime[ktrack]/newtime[ktrack]);
      }  
   }   

   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(lmax == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {   
      lmax = log(maxnewratio);
   }
   if(othk > -2.0*lmax)
   {
      lownum = 0.5+ lmax*invothk;
   }
   else
   {   
      lownum = 0.0;
   }
   ratio = (jterm /(highnum-lownum))*
      exp(((ngenes+1.0)*belownode[nnodes][choosenode]- (
      belownode[nnodes][choosenode]-belownode[choosenode][nnodes]))*loginvfac)*
      multiratio(newtime,newrate,newlogr,time, rate, logr);
   summvn = 0.0;
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      newmvn[ktrack] = logmvn(propbrlen[ktrack],ktrack);
      summvn += newmvn[ktrack] - curmvn[ktrack];
   }
   ratio *= exp( summvn);

   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         rate[ktrack2][nnodes] = newrate[ktrack2][nnodes];
         curmvn[ktrack2] = newmvn[ktrack2];
      }
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {  
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            rate[ktrack2][ktrack] = newrate[ktrack2][ktrack];
            logr[ktrack2][ktrack] = newlogr[ktrack2][ktrack];
         }
         time[ktrack] = newtime[ktrack];
      }  
      sucpartboth++;
   }
   else
   {   
      failpartboth++;
   }
}

void partrateflate(time,rate,logr, whichgene)
int whichgene;
double *time, **rate, **logr;
{
   double probratio();
   double logmvn();
   double rnd();

   double ratio, randnum, invfac, flatefac, holdthis;
   double loginvfac;
   double *trate, *tlogr,runlog;
   static double *newmvn;
   static int **belownode;
   static double halfk, neghalfk, *brlen, *propbrlen;
   static double invothk;
   static double *newtime, *newrate, *newlogr;
   int ktrack, ktrack2;
   int thisnode, parentnode, thisparent;
   int chooseparent, choosenode;
   static int hereyet=0;
 
   if(hereyet ==0)
   {
      brlen = dvec(nbranches);
      propbrlen = dvec(nbranches);
      newmvn = dvec(ngenes);
      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      belownode = imat(nnodes,nnodes);
      for(ktrack = 0; ktrack <= nnodes; ktrack++)
      {
         for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
         {
            belownode[ktrack][ktrack2] = 0;
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         belownode[ktrack][ktrack] = 1;
/* belownode[k][nnodes] counts the number of nodes that are tips below node k
   belownode[nnodes][k] counts total number of node below and including node k
*/
         for(ktrack2= ktrack+1; ktrack2 < nnodes; ktrack2++)
         {
            if(belownode[ktrack2][tonode[ktrack][MAXOFF]] == 1)
            {
               belownode[ktrack2][ktrack] = 1;
            }
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         for(ktrack2 = 0; ktrack2 <= ktrack ; ktrack2++)
         {
            if(belownode[ktrack][ktrack2] == 1 )
            {
               if(ktrack2 < ntips)
               {
                  belownode[ktrack][nnodes]++;
               }
               belownode[nnodes][ktrack]++;
            }
         }
      }   

      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);
   }
   trate = rate[whichgene];
   tlogr = logr[whichgene];
   choosenode = consnode;
   while(choosenode == consnode)
   {
      choosenode = (int) ((nnodes-ntips-1)*rnd(seed)+ntips);
   }
   runlog = 0.0;
   chooseparent = tonode[choosenode][MAXOFF];


   loginvfac = halfk - othk*rnd(seed);
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      newtime[ktrack] = time[ktrack];
      if(belownode[choosenode][ktrack] == 1)
      {  
         newrate[ktrack] = flatefac*trate[ktrack];
         newlogr[ktrack] = tlogr[ktrack] - loginvfac;
      }   
      else
      {   
         newrate[ktrack] = trate[ktrack];
         newlogr[ktrack] = tlogr[ktrack];
      }  
   }   
   newrate[nnodes] = trate[nnodes];
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
         thisparent = tonode[ktrack][MAXOFF];
         holdthis = 0.5*(time[thisparent] - time[ktrack]);
         brlen[ktrack] = holdthis*(trate[thisparent]+trate[ktrack]);
         propbrlen[ktrack] = holdthis*(newrate[thisparent] + newrate[ktrack]);
   }

   newmvn[whichgene] = logmvn(propbrlen,whichgene);
   ratio = exp(newmvn[whichgene] -curmvn[whichgene]- loginvfac*
      belownode[nnodes][choosenode]) *
      probratio(newtime,newrate,newlogr,time, trate, tlogr);
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack <= choosenode; ktrack++)
      {  
         trate[ktrack] = newrate[ktrack];
         tlogr[ktrack] = newlogr[ktrack];
      }  
      curmvn[whichgene] = newmvn[whichgene];
      sucpartrate[whichgene]++;
   }
   else
   {   
      failpartrate[whichgene]++;
   }
}

void partconsflate(time,rate,logr)
double *time, **rate, **logr;
{
   double multiratio();
   double logcondmvn();
   double logmvn();
   double rnd();

   double ratio, randnum, invfac, flatefac ;
   double jterm, maxcurratio, maxnewratio;
   double lbound,ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   double runlog, lmin, lmax, blen;
   static double *newmvn;
   static double summvn;
   static int **belownode;
   static double halfk, neghalfk, **brlen, **propbrlen;
   static double invothk;
   static double *newtime, **newrate, **newlogr;
   int ktrack, ktrack2;
   int thisnode, parentnode, thisparent;
   int chooseparent, choosenode;
   static int hereyet=0;

   if(hereyet ==0)
   {
      brlen = dmat(ngenes,nbranches);
      propbrlen = dmat(ngenes,nbranches);
      newmvn = dvec(ngenes);

      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      belownode = imat(nnodes,nnodes);
      for(ktrack = 0; ktrack <= nnodes; ktrack++)
      {
         for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
         {
            belownode[ktrack][ktrack2] = 0;
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         belownode[ktrack][ktrack] = 1;
/* belownode[k][nnodes] counts the number of nodes that are tips below node k
   belownode[nnodes][k] counts total number of node below and including node k
*/
         for(ktrack2= ktrack+1; ktrack2 < nnodes; ktrack2++)
         {
            if(belownode[ktrack2][tonode[ktrack][MAXOFF]] == 1)
            {
               belownode[ktrack2][ktrack] = 1;
            }
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         for(ktrack2 = 0; ktrack2 <= ktrack ; ktrack2++)
         {
            if(belownode[ktrack][ktrack2] == 1 )
            {
               if(ktrack2 < ntips)
               {
                  belownode[ktrack][nnodes]++;
               }
               belownode[nnodes][ktrack]++;
            }
         }
      }   

   
      newtime = dvec(nnodes);
      newrate = dmat(ngenes,nnodes);
      newlogr = dmat(ngenes,nnodes);

   }
   choosenode = consnode;
   while(choosenode == consnode)
   {
      choosenode = (int) ((nnodes-ntips-1)*rnd(seed)+ntips);
   }
   runlog = 0.0;
   chooseparent = tonode[choosenode][MAXOFF];
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      blen = time[thisparent] - time[ktrack];
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         brlen[ktrack2][ktrack] = 0.5*(rate[ktrack2][thisparent]+
            rate[ktrack2][ktrack])*blen;         
         propbrlen[ktrack2][ktrack] = brlen[ktrack2][ktrack];
      }
   }

   mincurratio = min2(time[chooseparent]/time[choosenode],
      hightime[choosenode]/time[choosenode]);
   maxcurratio = lowtime[choosenode]/time[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {
         maxcurratio = max2(maxcurratio, lowtime[ktrack]/time[ktrack]);
         mincurratio = min2(mincurratio, hightime[ktrack]/time[ktrack]);
      }                        
   }      
   for(ktrack =0 ; ktrack < ntips; ktrack++)
   {
      if( time[ktrack] > 0.0 &&  belownode[choosenode][ktrack] ==1)
      {
         parentnode = tonode[ktrack][MAXOFF];
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            if( time[parentnode] * rate[ktrack2][ktrack] < time[ktrack]*(
               rate[ktrack2][ktrack]+rate[ktrack2][parentnode]) )
            {
               mincurratio = min2(mincurratio, (time[ktrack]*
                  rate[ktrack2][parentnode])/ (time[ktrack]*(
                  rate[ktrack2][ktrack]+rate[ktrack2][parentnode]) -
                  time[parentnode]*rate[ktrack2][ktrack]) );
            }
         }
      }
   }

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   } 
   else
   {
      highnum = 0.5+ lmin*invothk;
   }
   if(maxcurratio == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {
      lmax = log(maxcurratio);
   }
   if(othk <  -2.0*lmax)
   {
      lownum = 0.0;
   }
   else
   {
      lownum = 0.5+ lmax*invothk;
   }
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;
 
   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(belownode[choosenode][ktrack] == 1)
      {
         if(ktrack < ntips && time[ktrack] > 0.0)
         {
            parentnode = tonode[ktrack][MAXOFF];
            newtime[ktrack] = time[ktrack];
            for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
            {
               newrate[ktrack2][ktrack] = (time[parentnode] - time[ktrack])*(
                  rate[ktrack2][parentnode] + rate[ktrack2][ktrack])/(
                  newtime[parentnode] - time[ktrack]) -
                  newrate[ktrack2][parentnode];
               newlogr[ktrack2][ktrack] = log(newrate[ktrack2][ktrack]);
            }
         }
         else
         {
            for(ktrack2=0; ktrack2 <= ngenes;ktrack2++)
            {
               newrate[ktrack2][ktrack] = invfac*rate[ktrack2][ktrack];
               newlogr[ktrack2][ktrack] = loginvfac + logr[ktrack2][ktrack];
            }
            newtime[ktrack] = flatefac*time[ktrack];
         } 
      }
      else
      {
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            newlogr[ktrack2][ktrack] = logr[ktrack2][ktrack];
            newrate[ktrack2][ktrack] = rate[ktrack2][ktrack];
         }
         newtime[ktrack] = time[ktrack];
      }
   }
   for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
   {
      newrate[ktrack2][nnodes] = rate[ktrack2][nnodes];
      propbrlen[ktrack2][choosenode] = (newtime[chooseparent] -
         newtime[choosenode])* 0.5*(newrate[ktrack2][chooseparent] + 
         newrate[ktrack2][choosenode]);
   }
   minnewratio = min2(newtime[chooseparent]/newtime[choosenode], 
      hightime[choosenode]/newtime[choosenode]);
   maxnewratio = lowtime[choosenode]/newtime[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {
         maxnewratio = max2(maxnewratio, lowtime[ktrack]/newtime[ktrack]);
         minnewratio = min2(minnewratio, hightime[ktrack]/newtime[ktrack]);
      }                        
   }      
   for(ktrack =0 ; ktrack < ntips; ktrack++)
   {
      if( belownode[choosenode][ktrack] ==1)
      {
         parentnode = tonode[ktrack][MAXOFF];
         if(time[ktrack] > 0.0)
         {
            runlog += (ngenes+1.0)* log((time[parentnode] - time[ktrack])/
               (newtime[parentnode] - newtime[ktrack]));   
         }
         else
         {
            runlog += (ngenes+1.0)*loginvfac;
         }
         if( newtime[ktrack] > 0.0 )
         {  
            parentnode = tonode[ktrack][MAXOFF];
            for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
            {
               if( newtime[parentnode] * newrate[ktrack2][ktrack] < 
                  newtime[ktrack]*(newrate[ktrack2][ktrack]+
                  newrate[ktrack2][parentnode]) )
               {
                  minnewratio = min2(minnewratio, (newtime[ktrack]*
                     newrate[ktrack2][parentnode])/ (newtime[ktrack]*(
                     newrate[ktrack2][ktrack]+newrate[ktrack2][parentnode]) -
                     newtime[parentnode]*newrate[ktrack2][ktrack]) );
               }
            }
         }
      }   
   }   

   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {  
      highnum = 1.0;
   }  
   else  
   {  
      highnum = 0.5+ lmin*invothk;
   }  
   if(lmax == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {
      lmax = log(maxnewratio);
   }
   if(othk > -2.0*lmax)
   {
      lownum = 0.5+ lmax*invothk;
   }
   else
   {
      lownum = 0.0;
   }
   ratio = (jterm /(highnum-lownum))* exp(runlog+loginvfac*ngenes*(
      belownode[nnodes][choosenode]-belownode[choosenode][nnodes]) ) *
      multiratio(newtime,newrate,newlogr,time, rate, logr); 
/* DEBUG
   if(fulldata ==1)
   {
      for(ktrack=0; ktrack <= ngenes; ktrack++)
      {
         ratio *= exp(logcondmvn(propbrlen[ktrack],choosenode, ktrack) -
            logcondmvn(brlen[ktrack],choosenode,ktrack)); 
      }
   }
   else
   {
*/
      summvn = 0.0;
      for(ktrack=0; ktrack <= ngenes; ktrack++)
      {
         newmvn[ktrack] = logmvn(propbrlen[ktrack],ktrack);
         summvn += newmvn[ktrack] - curmvn[ktrack];
      }
      ratio *= exp( summvn);


/* DEBUG
   }
*/
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack <= choosenode; ktrack++)
      {
         time[ktrack] = newtime[ktrack];
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            rate[ktrack2][ktrack] = newrate[ktrack2][ktrack];
            logr[ktrack2][ktrack] = newlogr[ktrack2][ktrack];
         }
      }
      sucpartcons++;
      for(ktrack=0; ktrack <= ngenes; ktrack++)
      {
         curmvn[ktrack] = newmvn[ktrack];
      }
   }   
   else
   {
      failpartcons++;
   }
}

void timeflate(time,rate,logr)
double *time, **rate, **logr;
{
   double multiratio();
   double rnd();
   double logmvn();

   double ratio, randnum, invfac, flatefac, holdthis ;
   double jterm, maxcurratio, maxnewratio;
   double lmin, lmax,lbound, ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   static double *newmvn;
   static double summvn;
   static double halfk, neghalfk, *thelowcon, invothk;
   static double *thehighcon;
   static double *newtime, **newrate, **newlogr;
   static double **brlen, **propbrlen;
   int ktrack, ktrack2, breakthis;
   int thisnode, parentnode;
   static int hereyet=0, numlowcons, *nodelowcon, numhighcons, *nodehighcon;

   if(hereyet ==0)
   { 
      newmvn = dvec(ngenes);
      brlen = dmat(ngenes,nnodes);
      propbrlen = dmat(ngenes,nnodes);
      thelowcon = dvec(nnodes);
      thehighcon = dvec(nnodes);
      nodelowcon = ivec(nnodes);
      nodehighcon = ivec(nnodes);
      hereyet = 1;
      numhighcons = 0;
      numlowcons = 0;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      for(ktrack = ntips; ktrack < nnodes; ktrack++)
      {
         breakthis = 0;
         for(ktrack2 = 1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thisnode = tonode[ktrack][ktrack2];
            if(lowtime[ktrack] == lowtime[thisnode] &&
               (thisnode >= ntips ||
               lowtime[thisnode] == 0.0) )
            {
               breakthis = 1;
               break;
            }
         }
         if( breakthis == 0)
         {
            numlowcons++;
            thelowcon[numlowcons] = lowtime[ktrack];
            nodelowcon[numlowcons] = ktrack;
         }
         if(ktrack == nnodes-1)
         {
            numhighcons++;
            thehighcon[numhighcons] = hightime[ktrack];
            nodehighcon[numhighcons] = ktrack;
         }
         else
         {
            if(hightime[ktrack] != hightime[tonode[ktrack][MAXOFF]])
            {
               numhighcons++;
               thehighcon[numhighcons] = hightime[ktrack];
               nodehighcon[numhighcons] = ktrack;
            }
         }
      }   
      newtime = dvec(nnodes);
      newrate = dmat(ngenes,nnodes);
      newlogr = dmat(ngenes,nnodes);
   }
   mincurratio = thehighcon[1]/time[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      mincurratio = min2( mincurratio,
         thehighcon[ktrack]/time[nodehighcon[ktrack]]);
   }

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons == 0)
   {
      lownum = 0.0;
      lmax = BIGNEGLOG;
   }
   else
   {   
      maxcurratio = thelowcon[1]/time[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxcurratio = max2(maxcurratio,
            thelowcon[ktrack]/time[nodelowcon[ktrack]]);
      }  
      if(maxcurratio == 0.0)
      {  
         lmax = BIGNEGLOG;
      }  
      else
      {   
         lmax = log(maxcurratio);
      }  
      if(othk < -2.0*lmax)
      {  
         lownum = 0.0;
      }  
      else
      {   
         lownum = 0.5+ lmax*invothk;
      }  
   }   
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(ktrack >= ntips)
      {
         newtime[ktrack] = flatefac*time[ktrack];
      }
      else
      {
         newtime[ktrack] = time[ktrack];
      }
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         newrate[ktrack2][ktrack] = rate[ktrack2][ktrack];
         newlogr[ktrack2][ktrack] = logr[ktrack2][ktrack];
      }
   }   
   for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
   {
      newrate[ktrack2][nnodes] = rate[ktrack2][nnodes];
   }
   minnewratio = thehighcon[1]/newtime[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      minnewratio = min2(minnewratio, thehighcon[ktrack]/
         newtime[nodehighcon[ktrack]]);
   }
   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons != 0)
   {
      maxnewratio = thelowcon[1]/newtime[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxnewratio = max2(maxnewratio, thelowcon[ktrack]/
            newtime[nodelowcon[ktrack]]);
      }  
      if(maxnewratio == 0.0)
      {  
         lmax = BIGNEGLOG;
      }  
      else
      {   
         lmax = log(maxnewratio);
      }  
      if(othk > -2.0*lmax)
      {  
         lownum = 0.5 + lmax*invothk;
      }  
      else
      {   
         lownum = 0.0;
      }  
   }   
   else
   {   
      lownum = 0.0;
   }
   for(ktrack=0;ktrack < nbranches; ktrack++)
   {
      parentnode = tonode[ktrack][MAXOFF];
      for(ktrack2 = 0; ktrack2 <= ngenes; ktrack2++)
      {
         holdthis = 0.5*(rate[ktrack2][ktrack]+rate[ktrack2][parentnode]);
         brlen[ktrack2][ktrack] = holdthis*( time[parentnode] - time[ktrack]);
         propbrlen[ktrack2][ktrack] = 
            holdthis* (newtime[parentnode] - newtime[ktrack]);
      }
   }
   jterm /= (highnum - lownum);
   ratio = exp(-1.0*(nnodes-ntips)*loginvfac)
      * jterm * multiratio(newtime,newrate,newlogr,time,rate, logr);
   summvn = 0.0;
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      newmvn[ktrack] = logmvn(propbrlen[ktrack],ktrack);
      summvn += newmvn[ktrack] - curmvn[ktrack];
   }
   ratio *= exp( summvn);

   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {  
         time[ktrack] = newtime[ktrack];
      }  
      suctimeflate++;
      for(ktrack=0; ktrack <= ngenes; ktrack++)
      {
         curmvn[ktrack] = newmvn[ktrack];
      }
   }
   else
   {   
      failtimeflate++;
   }
}

void bothflate(time,rate,logr)
double *time, **rate, **logr;
{
   double multiratio();
   double rnd();
   double logmvn();

   double ratio, randnum, invfac, flatefac ;
   double jterm, maxcurratio, maxnewratio;
   double lmin, lmax,lbound, ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   static double *newmvn;
   static double summvn;
   static double halfk, neghalfk, *thelowcon, invothk;
   static double *thehighcon;
   static double *newtime, **newrate, **newlogr;
   static double **brlen, **propbrlen;
   int ktrack, ktrack2, breakthis;
   int thisnode, parentnode;
   static int hereyet=0, numlowcons, *nodelowcon, numhighcons, *nodehighcon;

   if(hereyet ==0)
   { 
      newmvn = dvec(ngenes);
      brlen = dmat(ngenes,nnodes);
      propbrlen = dmat(ngenes,nnodes);
      thelowcon = dvec(nnodes);
      thehighcon = dvec(nnodes);
      nodelowcon = ivec(nnodes);
      nodehighcon = ivec(nnodes);
      hereyet = 1;
      numhighcons = 0;
      numlowcons = 0;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      for(ktrack = ntips; ktrack < nnodes; ktrack++)
      {
         breakthis = 0;
         for(ktrack2 = 1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thisnode = tonode[ktrack][ktrack2];
            if(lowtime[ktrack] == lowtime[thisnode] &&
               (thisnode >= ntips ||
               lowtime[thisnode] == 0.0) )
            {
               breakthis = 1;
               break;
            }
         }
         if( breakthis == 0)
         {
            numlowcons++;
            thelowcon[numlowcons] = lowtime[ktrack];
            nodelowcon[numlowcons] = ktrack;
         }
         if(ktrack == nnodes-1)
         {
            numhighcons++;
            thehighcon[numhighcons] = hightime[ktrack];
            nodehighcon[numhighcons] = ktrack;
         }
         else
         {
            if(hightime[ktrack] != hightime[tonode[ktrack][MAXOFF]])
            {
               numhighcons++;
               thehighcon[numhighcons] = hightime[ktrack];
               nodehighcon[numhighcons] = ktrack;
            }
         }
      }   
      newtime = dvec(nnodes);
      newrate = dmat(ngenes,nnodes);
      newlogr = dmat(ngenes,nnodes);
   }
   mincurratio = thehighcon[1]/time[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      mincurratio = min2( mincurratio,
         thehighcon[ktrack]/time[nodehighcon[ktrack]]);
   }

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons == 0)
   {
      lownum = 0.0;
      lmax = BIGNEGLOG;
   }
   else
   {   
      maxcurratio = thelowcon[1]/time[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxcurratio = max2(maxcurratio,
            thelowcon[ktrack]/time[nodelowcon[ktrack]]);
      }  
      if(maxcurratio == 0.0)
      {  
         lmax = BIGNEGLOG;
      }  
      else
      {   
         lmax = log(maxcurratio);
      }  
      if(othk < -2.0*lmax)
      {  
         lownum = 0.0;
      }  
      else
      {   
         lownum = 0.5+ lmax*invothk;
      }  
   }   
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(ktrack >= ntips)
      {
         newtime[ktrack] = flatefac*time[ktrack];
      }
      else
      {
         newtime[ktrack] = time[ktrack];
      }
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         newrate[ktrack2][ktrack] = invfac*rate[ktrack2][ktrack];
         newlogr[ktrack2][ktrack] = logr[ktrack2][ktrack] +loginvfac;
      }
   }   

   for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
   {
      newrate[ktrack2][nnodes] = rate[ktrack2][nnodes];
   }
   minnewratio = thehighcon[1]/newtime[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      minnewratio = min2(minnewratio, thehighcon[ktrack]/
         newtime[nodehighcon[ktrack]]);
   }
   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons != 0)
   {
      maxnewratio = thelowcon[1]/newtime[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxnewratio = max2(maxnewratio, thelowcon[ktrack]/
            newtime[nodelowcon[ktrack]]);
      }  
      if(maxnewratio == 0.0)
      {  
         lmax = BIGNEGLOG;
      }  
      else
      {   
         lmax = log(maxnewratio);
      }  
      if(othk > -2.0*lmax)
      {  
         lownum = 0.5 + lmax*invothk;
      }  
      else
      {   
         lownum = 0.0;
      }  
   }   
   else
   {   
      lownum = 0.0;
   }
   for(ktrack=0;ktrack < nbranches; ktrack++)
   {
      parentnode = tonode[ktrack][MAXOFF];
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         brlen[ktrack2][ktrack] = 0.5*(rate[ktrack2][ktrack]+
            rate[ktrack2][parentnode])*(time[parentnode] - time[ktrack]);
         propbrlen[ktrack2][ktrack] = 0.5*(newrate[ktrack2][ktrack]+
            newrate[ktrack2][parentnode])*
            (newtime[parentnode] - newtime[ktrack]);
      }
   }
   jterm /= (highnum - lownum);
   if(brownmean > SMALLNUM)
   {
      ratio = exp(((ngenes+1.0)*(nnodes-1.0)-(nnodes-ntips))*loginvfac) * jterm
         * multiratio(newtime,newrate,newlogr,time,rate, logr);
   }
   else
   {
      ratio = exp(((ngenes+1.0)-(nnodes-ntips))*loginvfac) * jterm
         * multiratio(newtime,newrate,newlogr,time,rate, logr);
   }   
   summvn = 0.0;
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      newmvn[ktrack] = logmvn(propbrlen[ktrack],ktrack);
      summvn += newmvn[ktrack] - curmvn[ktrack];
   }
   ratio *= exp( summvn);
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {  
         time[ktrack] = newtime[ktrack];
         for(ktrack2 = 0; ktrack2 <= ngenes; ktrack2++)
         {
            rate[ktrack2][ktrack] = newrate[ktrack2][ktrack];
            logr[ktrack2][ktrack] = newlogr[ktrack2][ktrack];
         }
      }  
      sucboth++;
      for(ktrack = 0; ktrack <= ngenes; ktrack++)
      {
         curmvn[ktrack] = newmvn[ktrack];
      }
   }
   else
   {   
      failboth++;
   }
}

void rateflate(time,rate,logr, whichgene)
int whichgene;
double *time, **rate, **logr;
{
   double probratio();
   double rnd();
   double logmvn();

   double ratio, randnum, invfac, flatefac ;
   double *trate, *tlogr,loginvfac;
   static double halfk, invothk;
   static double *newtime, *newrate, *newlogr;
   static double *brlen, *propbrlen;
   static double *newmvn;
   static double summvn;
   int ktrack;
   int thisnode, parentnode;
   static int hereyet=0;

   if(hereyet ==0)
   { 
      newmvn = dvec(ngenes);
      brlen = dvec(nnodes);
      propbrlen = dvec(nnodes);
      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);
   }
   loginvfac = halfk - othk*rnd(seed);
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;
   trate = rate[whichgene];
   tlogr = logr[whichgene];
   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      newtime[ktrack] = time[ktrack];
      newrate[ktrack] = flatefac*trate[ktrack];
      newlogr[ktrack] = tlogr[ktrack]-loginvfac;
   }   

   newrate[nnodes] = trate[nnodes];
   for(ktrack=0;ktrack < nbranches; ktrack++)
   {
      parentnode = tonode[ktrack][MAXOFF];
      brlen[ktrack] = 0.5*(trate[ktrack]+trate[parentnode])*(
         time[parentnode] - time[ktrack]);
      propbrlen[ktrack] = 0.5*(newrate[ktrack]+newrate[parentnode])*
         (newtime[parentnode] - newtime[ktrack]);
   }
   newmvn[whichgene] = logmvn(propbrlen,whichgene);
   if(brownmean > SMALLNUM)
   {
      ratio = exp(-1.0*(nnodes-1)*loginvfac + newmvn[whichgene] - 
         curmvn[whichgene] ) 
         * probratio(newtime,newrate,newlogr,time,trate, tlogr);
   }
   else
   {
      ratio = exp(-1.0*loginvfac + newmvn[whichgene]
         -curmvn[whichgene])
         * probratio(newtime,newrate,newlogr,time,trate, tlogr);
   }   
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {  
         trate[ktrack] = newrate[ktrack];
         tlogr[ktrack] = newlogr[ktrack];
      }  
      sucrateflate[whichgene]++;
      curmvn[whichgene] = newmvn[whichgene];
   }
   else
   {   
      failrateflate[whichgene]++;
   }
}



void consflate(time,rate,logr)
double *time, **rate, **logr;
{
   double multiratio();
   double rnd();

   double ratio, runlog, randnum, invfac, flatefac ;
   double jterm, maxcurratio, maxnewratio;
   double lmin, lmax,lbound, ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   static double halfk, neghalfk, *thelowcon, invothk;
   static double *thehighcon;
   static double *newtime, **newrate, **newlogr;
   int ktrack, ktrack2, breakthis;
   int thisnode, parentnode;
   static int hereyet=0, numlowcons, *nodelowcon, numhighcons, *nodehighcon;

   if(hereyet ==0)
   {
      thelowcon = dvec(nnodes);
      thehighcon = dvec(nnodes);
      nodelowcon = ivec(nnodes);
      nodehighcon = ivec(nnodes);
      hereyet = 1;
      numhighcons = 0;
      numlowcons = 0;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      for(ktrack = ntips; ktrack < nnodes; ktrack++)
      {
         breakthis = 0;
         for(ktrack2 = 1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thisnode = tonode[ktrack][ktrack2];
            if(lowtime[ktrack] == lowtime[thisnode] &&
               (thisnode >= ntips || 
               lowtime[thisnode] == 0.0) )
            {
               breakthis = 1;
               break;
            }
         }
         if( breakthis == 0) 
         {
            numlowcons++;
            thelowcon[numlowcons] = lowtime[ktrack];
            nodelowcon[numlowcons] = ktrack;
         }
         if(ktrack == nnodes-1)
         {
            numhighcons++;
            thehighcon[numhighcons] = hightime[ktrack];
            nodehighcon[numhighcons] = ktrack;
         }
         else
         {
            if(hightime[ktrack] != hightime[tonode[ktrack][MAXOFF]])
            {
               numhighcons++; 
               thehighcon[numhighcons] = hightime[ktrack];   
               nodehighcon[numhighcons] = ktrack; 
            }
         }
      }
      newtime = dvec(nnodes);
      newrate = dmat(ngenes,nnodes);
      newlogr = dmat(ngenes,nnodes);
   }
   runlog = 0.0;
   mincurratio = thehighcon[1]/time[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      mincurratio = min2( mincurratio,
         thehighcon[ktrack]/time[nodehighcon[ktrack]]);
   }
   for(ktrack =0 ; ktrack < ntips; ktrack++)
   {
      if(time[ktrack] > 0.0)
      {
         parentnode = tonode[ktrack][MAXOFF];
         for(ktrack2 = 0; ktrack2 <= ngenes; ktrack2++)
         {
            if( time[parentnode] * rate[ktrack2][ktrack] < time[ktrack]*(
               rate[ktrack2][ktrack]+rate[ktrack2][parentnode]) )
            {
               mincurratio = min2(mincurratio, (time[ktrack]*
                  rate[ktrack2][parentnode])/ (time[ktrack]*(
                  rate[ktrack2][ktrack]+rate[ktrack2][parentnode]) -
                  time[parentnode]*rate[ktrack2][ktrack]) );
            }
         }
      }
   }   

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   } 
   else
   {
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons == 0)
   {
      lownum = 0.0;
      lmax = BIGNEGLOG;
   }
   else
   {
      maxcurratio = thelowcon[1]/time[nodelowcon[1]];   
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {
         maxcurratio = max2(maxcurratio, 
            thelowcon[ktrack]/time[nodelowcon[ktrack]]);
      }
      if(maxcurratio == 0.0)
      {
         lmax = BIGNEGLOG;
      }
      else
      {
         lmax = log(maxcurratio);
      }
      if(othk < -2.0*lmax)
      {
         lownum = 0.0;
      }
      else
      {
         lownum = 0.5+ lmax*invothk;
      }
   }   
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(ktrack < ntips && time[ktrack] > 0.0)
      {
         parentnode = tonode[ktrack][MAXOFF];
         newtime[ktrack] = time[ktrack];
         for(ktrack2 = 0; ktrack2 <= ngenes; ktrack2++)
         {
            newrate[ktrack2][ktrack] = (time[parentnode] - time[ktrack])*
               (rate[ktrack2][parentnode] + rate[ktrack2][ktrack])/(
               newtime[parentnode]-time[ktrack]) -newrate[ktrack2][parentnode]; 
            newlogr[ktrack2][ktrack] = log(newrate[ktrack2][ktrack]);
         }
      } 
      else
      {  
         newtime[ktrack] = flatefac*time[ktrack];
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            newrate[ktrack2][ktrack] = invfac*rate[ktrack2][ktrack];
            newlogr[ktrack2][ktrack] = loginvfac + logr[ktrack2][ktrack];
         }
      }
   }

   for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
   {
      newrate[ktrack2][nnodes] = rate[ktrack2][nnodes];
   }
   minnewratio = thehighcon[1]/newtime[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {  
      minnewratio = min2(minnewratio, thehighcon[ktrack]/
         newtime[nodehighcon[ktrack]]);
   }  
   for(ktrack =0 ; ktrack < ntips; ktrack++)
   {
      parentnode = tonode[ktrack][MAXOFF];
      if(time[ktrack] > 0.0)
      {
         runlog += (ngenes+1.0)*log((time[parentnode] - time[ktrack])/ 
            (newtime[parentnode] - newtime[ktrack]));
      }
      else
      {
         runlog += (ngenes+1.0)*loginvfac;
      }
      if(newtime[ktrack] > 0.0)
      {
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            if( newtime[parentnode]*newrate[ktrack2][ktrack] < newtime[ktrack]*(
               newrate[ktrack2][ktrack]+newrate[ktrack2][parentnode]) )
            {
               minnewratio =min2(minnewratio, (newtime[ktrack]*newrate[ktrack2][
                  parentnode])/ (newtime[ktrack]*(newrate[ktrack2][ktrack]+ 
                  newrate[ktrack2][parentnode]) - newtime[parentnode]*
                  newrate[ktrack2][ktrack]) );
            }
         }
      }
   }   
   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {  
      highnum = 1.0;
   }  
   else  
   {  
      highnum = 0.5+ lmin*invothk;
   }  
   if(numlowcons != 0)
   {
      maxnewratio = thelowcon[1]/newtime[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxnewratio = max2(maxnewratio, thelowcon[ktrack]/
            newtime[nodelowcon[ktrack]]);
      }  
      if(maxnewratio == 0.0)
      {
         lmax = BIGNEGLOG;
      }
      else
      {
         lmax = log(maxnewratio);
      }
      if(othk > -2.0*lmax)
      {
         lownum = 0.5 + lmax*invothk;
      }
      else
      {
         lownum = 0.0;
      }
   }
   else
   {
      lownum = 0.0;
   }
   jterm /= (highnum - lownum);
   ratio = exp(runlog+((nintnodes)*(ngenes+1.0)-(nintnodes+1))*loginvfac) * 
      jterm * multiratio(newtime,newrate,newlogr,time,rate, logr);
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {
         time[ktrack] = newtime[ktrack];
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            rate[ktrack2][ktrack] = newrate[ktrack2][ktrack];
            logr[ktrack2][ktrack] = newlogr[ktrack2][ktrack];
         }
      }
      succons++;
   }   
   else
   {
      failcons++;
   }
}

void changebrown(time,rate, logr)
double *time, **rate, **logr;
{
   double probratio();
   double multiratio();
   double rnd();
   
   double *trate, *tlogr,ratio, randnum;
   static double  **newrate;
   int ktrack, ktrack2;
   static int hereyet=0;

   if(hereyet ==0)
   {
      hereyet = 1;
      newrate = dmat(ngenes,nnodes);
   }
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
      {
         newrate[ktrack][ktrack2] = rate[ktrack][ktrack2];
      }
   }

   if(commonbrown ==1)
   {
/* all genes have same tendency to change rate */
      newrate[0][nnodes] = rate[0][nnodes]*exp(thek *(rnd(seed)-0.5));
      for(ktrack2=1;ktrack2 <= ngenes; ktrack2++)
      {
         newrate[ktrack2][nnodes] = newrate[0][nnodes];
      }
      ratio = (newrate[0][nnodes]/rate[0][nnodes])*
      multiratio(time,newrate,logr,time,rate,logr);
      randnum = rnd(seed);
      if(randnum < ratio)
      {
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            rate[ktrack2][nnodes] = newrate[ktrack2][nnodes];
            sucrate[ktrack2][nnodes]++;
         }
      }
      else
      {
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            failrate[ktrack2][nnodes]++;
         }
      }
   }
   else
   {
/* commonbrown =0, so all genes have individual predilections
   to change rates */
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         newrate[ktrack2][nnodes] = rate[ktrack2][nnodes]*
            exp(thek *(rnd(seed)-0.5));
         ratio = (newrate[ktrack2][nnodes]/rate[ktrack2][nnodes])*
            probratio(time,newrate[ktrack2],logr[ktrack2],time,rate[ktrack2],
            logr[ktrack2]);
         randnum = rnd(seed);
         if(randnum < ratio)
         {
            rate[ktrack2][nnodes] = newrate[ktrack2][nnodes];
            sucrate[ktrack2][nnodes]++;
         }   
         else
         {
            failrate[ktrack2][nnodes]++;  
         }
      }
   }

}


void timechange(time,rate,logr)
double *time,**rate,**logr;
{
   void freedvec();
   double *dvec();
   double **dmat();
   double logcondmvn();
   double logmvn();
   double rnd();

   double blen, randnum, ratio; 
   double proptime;
   double minposs,maxposs, maxtime, maxchild, numratio, denomratio, oldfrac;
   double newfrac, junk, oldwin,newwin;
   double denregfac, numregfac,rtratio, expratio, regratio;
   double lhold,lbound, ubound,newrdif, nowrdif, newvar, nowvar, newlen, nowlen;
   static double *newmvn;
   static double halfk, neghalfk, **brlen, **propbrlen, invthek;
   int thisbranch, thisparent, thischild, ktrack, ktrack2, ktrack3;
   static int hereyet = 0;


   if(hereyet ==0)
   {
      newmvn = dvec(ngenes);
      brlen = dmat(ngenes,nbranches);
      propbrlen = dmat(ngenes,nbranches);
      invthek = 1.0/thek;
      halfk = 0.5*thek;
      neghalfk = -1.0*halfk;
      hereyet=1;
   }
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      blen = time[thisparent] - time[ktrack];
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         brlen[ktrack2][ktrack] = 0.5*(rate[ktrack2][thisparent]+
            rate[ktrack2][ktrack])*blen;
         propbrlen[ktrack2][ktrack] = brlen[ktrack2][ktrack];
      }
   }
   for(ktrack = ntips; ktrack < nnodes; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      maxchild = time[tonode[ktrack][1]];
      for(ktrack2=2; ktrack2 <= tonode[ktrack][0]; ktrack2++)
      {
         maxchild = max2(maxchild, time[tonode[ktrack][ktrack2]]);
      }
      minposs = max2(maxchild, lowtime[ktrack]);
      if(ktrack < nnodes -1)
      {
         maxposs = min2(hightime[ktrack], time[thisparent]);
         maxtime = maxposs - minposs;
         oldfrac = (time[ktrack]-minposs)/maxtime;
         oldwin = min2(oldfrac,1.0-oldfrac);
         junk = rnd(seed);
         if(junk < newk)
         {
            newfrac = rnd(seed);
         }
         else
         {
            newfrac = oldfrac + (2.0*rnd(seed) - 1.0)*oldwin;
         }
         newwin = min2(newfrac, 1.0-newfrac);
         proptime = minposs + maxtime*newfrac;
         if((oldfrac > (newfrac-newwin)) && (oldfrac < (newfrac+newwin)))
         {
            numratio = newk + (1.0-newk )*0.5/newwin;
         }
         else
         {
            numratio = newk;
         }
         if((newfrac > (oldfrac-oldwin)) && (newfrac < (oldfrac+oldwin)))
         {
            denomratio = newk+ (1.0-newk)*0.5/oldwin;
         }
         else
         {
            denomratio = newk;
         }
         for(ktrack2 = 0; ktrack2<= ngenes; ktrack2++)
         {
            propbrlen[ktrack2][ktrack] = (time[thisparent] - proptime)*0.5*(
               rate[ktrack2][thisparent]+rate[ktrack2][ktrack]);
         }
         rtratio = 1.0;
         if(brownmean > SMALLNUM)
         {
            expratio = 0.0;
            regratio = 1.0;
            newlen = time[thisparent] - proptime;
            nowlen = time[thisparent] - time[ktrack];
            nowvar = rate[0][nnodes];
            for(ktrack2=1; ktrack2 <= ngenes; ktrack2++)
            {
               nowvar = min2(nowvar, rate[ktrack2][nnodes]);
            }
            nowvar *= nowlen;
          
            if(nowvar > 0.0)
            {
               regratio = numratio/ denomratio;
               expratio = a[ktrack]*log(proptime/time[ktrack])+ b[ktrack]*
                  log((time[thisparent] - proptime)/(time[thisparent] - 
                  time[ktrack])) ;
               for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
               {
                  newvar = newlen*rate[ktrack2][nnodes];
                  nowvar = nowlen*rate[ktrack2][nnodes];
                  nowrdif = logr[ktrack2][ktrack] - logr[ktrack2][thisparent] + 
                     0.5*nowvar;
                  newrdif = logr[ktrack2][ktrack] - logr[ktrack2][thisparent] + 
                     0.5*newvar;
                  if( ktrack != consnode)
                  {
                     expratio += 0.5*(nowrdif*nowrdif/nowvar - newrdif*newrdif/
                        newvar);
                     rtratio *= nowvar/newvar;
                  }
               }
            }
            else
            {
               rtratio = 0.0;
               expratio = 0.0;
               regratio = 0.0;
            }
         }
         else
         {
            regratio = numratio/ denomratio;
            expratio = a[ktrack]*log(proptime/time[ktrack])+ b[ktrack]*
               log((time[thisparent] - proptime)/(time[thisparent] -
               time[ktrack])) ;
         }
      }
      else 
      {
/* ktrack = nnodes -1 ... try to change root depth */  
         maxposs = hightime[nnodes-1];
         lbound = neghalfk;
         lhold = log((maxposs-minposs)/(time[ktrack] - minposs)); 
         ubound = min2(halfk,lhold);
         proptime = minposs+(time[ktrack]-minposs)*exp(lbound+
            (ubound-lbound)*rnd(seed) );
         denregfac = min2(1.0, invthek*lhold+0.5);
         numregfac = min2(1.0, invthek*log((maxposs-minposs)/(proptime -
            minposs)) + 0.5); 

         regratio = ((proptime - minposs)*numregfac)/(denregfac*(
            time[ktrack] - minposs));
         expratio = (timegamr - 1.0)*log(proptime/time[nnodes-1])
         + timegamlam*(time[nnodes-1] - proptime);
         rtratio = 1.0;
      }
      for(ktrack2=1; ktrack2 <=tonode[ktrack][0]; ktrack2++)
      {
         thischild = tonode[ktrack][ktrack2];
         newlen = proptime - time[thischild];
         nowlen = time[ktrack] - time[thischild];
         for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
         {
            propbrlen[ktrack3][thischild] = newlen*0.5*(
               rate[ktrack3][thischild] + rate[ktrack3][ktrack]);
         }
         if(thischild >= ntips)
         {
            expratio += a[thischild]*log(time[ktrack]/proptime) +
               b[thischild]*log(newlen*time[ktrack]/(nowlen*proptime));
            regratio *= time[ktrack]/proptime;
         }
         if(brownmean > SMALLNUM)
         {
            for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
            {
               newvar = newlen*rate[ktrack3][nnodes];
               if(newvar > 0.0)
               {
                  if(thischild != consnode)
                  {
                     nowvar = nowlen*rate[ktrack3][nnodes];
                     newrdif =logr[ktrack3][thischild]- logr[ktrack3][ktrack] + 
                        0.5*newvar;
                     nowrdif =logr[ktrack3][thischild]-logr[ktrack3][ktrack] + 
                        0.5*nowvar;
                     expratio += 0.5*((nowrdif*nowrdif)/nowvar -(newrdif*
                        newrdif)/ newvar);
                     rtratio *= nowvar/newvar;
                  }
               }
               else
               {
                  rtratio = 0.0;
                  expratio = 0.0;
                  regratio = 0.0;
               }  
            }
         }
      }
/* DEBUG
      if(fulldata==1)
      {
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            expratio += logcondmvn(propbrlen[ktrack2],ktrack,ktrack2)-
               logcondmvn(brlen[ktrack2],ktrack, ktrack2);
         }
      }
      else
      {
*/
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            newmvn[ktrack2] = logmvn(propbrlen[ktrack2],ktrack2);
            expratio += newmvn[ktrack2] - curmvn[ktrack2];
         }

/* DEBUG
      } 
*/
      ratio = exp(expratio)*sqrt(rtratio)* regratio;
      randnum = rnd(seed);
      if(randnum < ratio)
      { 
         suctime[ktrack]++;
         for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
         {
            curmvn[ktrack2] = newmvn[ktrack2];
         }

         time[ktrack] = proptime;
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thisbranch = tonode[ktrack][ktrack2];
            for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
            {
               brlen[ktrack3][thisbranch] =propbrlen[ktrack3][thisbranch];
            }
         }
         if(ktrack < nnodes-1)
         {
            for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
            {
               brlen[ktrack3][ktrack] = propbrlen[ktrack3][ktrack];
            }
         }
      }
      else
      {
         failtime[ktrack]++;
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++) 
         { 
            thisbranch = tonode[ktrack][ktrack2];
            for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
            {
               propbrlen[ktrack3][thisbranch] = brlen[ktrack3][thisbranch];        
            }
         } 
         if(ktrack < nnodes-1) 
         { 
            for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
            {
               propbrlen[ktrack3][ktrack] = brlen[ktrack3][ktrack];
            }
         } 
      }
   }
}
 
void ratechange(time,rate,logr, whichgene)
double *time,**rate, **logr;
int whichgene;
{
   double *trate, *tlogr, thelen, var, holdlen;
   double rfac, logrfac, proprate, proplogr, randnum, ratio;
   static double *brlen, *propbrlen;
   double expratio, regratio, newrdif, nowrdif;
   int thischild, thisparent, ktrack, ktrack2;
   static double *newmvn;
   static double summvn;
   static int hereyet = 0;

   void freedvec();
   double *dvec();
   double logcondmvn();
   double logmvn();
   double rnd();

   if(hereyet == 0)
   {
      newmvn = dvec(ngenes);
      brlen = dvec(nbranches);
      propbrlen = dvec(nbranches);
      hereyet = 1;
   }
   trate = rate[whichgene];
   tlogr = logr[whichgene];
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      holdlen =  time[thisparent] - time[ktrack];
      brlen[ktrack] = holdlen*0.5*(trate[ktrack]+trate[thisparent]);
      propbrlen[ktrack] = brlen[ktrack];
   }
   for(ktrack = 0; ktrack < nnodes-1; ktrack++)
   {
      logrfac = thek*(rnd(seed)-0.5);
      rfac = exp(logrfac);
      proprate = rfac*trate[ktrack];
      proplogr = tlogr[ktrack]+logrfac;
      expratio = 0.0;
      regratio = 1.0;
      thisparent = tonode[ktrack][MAXOFF];
      thelen = time[thisparent] - time[ktrack];
      if(ktrack != consnode)
      {
         var = thelen*trate[nnodes];
         propbrlen[ktrack] = thelen*0.5*(trate[thisparent]+proprate);
         if(var > 0.0)
         {
            newrdif = proplogr - tlogr[thisparent] + 0.5*var;
            nowrdif = tlogr[ktrack] - tlogr[thisparent] + 0.5*var;
            regratio *= trate[ktrack]/ proprate;
            expratio += 0.5*(nowrdif*nowrdif - newrdif*newrdif)/var;
         }
         else
         {
            expratio = 0.0;
            regratio = 0.0;
         }
      }          
      else
      {
         propbrlen[ktrack] = thelen*proprate;
         expratio += (rategamr - 1.0)*logrfac+
            rategamlam*(trate[ktrack] - proprate);   
         for(ktrack2 = 1; ktrack2 <= tonode[nnodes-1][0]; ktrack2++)
         {
            thischild = tonode[nnodes-1][ktrack2];
            if(thischild != consnode)
            {
               thelen = time[nnodes-1] - time[thischild];
               var = thelen*trate[nnodes];
               if( var > 0.0)
               {
                  newrdif = tlogr[thischild] - proplogr + 0.5*var;
                  nowrdif = tlogr[thischild] - tlogr[ktrack] + 0.5*var;
                  propbrlen[thischild ] =0.5*(proprate+trate[thischild])*thelen;
                  expratio += 0.5*(nowrdif*nowrdif - newrdif*newrdif)/var;
               }
               else
               {
                  expratio = 0.0;
                  regratio = 0.0;
               }              
            }
         }
      }
      if(ktrack >= ntips)
      {
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thischild = tonode[ktrack][ktrack2];
            thelen = time[ktrack] - time[thischild];
            var = thelen*trate[nnodes];
            if( var > 0.0)
            {
               newrdif = tlogr[thischild] - proplogr + 0.5*var;
               nowrdif = tlogr[thischild] - tlogr[ktrack] + 0.5*var;
               propbrlen[thischild ] = 0.5*(proprate+ trate[thischild])*thelen;
               expratio += 0.5*(nowrdif*nowrdif - newrdif*newrdif)/var;
            }
            else
            {
               expratio = 0.0;
               regratio = 0.0;
            }
         }
      }
/* DEBUG
      if(fulldata ==1)
      {
         expratio += logcondmvn(propbrlen, ktrack, whichgene) - 
            logcondmvn(brlen, ktrack, whichgene);
      }
      else
      {
*/
         newmvn[whichgene] = logmvn(propbrlen,whichgene);
         expratio += newmvn[whichgene] - curmvn[whichgene];
/* DEBUG
      }
*/
      ratio = regratio*exp(expratio)* rfac;
      randnum = rnd(seed);

      if(randnum < ratio)
      { 
         curmvn[whichgene] = newmvn[whichgene];
         sucrate[whichgene][ktrack]++;
         trate[ktrack] = proprate;
         tlogr[ktrack] = proplogr;
         brlen[ktrack] = propbrlen[ktrack];
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thischild = tonode[ktrack][ktrack2];
            brlen[thischild] =propbrlen[thischild];
         }         
         if(ktrack == consnode)
         {
            trate[nnodes-1] = proprate;
            tlogr[nnodes-1] = proplogr;
            for(ktrack2 = 1; ktrack2 <= tonode[nnodes-1][0]; ktrack2++)
            {
               thischild = tonode[nnodes-1][ktrack2];
               brlen[thischild] = propbrlen[thischild];
            }
         }
      }
      else
      {
         failrate[whichgene][ktrack]++;
         propbrlen[ktrack] = brlen[ktrack];
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++)  
         {         
            thischild = tonode[ktrack][ktrack2];
            propbrlen[thischild] =brlen[thischild];
         }        
         if(ktrack == consnode)
         {
            for(ktrack2=1; ktrack2 <= tonode[nnodes-1][0]; ktrack2++)
            {
               thischild = tonode[nnodes-1][ktrack2];
               propbrlen[thischild] = brlen[thischild];
            }
         }
      }
   }
}
 
void testcors(postime,posrate)
double *postime, **posrate;
{
   
   double **logposrate, **logdevrate, **fklogposrate;
   double ***thecor, **truecor, magicnum, ransign;
   double **rateranks;
   int count, thisrep;
   int  ktrack, ktrack2;

   void dotherank();
   double dothecor();
   double rnd();

   truecor = dmat(ngenes,ngenes);
   thecor = threedmat(NUMCORREPS,ngenes,ngenes);
   rateranks = dmat(ngenes,nnodes);
   logposrate = dmat(ngenes,nnodes);
   logdevrate = dmat(ngenes,nnodes);
   fklogposrate = dmat(ngenes,nnodes);

   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      for(ktrack2=0; ktrack2 < nnodes; ktrack2++)
      {
         logposrate[ktrack][ktrack2] = log(posrate[ktrack][ktrack2]);
      }
      for(ktrack2 =0 ; ktrack2 < nnodes-1; ktrack2++)
      {
/* adjusting logdevrate below for the fact that the mean of the logarithm
of the rate is not the same as the logarithm of the mean */
         logdevrate[ktrack][ktrack2] = logposrate[ktrack][ktrack2] -
            logposrate[ktrack][tonode[ktrack2][MAXOFF]] + 
            0.5* posrate[ktrack][nnodes]*(postime[ tonode[ktrack2][MAXOFF]]-
            postime[ktrack2]);

      }
   }
for(thisrep=1; thisrep<= NUMCORREPS; thisrep++)
{
   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {  
      fklogposrate[ktrack][nnodes-1] = logposrate[ktrack][nnodes-1];
      for(ktrack2=nnodes-2; ktrack2 >= 0; ktrack2--)
      {  
/* DEBUG BELOW */
         if(ktrack2 == consnode)
         {
            fklogposrate[ktrack][ktrack2] = fklogposrate[ktrack][nnodes-1];
         }
         else
         {
            magicnum = rnd(seed);
            if(magicnum < 0.5)
            {
               ransign = 1.0;
            }
            else
            {
               ransign = -1.0;
            }
/* DEBUG Modification below !! */
            fklogposrate[ktrack][ktrack2] = 
               fklogposrate[ktrack][tonode[ktrack2][ MAXOFF]] - 
               0.5* posrate[ktrack][nnodes]*(postime[ tonode[ktrack2][MAXOFF]]-
               postime[ktrack2])+ ransign*logdevrate[ktrack][ktrack2];
         }
         dotherank(fklogposrate[ktrack],nnodes-1,rateranks[ktrack]); 
      }
   }
   for(ktrack=0; ktrack < ngenes; ktrack++)
   {
      for(ktrack2=ktrack+1; ktrack2 <= ngenes; ktrack2++)
      {
         thecor[thisrep][ktrack][ktrack2] = dothecor(rateranks[ktrack],
            rateranks[ktrack2], nnodes-1);
      }
   }
}
   for(ktrack = 0; ktrack<= ngenes; ktrack++)
   {
      dotherank(logposrate[ktrack],nnodes-1,rateranks[ktrack]);
   }
   printf("\n");
/* 
for(ktrack = 0; ktrack <= ngenes; ktrack++)
{
   for(ktrack2=0; ktrack2 < nnodes-1; ktrack2++)
   {
printf("posrate[%3d][%3d]= %6.4lf rank= %5.2lf\n", ktrack, ktrack2,
posrate[ktrack][ktrack2], rateranks[ktrack][ktrack2]);
   }
}
*/
   for(ktrack=0; ktrack < ngenes; ktrack++)
   {
      for(ktrack2=ktrack+1; ktrack2 <= ngenes; ktrack2++)
      {
         truecor[ktrack][ktrack2] = dothecor(rateranks[ktrack],
            rateranks[ktrack2], nnodes-1);
         count = 0;
         for(thisrep=1; thisrep <= NUMCORREPS; thisrep++)
         {
            if(thecor[thisrep][ktrack][ktrack2] >= truecor[ktrack][ktrack2])
            {
               count++;
            }
         }
         printf("For rank corr. between genes %2d and %2d, the observed\n",
            ktrack, ktrack2);
         printf("value is %10.7lf and this is equalled or exceeded %4.3lf of\n",
            truecor[ktrack][ktrack2], (1.0*count)/(1.0*NUMCORREPS) );
         printf("the time when these 2 genes evolve independently\n\n");
      }
   }
    
   freethreedmat(thecor,NUMCORREPS,ngenes,ngenes);
   freedmat(rateranks,ngenes,nnodes);
   freedmat(logposrate,ngenes,nnodes);
   freedmat(logdevrate,ngenes,nnodes);
   freedmat(fklogposrate,ngenes,nnodes);
   freedmat(truecor,ngenes,ngenes);
}

double dothecor(xvec,yvec,veclen)
double *xvec, *yvec;
int veclen;
{
   int ktrack;
   double sumsqx, sumsqy, sumx, sumy, sumxy;

   sumx = 0.0;
   sumy = 0.0;
   sumsqx = 0.0;
   sumsqy = 0.0;
   sumxy = 0.0;

   for(ktrack=0; ktrack < veclen; ktrack++)
   {
      sumx += xvec[ktrack];
      sumy += yvec[ktrack];
      sumxy += xvec[ktrack]*yvec[ktrack];
      sumsqx += xvec[ktrack]*xvec[ktrack];
      sumsqy += yvec[ktrack]*yvec[ktrack];
   }
   return((sumxy-(sumx*sumy)/veclen)/(
      sqrt(sumsqx-sumx*sumx/veclen)*sqrt(sumsqy-sumy*sumy/veclen)));
}

void summarize(samptime, samprate,sufname)
double **samptime, ***samprate;
char *sufname;
{
   int jtrack,ktrack, ktrack2, ktrack3;
   int lowval, highval;
   int whichgene, whichnode;
   double thelowval, thehighval;
   double *sumsq,*sum, *sampmean, *sampmeansq, *sumtime, *sumsqtime;
   double holdrt, holdtemp;
   double holdparent, holdbrlen;
   double holdtime;
   double *oldtime,*newtime,**oldrate,**newrate,theratio;
   double *actualtime, **oldlogr, **newlogr;
   double **prtime, **hbrlen;
   double *dumvec;
   double **posrate,*postime;
   FILE *ratiofile, *treefile, *nodefile;
   char ratiofilename[100],nodefilename[100], treefilename[100];

   void fprinttree();
   double multiratio();
   void sort();
   void testcors();

   posrate = dmat(ngenes,nnodes);
   postime = dvec(nnodes);

   lowval = ((int) floor(numsamps*LOWFRAC +0.5)) -1;
   highval = ((int) floor(numsamps*HIGHFRAC + 0.5))-1;
   printf("POSTERIOR INTERVAL from probability %9.5lf to %9.5lf\n",
      LOWFRAC, HIGHFRAC);
   printf("(i.e. %5d -th lowest value to %5d -th lowest value)\n",
      lowval+1, highval+1);
   printf("Consflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      succons, failcons, (1.0*succons)/ (succons+ failcons) ); 
   printf("Bothflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucboth, failboth, (1.0*sucboth)/ (sucboth+ failboth) );

   for(ktrack=0; ktrack <= ngenes; ktrack++)
   {
      printf("Rateflate[%1d]: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n", ktrack,
         sucrateflate[ktrack],failrateflate[ktrack],(1.0*sucrateflate[ktrack])/ 
         (sucrateflate[ktrack]+ failrateflate[ktrack]) );

      printf("Partrateflate[%1d]: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",ktrack,
         sucpartrate[ktrack], failpartrate[ktrack], (1.0*sucpartrate[ktrack])/ 
         (sucpartrate[ktrack]+ failpartrate[ktrack]) );   
   }

   printf("Timeflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      suctimeflate, failtimeflate, (1.0*suctimeflate)/ (suctimeflate+ 
      failtimeflate) );
   printf("Partconsflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucpartcons, failpartcons, (1.0*sucpartcons)/ (sucpartcons+
      failpartcons) );   
   printf("Partbothflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucpartboth, failpartboth, (1.0*sucpartboth)/ (sucpartboth+
      failpartboth) );   
   printf("Parttimeflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucparttime, failparttime, (1.0*sucparttime)/ (sucparttime+
      failparttime) );   


   if(brownmean > SMALLNUM)
   {
      for(ktrack=0; ktrack <= nnodes; ktrack++)
      {
         if(ktrack != nnodes-1)
         {
            for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
            {
               if(ktrack == nnodes)
               {
                  printf("BROWNIAN");
               }
               else
               {
                  printf("RATE");
               }
               printf("[gene %1d][node %2d] ", ktrack2, ktrack);
               printf("nsuc=%10ld nfail=%10ld psuc= %7.5lf\n", 
                  sucrate[ktrack2][ktrack], failrate[ktrack2][ktrack], 
                  (1.0*sucrate[ktrack2][ktrack])/ (sucrate[ktrack2][ktrack]+
                  failrate[ktrack2][ktrack]) );
            }
         }
         else
         {
            printf("ROOT rate constrained to equal rate of node %2d\n",
               consnode);
         } 
      }
   }
   for(ktrack=ntips; ktrack <= nnodes-1; ktrack++) 
   { 
      printf("TIME %2d: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
         ktrack, suctime[ktrack], failtime[ktrack], (1.0*suctime[ktrack])/ 
         (suctime[ktrack]+failtime[ktrack]) ); 
   }

   oldlogr = dmat(ngenes,nnodes);
   newlogr = dmat(ngenes,nnodes);
   oldtime = dvec(nnodes);
   newtime = dvec(nnodes);
   oldrate = dmat(ngenes,nnodes);
   newrate = dmat(ngenes,nnodes);

   strcpy(nodefilename,"node.");
   strcat(nodefilename,sufname);
   strcpy(ratiofilename,"ratio.");
   strcat(ratiofilename,sufname);
   nodefile = gfopen(nodefilename, "w");
   ratiofile = gfopen(ratiofilename,"w");
   for(ktrack=0; ktrack<= nnodes; ktrack++)
   {
      if(ktrack < nnodes)
      {
         oldtime[ktrack] = samptime[ktrack][0];
      }
      for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
      {
         oldrate[ktrack2][ktrack] = samprate[ktrack2][ktrack][0];
         oldlogr[ktrack2][ktrack] = log(oldrate[ktrack2][ktrack]);
      }
   }
   for(ktrack2=0; ktrack2 < numsamps; ktrack2++)
   {
      for(ktrack=0; ktrack<= nnodes; ktrack++) 
      { 
         if(ktrack < nnodes)
         {
            newtime[ktrack] = samptime[ktrack][ktrack2]; 
         }
         for(ktrack3=0; ktrack3 <= ngenes; ktrack3++)
         { 
            newrate[ktrack3][ktrack] = samprate[ktrack3][ktrack][ktrack2]; 
            newlogr[ktrack3][ktrack] = log(newrate[ktrack3][ktrack]);
         }
      }
      theratio = multiratio(newtime,newrate, newlogr,oldtime,oldrate,oldlogr);
      fprintf(ratiofile, "%12.6lf\n", theratio);
   } 
   fclose(ratiofile);
   freedvec(oldtime);
   freedvec(newtime);
   freedmat(oldrate, ngenes, nnodes);
   freedmat(newrate,ngenes, nnodes);
   freedmat(oldlogr, ngenes, nnodes);
   freedmat(newlogr,ngenes, nnodes);
   sumtime = dvec(nnodes);
   sumsqtime = dvec(nnodes);
   sumsq = dvec(nnodes+(ngenes+1)*(nnodes+1));
   sum = dvec(nnodes+(ngenes+1)*(nnodes+1));
   sampmean = dvec(nnodes);
   sampmeansq = dvec(nnodes);
   prtime = dmat(nnodes,numsamps);
   hbrlen = dmat(nnodes,numsamps);
   dumvec = dvec(numsamps);
   for(ktrack=0; ktrack <= nnodes+ (ngenes+1)*(nnodes+1)-1; ktrack++)
   {
      sum[ktrack] = 0.0;
      sumsq[ktrack] = 0.0;
      if(ktrack < nnodes)
      {
         sumtime[ktrack] = 0.0;
         sumsqtime[ktrack] = 0.0;
      }
      else
      {
         whichgene = (ktrack-nnodes)/(nnodes+1);
         whichnode = (ktrack-nnodes)%(nnodes+1);
      }

      if(ktrack < nnodes-1)
      {
         sampmean[ktrack] = 0.0;
         sampmeansq[ktrack] = 0.0;
      }
      for(ktrack2=0; ktrack2 < numsamps; ktrack2++)
      {
         if(ktrack < nnodes)
         {
            holdtime = samptime[ktrack][ktrack2];
            sumtime[ktrack] += holdtime;
            sumsqtime[ktrack] += holdtime*holdtime;
         }
         if(ktrack < nnodes-1)
         {
/* ONLY BRANCH LENGTHS FOR 1st Gene (Gene 0) are directly reported now */
            holdparent = samptime[tonode[ktrack][MAXOFF]][ktrack2];
            holdrt = 0.5*(samprate[0][tonode[ktrack][MAXOFF]][ktrack2] +
               samprate[0][ktrack][ktrack2]);
            holdbrlen = holdrt*(holdparent-holdtime);
            hbrlen[ktrack][ktrack2] = holdbrlen;
            sampmean[ktrack] += holdbrlen;
            sampmeansq[ktrack] += holdbrlen*holdbrlen; 
            holdrt = samptime[nnodes-1][ktrack2];
            holdtemp =  holdtime/holdrt;
            prtime[ktrack][ktrack2] = holdtemp;
            sum[ktrack] += holdtemp;
            sumsq[ktrack] += holdtemp*holdtemp;
         }
         else
         {
            if(ktrack == nnodes -1)
            {
               holdrt = samptime[ktrack][ktrack2];
               sum[ktrack] += holdrt;
               sumsq[ktrack] += holdrt*holdrt; 
            }
            else
            {

               holdrt = samprate[whichgene][whichnode][ktrack2];
               sum[ktrack] += holdrt;
               sumsq[ktrack] += holdrt*holdrt; 
            } 
         }      
      }
      if(ktrack < nnodes-1)
      {
         sort(prtime[ktrack],numsamps,dumvec);
         thelowval = dumvec[lowval];
         thehighval = dumvec[highval];
         fprintf(nodefile,"p %2d", ktrack);
         printf("Prop. Node %2d to Root= ", ktrack);
      }
      else
      {
         if(ktrack==nnodes-1)
         {
/* DEBUG HAVE TO CREATE DUMMY VECTORS TO AVOID DESTRUCTION */
            sort(samptime[nnodes-1],numsamps,dumvec);
            thelowval = dumvec[lowval];
            thehighval = dumvec[highval];
            fprintf(nodefile, "p %2d", ktrack);
            printf("Root Depth = ");
         }
         else
         {
            sort(samprate[whichgene][whichnode],numsamps,dumvec);
            thelowval = dumvec[lowval];
            thehighval = dumvec[highval];
            fprintf(nodefile,"R %2d %2d ", whichgene, whichnode);
            posrate[whichgene][whichnode] = sum[ktrack]/numsamps;
            if(whichnode != nnodes )
            {
               printf("Rt[Gene %2d,Node %2d]= ", whichgene, whichnode);
            }
            else
            {
               printf("Brown[Gene %2d] = ",whichgene);
            }
         }
      }
      if(sumsq[ktrack] - sum[ktrack]*sum[ktrack]/numsamps< 0.0)
      {
         printf("prob in summary, can't take sqr. root of %18.12lf\n",
            (sumsq[ktrack] -(sum[ktrack]*sum[ktrack]/numsamps))/(numsamps-1.0));
      }
      fprintf(nodefile, " %9.5lf %9.5lf %9.5lf %9.5lf\n", sum[ktrack]/numsamps
         , sqrt((sumsq[ktrack] - sum[ktrack]*sum[ktrack]/numsamps
         )/(numsamps-1.0)), thelowval, thehighval);
      printf(" %9.5lf (Std. Dev. = %9.5lf) (%9.5lf,%9.5lf)\n", 
         sum[ktrack]/numsamps
         , sqrt((sumsq[ktrack] - sum[ktrack]*sum[ktrack]/numsamps
         )/(numsamps-1.0)), dumvec[lowval],dumvec[highval]);
   }
   printf("\n");

   for(ktrack=0 ; ktrack < tnnodes[0]-1; ktrack++)
   { 
      sort(hbrlen[ktrack],numsamps,dumvec);     
      thelowval = dumvec[lowval];
      thehighval = dumvec[highval];
printf("br %3d samp mean= %8.5lf (%8.5lf) est. = %8.5lf (%9.5lf, %9.5lf)\n", 
         ktrack, sampmean[ktrack]/numsamps, sqrt((sampmeansq[ktrack] -
         sampmean[ktrack]*sampmean[ktrack]/numsamps)/(numsamps-1.0)), 
         mean[0][ktrack], dumvec[lowval], dumvec[highval]);
   }

   for(ktrack = 0; ktrack < nnodes; ktrack++)
   {
      sort(samptime[ktrack],numsamps,dumvec);
      thelowval = dumvec[lowval];
      thehighval = dumvec[highval];
      fprintf(nodefile,"T %2d %9.5lf %9.5lf %9.5lf %9.5lf\n", ktrack, 
         sumtime[ktrack]/numsamps,  sqrt((sumsqtime[ktrack] - 
         sumtime[ktrack]* sumtime[ktrack]/numsamps)/(numsamps-1.0)),
         thelowval, thehighval);
      printf("Actual time node %2d = %9.5lf (Std. Dev. = ", 
         ktrack, sumtime[ktrack]/numsamps);
      postime[ktrack] = sumtime[ktrack]/numsamps;
      if(sumsqtime[ktrack] - sumtime[ktrack]*sumtime[ktrack]/numsamps< 
         NEGSMALLNUM)
      {
         printf(" UNDEFINED)\n");
         printf("Can't take square root of ... %25.20lf\n", 
            sumsqtime[ktrack] - sumtime[ktrack]*sumtime[ktrack]/numsamps);
      }
      else
      {
         if(sumsqtime[ktrack] - sumtime[ktrack]*sumtime[ktrack]/numsamps<=
            0.0)
         {
            printf(" %9.5lf) (%9.5lf, %9.5lf)\n", 0.0, dumvec[lowval], 
               dumvec[highval]);
         }
         else
         {
            printf(" %9.5lf) (%9.5lf, %9.5lf)\n", 
               sqrt((sumsqtime[ktrack] - sumtime[ktrack]* sumtime[ktrack]/
               numsamps)/(numsamps-1.0)), dumvec[lowval], dumvec[highval]);
         }
      }
   }
   actualtime = dvec(nnodes);
   for(ktrack = 0; ktrack < nnodes-1; ktrack++)
   {
      actualtime[ktrack] = (sumtime[tonode[ktrack][MAXOFF]] - sumtime[ktrack])
         /numsamps;
   }
   strcpy(treefilename,"tree.");
   strcat(treefilename,sufname);
   treefile = gfopen(treefilename, "w");
   fprinttree(treefile,nnodes-1,1,nnodes-1,actualtime);
   fclose(treefile);
   fclose(nodefile);

   if(brownmean > SMALLNUM && nodata ==0)
   {
      testcors(postime,posrate);
   }

   freedvec(postime);
   freedmat(posrate,ngenes,nnodes);
   freedvec(dumvec);
   freedmat(hbrlen,nnodes,numsamps);
   freedmat(prtime,nnodes,numsamps);
   freedvec(actualtime);
   freedvec(sumtime);
   freedvec(sumsqtime);
   freedvec(sampmean);
   freedvec(sampmeansq);
   freedvec(sumsq);
   freedvec(sum);

}


 
void
fprintname(outfile,whichseq)
FILE *outfile;
int whichseq;
   /* prints the name of sequence number whichseq */
{
   int ttrack;
   char c;
 
   ttrack = 0;
   c = ' ';
   while (c != '\n')
   {
      c= name[whichseq][ttrack];
      if( c != '\n') fprintf(outfile,"%1c",c);
      ttrack++;
   }
}

/* print the tree */
void fprinttree(outfile,whichnode,whichoption,highnodenum,param)
FILE *outfile;
int whichnode, whichoption,highnodenum;
double *param;
/* if whichoption != 0 then print sequence names else print sequence
numbers */
{
   int ktrack;

   if(tonode[whichnode][0] ==0)
   {
      if(whichoption != 0)
      {
         fprintname(outfile,whichnode);
      }
      else
      {   
         if(whichoption==0)
         {
            fprintf(outfile,"%d", whichnode);
         }
      }   
   }   
   else
   {   
      for(ktrack=1; ktrack <= tonode[whichnode][0]; ktrack++)
      {
         if(ktrack==1)
         {
            fprintf(outfile,"(");
         }

         else
         {   
            fprintf(outfile,",");
         }

         fprinttree(outfile,tonode[whichnode][ktrack],whichoption,
            highnodenum, param);
      }
      fprintf(outfile, ")");
   }
   if(whichnode==highnodenum)
   {
      fprintf(outfile, ";\n");
   }
   else
   {   
      if(whichoption !=2)
      {
         fprintf(outfile, ":%9.5f",param[whichnode]);
      }
      else
      {
         fprintf(outfile,":%d", ((int) param[whichnode]));
      }
   }
}



void
printname(whichseq)
int whichseq;
   /* prints the name of sequence number whichseq */
{
   int ttrack;
   char c;

   ttrack = 0;
   c = ' ';

   while (c != '\n')
   {
      c= name[whichseq][ttrack];
      if( c != '\n') printf("%1c",c);
      ttrack++;
   }
}

/* CHANGE BELOW SO THAT THERE ARE ROUTINES FOR BOTH
1. case where rates of one gene have been changed.
2. case where rates of all genes have been changed.

   ???

*/

/*  This routine calculates proposed state probability divided by current
state probability except for the fact the branch length terms are ignored.
It is assumed that ratio of proposal densitites is handled by routine
that calls this routine */
double probratio(proptime, proprate, proplogr, oldtime, oldrate, oldlogr)
double *proptime, *proprate, *proplogr, *oldtime, *oldrate, *oldlogr;
{
   double oldvar,newvar;
   double oldprop, newprop;
   double theratio, oldlength, newlength, oldrdif, newrdif;
   double rtratio, regratio, expratio;
   int panode, ktrack;

   expratio = 0.0;
   rtratio = 1.0;
   regratio = 1.0;
   if(brownmean > SMALLNUM)
   {
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         if(ktrack != consnode)
         {
            panode = tonode[ktrack][MAXOFF];
            oldlength = oldtime[panode] - oldtime[ktrack];
            oldvar = oldlength*oldrate[nnodes];
            oldrdif =  oldlogr[ktrack] - oldlogr[panode]+0.5*oldvar;
            newlength = proptime[panode] - proptime[ktrack];
            newvar =  newlength*proprate[ nnodes];
            newrdif =  proplogr[ktrack] - proplogr[panode]+0.5*newvar;
            if(newvar > 0.0)
            {
               rtratio *= oldvar/newvar;
               expratio += (newrdif*newrdif/newvar)
                  - (oldrdif*oldrdif/oldvar);
               regratio *= oldrate[ktrack]/proprate[ktrack];
            }
            else
            {
               rtratio = 0.0;
               expratio = 0.0;
               regratio = 0.0;
            }
         }
      }
      expratio *= (-0.5);
   }
   expratio += (timegamr - 1.0)*log(proptime[nnodes-1]/oldtime[nnodes-1])
      + timegamlam*(oldtime[nnodes-1] - proptime[nnodes-1]) +
      (rategamr - 1.0)*log(proprate[nnodes-1]/oldrate[nnodes-1]) +
      rategamlam*(oldrate[nnodes-1] - proprate[nnodes-1]); 
   if(brownmean > SMALLNUM)
   {
      expratio +=  browngamlam*(oldrate[nnodes] - proprate[nnodes]) +
         (browngamr - 1.0)*log(proprate[nnodes]/oldrate[nnodes]);
   }
   for(ktrack = ntips; ktrack < nnodes-1; ktrack++)
   {
         panode = tonode[ktrack][MAXOFF];
         newprop = proptime[ktrack]/proptime[panode];
         oldprop = oldtime[ktrack]/oldtime[panode];
         expratio += a[ktrack]*log(newprop/oldprop) + b[ktrack]*
            log((1.0-newprop)/(1.0-oldprop));
         regratio *= oldtime[panode]/proptime[panode];
   }
   theratio = sqrt(rtratio)*exp(expratio)*regratio;
   return(theratio);
}

/*  This routine calculates proposed state probability divided by current
state probability except for the fact the branch length terms are ignored.
It is assumed that ratio of proposal densitites is handled by routine
that calls this routine */
double multiratio(proptime, proprate, proplogr, oldtime, oldrate, oldlogr)
double *proptime, **proprate, **proplogr, *oldtime, **oldrate, **oldlogr;
{
   static double *oldvar,*newvar, *oldrdif, *newrdif;
   double oldprop, newprop;
   double theratio, oldlength, newlength;
   double rtratio, regratio, expratio;
   static int hereyet = 0;
   int panode, ktrack, ktrack2;

   if(hereyet ==0)
   {
      oldvar = dvec(ngenes);
      newvar = dvec(ngenes);
      oldrdif = dvec(ngenes);
      newrdif = dvec(ngenes);
      hereyet = 1;
   }

   expratio = 0.0;
   rtratio = 1.0;
   regratio = 1.0;
   if(brownmean > SMALLNUM)
   {
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         if(ktrack != consnode)
         {
            panode = tonode[ktrack][MAXOFF];
            oldlength = oldtime[panode] - oldtime[ktrack];
            newlength = proptime[panode] - proptime[ktrack];
            for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
            {
               oldvar[ktrack2] = oldlength*oldrate[ktrack2][nnodes];
               oldrdif[ktrack2] =  oldlogr[ktrack2][ktrack] - 
                  oldlogr[ktrack2][panode]+0.5*oldvar[ktrack2];
               newvar[ktrack2] =  newlength*proprate[ktrack2][ nnodes];
               newrdif[ktrack2] =  proplogr[ktrack2][ktrack] - 
                  proplogr[ktrack2][panode]+0.5*newvar[ktrack2];
               if(newvar[ktrack2] > 0.0)
               {
                  rtratio *= oldvar[ktrack2]/newvar[ktrack2];
                  expratio +=(newrdif[ktrack2]*newrdif[ktrack2]/newvar[ktrack2])
                     - (oldrdif[ktrack2]*oldrdif[ktrack2]/oldvar[ktrack2]);
                  regratio *=oldrate[ktrack2][ktrack]/proprate[ktrack2][ktrack];
               }
               else
               {
                  rtratio = 0.0;
                  expratio = 0.0;
                  regratio = 0.0;
               }
            }
         }
      }
      expratio *= (-0.5);
   }
   expratio += (timegamr - 1.0)*log(proptime[nnodes-1]/oldtime[nnodes-1])
      + timegamlam*(oldtime[nnodes-1] - proptime[nnodes-1]);
   for(ktrack2=0; ktrack2 <= ngenes; ktrack2++)
   {  
      expratio += (rategamr - 1.0)*log(proprate[ktrack2][nnodes-1]/
         oldrate[ktrack2][nnodes-1]) + rategamlam*(oldrate[ktrack2][nnodes-1] - 
         proprate[ktrack2][nnodes-1]); 
   }
   if(brownmean > SMALLNUM)
   {
      if(commonbrown ==1)
      {
/* all genes share same tendency to change rates */
         expratio +=  browngamlam*(oldrate[0][nnodes] -
            proprate[0][nnodes]) + (browngamr - 1.0)*
            log(proprate[0][nnodes]/oldrate[0][nnodes]);
      }
      else
      {
/* commonbrown = 0, so all genes have individual tendencies to change
   rates */
         for(ktrack2=0;ktrack2<=ngenes;ktrack2++)
         {
            expratio +=  browngamlam*(oldrate[ktrack2][nnodes] - 
               proprate[ktrack2][nnodes]) + (browngamr - 1.0)*
               log(proprate[ktrack2][nnodes]/oldrate[ktrack2][nnodes]);
         }
      }
   }
   for(ktrack = ntips; ktrack < nnodes-1; ktrack++)
   {
         panode = tonode[ktrack][MAXOFF];
         newprop = proptime[ktrack]/proptime[panode];
         oldprop = oldtime[ktrack]/oldtime[panode];
         expratio += a[ktrack]*log(newprop/oldprop) + b[ktrack]*
            log((1.0-newprop)/(1.0-oldprop));
         regratio *= oldtime[panode]/proptime[panode];
   }
   theratio = sqrt(rtratio)*exp(expratio)*regratio;
   return(theratio);
}

void rdtopology(treefile,ttonode,tnnodes,tntips,tmean,tname)
int **ttonode, *tnnodes,*tntips;
double *tmean;
char **tname;
FILE *treefile;
{
   static hereyet=0;
   int ktrack, ktrack2;
   char thischar;
 
   void docomma();
   void doleftpar();

   if(hereyet==0)
   {
      for(ktrack=0;ktrack < MAXNODE; ktrack++)
      {
         ttonode[ktrack][0]=0;
      }
      hereyet = 1;
   }
   else
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {
         ttonode[ktrack][0]=0;
      }
   }

   ttonode[0][MAXOFF]= -1;

   currentnode=0;
   (*tntips) =1;
   (*tnnodes) =1;
   ktrack=0;
   thischar= ' ';
   while(thischar != ';' )
   {
      thischar=getc(treefile);
      while(thischar == '\n')
      {
        thischar=getc(treefile);
      }
      switch(thischar)
      {
         case ')':
            currentnode = ttonode[currentnode][MAXOFF];
            ktrack=0;
            break;
         case '(':
            doleftpar(ttonode,tnnodes,tntips,tmean);
            ktrack=0;
            break;
         case ',':
            docomma(ttonode, tnnodes, tntips,tmean);
            ktrack=0;
            break;
         case ':':
            fscanf(treefile,"%lf",&tmean[currentnode]);
/* Do we need to do something somewhere to make sure correct number
of changes is assigned to each branch as tree grows? */

            ktrack=0; /* docolon() */
            break;
         case ';':
            ktrack=0;
            break;
         case ' ':
            break;
         default:
            tname[currentnode][ktrack] = thischar;
            ktrack++;
            tname[currentnode][ktrack] = '\n';
            break;
      }
   }   
/* CHANGE OR MOVE BELOW ELSEWHERE?
      for(ktrack = 0; ktrack < (*tnnodes); ktrack++)
      {
         for(ktrack2 = 0; ktrack2 <= MAXOFF ; ktrack2++)
         {
            if( ttonode[ktrack][ktrack2] != tonode[ktrack][ktrack2] &&
               ((ktrack2 == MAXOFF) || (ktrack2 <= tonode[ktrack][0])) )
            {
               printf("PROB:tonode[%3d][%3d] = %3d != %3d =ttonode[%3d][%3d]\n"
                  , ktrack, ktrack2, tonode[ktrack][ktrack2], 
                  ttonode[ktrack][ktrack2], ktrack, ktrack2);
            }
         }
      }
*/
}

void docomma(ttonode,tnnodes,tntips,tmean)
double *tmean;
int **ttonode, *tnnodes, *tntips;
{
   int ktrack, ktrack2;

   currentnode = ttonode[currentnode][MAXOFF];
   (*tnnodes)++;
   (*tntips)++;
   for(ktrack= (*tnnodes)-2;ktrack >= (*tntips)-1; ktrack--)
   {
      if(ktrack== (*tnnodes)-2)
      {
         ttonode[ (*tnnodes)-1][MAXOFF] = -1;
      }

      else
      {   
         tmean[ktrack+1] = tmean[ktrack];
         ttonode[ktrack+1][MAXOFF] = ttonode[ktrack][MAXOFF]+1;
      }
      ttonode[ktrack+1][0] = ttonode[ktrack][0];
      for(ktrack2=1; ktrack2 <= ttonode[ktrack][0];ktrack2++)
      {
         if(ttonode[ktrack][ktrack2] >= (*tntips)-1)
         {
            ttonode[ktrack+1][ktrack2] = ttonode[ktrack][ktrack2] +1;
         }
         else
         {   
            ttonode[ktrack+1][ktrack2] = ttonode[ktrack][ktrack2];
         }
      }   
   }   
   ttonode[ (*tntips)-1][MAXOFF] = currentnode;
   ttonode[ (*tntips)-1][0]=0;
   ttonode[currentnode+1][0]++;
   ttonode[currentnode+1][ttonode[currentnode+1][0]] = (*tntips)-1;
   currentnode =  (*tntips)-1;
   for(ktrack= (*tntips)-1; ktrack >=0;ktrack--)
   {
      ttonode[ktrack][MAXOFF]++;
   }
}

 
void doleftpar(ttonode, tnnodes,tntips,tmean)
double *tmean;
int **ttonode, *tnnodes, *tntips;
{
   int ktrack, ktrack2;

   (*tnnodes)++;
   for(ktrack= (*tnnodes)-2;ktrack >= *tntips; ktrack--)
   {
      if(ktrack== (*tnnodes)-2)
      {
         ttonode[ (*tnnodes)-1][MAXOFF] = -1;
      }
      else
      {   
         tmean[ktrack+1] = tmean[ktrack];
         ttonode[ktrack+1][MAXOFF] = ttonode[ktrack][MAXOFF]+1;
      }
      ttonode[ktrack+1][0] = ttonode[ktrack][0];
      for(ktrack2=1;ktrack2 <= ttonode[ktrack][0];ktrack2++)
      {
         if(ttonode[ktrack][ktrack2] >= *tntips)
         {
            ttonode[ktrack+1][ktrack2] = ttonode[ktrack][ktrack2] +1;
         }
         else
         {   
            if(ttonode[ktrack][ktrack2] == currentnode)
            {
               ttonode[ktrack+1][ktrack2] = *tntips;
            }
            else
            {   
               ttonode[ktrack+1][ktrack2] = ttonode[ktrack][ktrack2] ;
            }
         }   
      }   
   }   
   if( ttonode[currentnode][MAXOFF] == -1)
   {
      ttonode[ *tntips][MAXOFF] = -1;
   }
   else
   {   
      ttonode[ *tntips][MAXOFF] = ttonode[currentnode][MAXOFF] + 1;
   }
   ttonode[ *tntips][0]=1;
   ttonode[ *tntips][1]=currentnode;
   for(ktrack= (*tntips)-1; ktrack >=0;ktrack--)
   {
      if(ktrack == currentnode)
      {
         ttonode[currentnode][MAXOFF] = *tntips;
      }

      else
      {   
         ttonode[ktrack][MAXOFF]++;
      }
   }   
}

double snorm()
{
   double rnd();

   static int onemore=0;
   static double rankeep;
   double u1,u2,x1,x2;

   if(onemore==0)
   {
      u1=rnd(seed);
      u2=rnd(seed);
      x1= sqrt(-2.0*log(u1))*cos(2.0*PI*u2);
      x2= sqrt(-2.0*log(u1))*sin(2.0*PI*u2);
      rankeep=x2;
   }
   else
   {
      x1=rankeep;
   }
   onemore= 1-onemore;
   return(x1);
}

double exptial(explambda)
double explambda;
{
   double expvalue;
   extern double rnd();
   expvalue = -(1/ explambda) * log( 1 - rnd(seed));
   return(expvalue);
}

  
/* From Joe Felsenstein's PHYLIP package ...
   check for "needed by rnd" for other lines of code this routine needs */
double rnd(seed)
long *seed;
{
  /* random number generator -- slow but machine independent */
  static long i, j, k, sum;
  static longer mult, newseed;
  static double x;
  mult[0] = 13;
  mult[1] = 24;
  mult[2] = 22;
  mult[3] = 6;
  for (i = 0; i <= 5; i++)
    newseed[i] = 0;
  for (i = 0; i <= 5; i++) {
    sum = newseed[i];
    k = i;
    if (i > 3)
      k = 3;
    for (j = 0; j <= k; j++)
      sum += mult[j] * seed[i - j];
    newseed[i] = sum;
    for (j = i; j <= 4; j++) {
      newseed[j + 1] += newseed[j] / 64;
      newseed[j] &= 63;
    }
  }  
  memcpy(seed, newseed, sizeof(longer));
  seed[5] &= 3;
  x = 0.0;
  for (i = 0; i <= 5; i++)
    x = x / 64.0 + seed[i];
  x /= 4.0;
  return x;
}   /* rnd */


/* returns the ranks for each entry
    of a vector of doubles named vec of length veclen 
   (running from vec[0] to vec[veclen-1]).  The ranks
   are placed into a vector of veclen named rankvec and
   the ranks range from 0 (lowest value) to veclen-1 (highest value) 
    */
void dotherank(vec,veclen,rankvec)
double *vec, *rankvec;
int veclen;
{
   int ktrack, ktrack2, rankdummy, *rank;
   double *orderedvec, dummy;

   rank = ivec(veclen);
   orderedvec = dvec(veclen); 
   for(ktrack=0; ktrack < veclen; ktrack++)
   {
      orderedvec[ktrack] = vec[ktrack];
      rank[ktrack] = ktrack;
   }
   for(ktrack=1; ktrack < veclen; ktrack++)
   {
      for(ktrack2=ktrack; ktrack2>=1; ktrack2--)
      {
         if(orderedvec[ktrack2] < orderedvec[ktrack2-1])
         {
            dummy = orderedvec[ktrack2];
            rankdummy = rank[ktrack2];
            orderedvec[ktrack2] = orderedvec[ktrack2-1];
            rank[ktrack2] = rank[ktrack2-1];
            orderedvec[ktrack2-1] = dummy;
            rank[ktrack2-1] = rankdummy;
         }
         else
         {
            break;
         }
      }
   }   
   for(ktrack =0; ktrack < veclen; ktrack++)
   {
      rankvec[rank[ktrack]] = ktrack * 1.0;
   }
   freedvec(orderedvec);
   freeivec(rank);
}



/* copies a vector of doubles named vec of length veclen 
   (running from vec[0] to vec[veclen-1]) into 
   orderedvec and then sorts orderedvec from lowest to highest. 
   -- notice that the vector vec DOES NOT
   run from vec[1]to vec[veclen] */
void sort(vec,veclen,orderedvec)
double *vec,*orderedvec;
long veclen;
{
   int ktrack, ktrack2;
   double dummy;
   for(ktrack=0; ktrack < veclen; ktrack++)
   {
      orderedvec[ktrack] = vec[ktrack];
   }
   for(ktrack=1; ktrack < veclen; ktrack++)
   {
      for(ktrack2=ktrack; ktrack2>=1; ktrack2--)
      {
         if(orderedvec[ktrack2] < orderedvec[ktrack2-1])
         {
            dummy = orderedvec[ktrack2];
            orderedvec[ktrack2] = orderedvec[ktrack2-1];
            orderedvec[ktrack2-1] = dummy;
         }
         else
         {
            break;
         }
      }
   }   
}



FILE *gfopen(char *filename, char *mode)
{
   FILE *fp=(FILE*)fopen(filename, mode);

   if(fp==NULL) {
      printf("\nerror when opening file %s\n", filename);
      if(!strchr(mode,'r')) exit(-1);
      printf("tell me the full path-name of the file? ");
      scanf("%s", filename);
      if((fp=(FILE*)fopen(filename, mode))!=NULL)  return(fp);
      puts("Can't find the file.  I give up.");
      exit(-1);
   }
   return(fp);
}
