/* (c) Copyright 1999, 2000 by Jeff ThorneQ

VERSION 10/4/00

As this program is very crude, please do not distribute it. Instead,
people can obtain hopefully improved versions by contacting me (Jeff Thorne)
(Also, please don't delete this message)
 
This version has the expected rate at a descendant node equal to the rate
at the ancestral node - this is different than the previous version which
had expected log rate at descendant equal to log rate of ancestor.

However, one arbitrarily selected branch that is attached to the
root has the rate of the node ending the branch being forced
to be identical to the rate at the root node.

To compile to analyze a data set, type:

cc -o divtime5b -fast divtime5b.c dlinpack.c arrayutl.c -lm
gcc divtime5b.c dlinpack.c arrayutl.c -lm -O  -o divtime5b

(use "cc" above instead of "gcc" if "cc" is the compiler to which you have
 access)

 MS VC++ 6.0:  
   cl -Ot -O2 -W4 divtime5b.c dlinpack.c arrayutl.c

To run divtime5b, type:

divtime5b infile treefile nodefile

where ...

infile contains the data to be analyzed ...

and ...

treefile is the name of the file in which to write the tree with 
actual estimated times ...

and ...

nodefile is the name of the file in which various summary output is
to be placed.

*/

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "arrayutl.h"
FILE *gfopen(char *filename, char *mode);

#define min2(A,B) ((A) < (B) ? (A) : (B))
#define min3(A,B,C) ((A) < (B) ? min2(A,C) : min2(B,C)) 
#define max2(A,B) ((A) > (B) ? (A) : (B))
#define max3(A,B,C) ((A) > (B) ? max2(A,C) : max2(B,C))

#define PI 3.1415927
#define BIGNEGLOG -1000 /* a big negative logarithm */
#define MAXNODE 200
#define MAXOFF 10 /* Maximum number of offspring nodes of a node */
#define SMALLNUM 1.0e-10 /* a number slightly greater than 0 */
#define NEGSMALLNUM -1.0e-10 /* a number slightly less than 0 */
#define NAMELEN 52 
#define TINYLEN 0.0001
#define TINYTIME 0.00001
#define SMALLRATE 0.0001
#define LOWFRAC 0.025 /* lowest 0.025 of samples cut from 
                         posterior interval */
#define HIGHFRAC 0.975 /* highest 0.025 of samples cut from
                          posterior interval */
/* needed by rnd() routine */
typedef long longer[6];
static longer seed;
 
int **tonode;
int nodata; /* value 0 to approximate posterior, value 1 to approx. prior */ 
int viral; /* value 0 means all tip sequences isolated at same time,
              value 1 means not all tip sequences isolated at same time */ 
int consnode, ntipcons;
int currentnode, nbranches, nnodes, ntips;
int starttype;
int nintnodes; /* number internal nodes on tree that are NOT root */
char **name;
double a[MAXNODE];
double b[MAXNODE];
double *mean, *varcovar, *hightime, *lowtime, *tipcons;
double timegamr, timegamlam, rategamr, rategamlam, browngamr, browngamlam;
double rttm, rttmsd,rtrate,rtratesd;
double brownmean; /* value 0.0 means constant rates (i.e., "perfect" clock)
                     whereas value > 0.0 means allow rates to evolve */
double brownsd, minab,newk, othk;
double thek, bigtime;
long failrate[MAXNODE], failtime[MAXNODE],failcons,sucrate[MAXNODE];
long suctime[MAXNODE],succons, sucboth, failboth, sucrateflate, failrateflate;
long suctimeflate, failtimeflate;
long sucpartcons, failpartcons, sucpartboth, failpartboth;
long sucparttime, failparttime, sucpartrate, failpartrate;
long numsamps, sampfreq, burnin;

main(int argc, char *argv[])
{

   void rddata();
   void initialize();
   void dothemcmc();
   void summarize();
   double rnd();

   double *time, *rate, **samprate, **samptime;
/* needed by rnd() routine */
   FILE *inseed; /* contains the seed */
   long i,initseed;
   double lastrnd;
   int ktrack;


   if(argc<2) argv[1]="o.estb";
   if(argc<3) argv[2]="o.tree";
   if(argc<4) argv[3]="o.node";
   printf("Input Data File = %s\n",argv[1]);
   printf("Output Treefile = %s\n",argv[2]);
   printf("Output Nodefile = %s\n",argv[3]);

/* needed by rnd() routine */
   inseed=(FILE*)gfopen("inseed","r");
   fscanf(inseed,"%ld",&initseed);
   fclose(inseed);
   printf("initseed = %ld\n", initseed);
   for (i = 0; i <= 5; i++) seed[i] = 0;
   i = 0;
   do {
     seed[i] = initseed & 63;
     initseed /= 64;
     i++;
   } while (initseed != 0);
   for(i=1; i<= 100; i++)
   {
      lastrnd = rnd(seed);
   }

   rddata(argv[1]);
   if(brownmean > SMALLNUM)
   {
      browngamr = brownmean*brownmean/(brownsd*brownsd);
      browngamlam = brownmean/(brownsd*brownsd);
   }

   timegamr = rttm*rttm/(rttmsd*rttmsd);
   timegamlam = rttm/(rttmsd*rttmsd);
   rategamr = rtrate*rtrate/(rtratesd*rtratesd);
   rategamlam = rtrate/(rtratesd*rtratesd);

   if(nodata==1)
   {
      printf("The NO DATA - EXPLORE YOUR PRIOR VERSION\n");
   }
   else
   {
      printf("Approximate the posterior version\n");
   }
   printf("sampfreq %6ld numsamps %6ld burnin %6ld\n",sampfreq,numsamps,burnin);
   printf("newk %6.3lf othk %6.3lf thek %6.3lf\n", newk, othk, thek);
   printf("brownmean %6.3lf brownsd %6.3lf\n", brownmean, brownsd);
   if(brownmean > SMALLNUM)
   {
      printf("browngamr %6.3lf browngamlam %6.3lf\n", browngamr, browngamlam);
   }
   else
   {
      printf("THIS ANALYSIS ASSUMES CONSTANT RATE OF EVOLUTION ");
      printf("(i.e., so-called perfect clock)\n");
   }
   printf("rttm %6.3lf rttmsd %6.3lf minab %6.3lf\n", rttm, rttmsd,
      minab);
   printf("timegamr %6.3lf timegamlam %6.3lf\n", timegamr, timegamlam);
   printf("rtrate %8.4lf rtratesd %8.4lf\n", rtrate, rtratesd);
   printf("rategamr %6.3lf rategamlam %6.3lf\n", rategamr, rategamlam);

   printf("nbranches= %2d nnodes= %2d\n", nbranches, nnodes);
   nintnodes = nnodes - ntips -1;
   printf("ntips = %2d nintnodes = %2d\n", ntips, nintnodes);
   printf("bigtime = %8.3lf\n", bigtime);

   time = dvec(nnodes);
   rate = dvec(nnodes);
   for(ktrack=0; ktrack <= nnodes; ktrack++)
   {
      if(brownmean > SMALLNUM)
      {
         sucrate[ktrack] = 0;
         failrate[ktrack] = 0;
      }
      suctime[ktrack] = 0;
      failtime[ktrack] = 0;
   }
   if(brownmean > SMALLNUM)
   {
      succons = 0;
      failcons = 0;
      sucpartcons = 0;
      failpartcons = 0;
      sucpartboth = 0;
      failpartboth = 0;
      sucpartrate = 0;
      failpartrate = 0;
   }
   sucboth = 0;
   failboth = 0;
   sucrateflate = 0;
   failrateflate = 0;
   suctimeflate = 0;
   failtimeflate = 0;
   sucparttime = 0;
   failparttime = 0;

   samprate = dmat(nnodes+5,numsamps);
   samptime = dmat(nnodes+5,numsamps);
   initialize(time, rate, starttype); /* 3rd parameter=1 means 
      rndstart, other values mean start at hopefully good guess */
   if(viral == 0)
   {
      printf("All sequences (tips) isolated at same time\n");
   }
   else
   {
      printf("Sequences (tips) are not all isolated at same time\n");
   }
   dothemcmc(time,rate,samptime, samprate);
   puts("\nhave done the mcmc, now summarize.");
   summarize(samptime, samprate, argv[2],argv[3]);

/* needed by rnd() routine */
   initseed = (long) (1000000000*rnd(seed));
   if(initseed % 2 ==0)
   {
      initseed++;
   }
   inseed=gfopen("inseed","w");
   fprintf(inseed,"%ld",initseed);  /* put in a new random seed */
   fclose(inseed);

   freedmat(samprate,nnodes+5,numsamps);
   freedmat(samptime,nnodes+5,numsamps);
   freedvec(time);
   freedvec(tipcons);
   freedvec(hightime);
   freedvec(lowtime);
   freedvec(rate);
   freedvec(mean);
   freecmat(name,MAXNODE, NAMELEN);
   freedvec(varcovar);
   freeimat(tonode,MAXNODE,MAXOFF);
}


/* this routine assumes that the log-likelihood surface is being approximated
by a multivariate normal distribution with mean equal to the
maximum likelihood values of the branch lengths and the variance
derived from the curvature of the log-likelihood surface.  
*/
/* nodata=1 forces this routine to return
the same value no matter what the data are -- this option
is for investigating the prior distribution */
double logmvn(values)
double *values;
{
 
   static int hereyet = 0;
   int ktrack, ktrack2;
   double thesum;
   static double *tr2values;
   static double *tempvec;
   static double **invvcv;

   void partialinvert();
   extern double **dmat();
   extern double *dvec();
   extern double ***threedmat();
   extern void freedmat();

   if(nodata != 1)
   {

      if(hereyet==0)
      {
         invvcv = dmat(nbranches-1, nbranches-1);
         tempvec = dvec(nbranches-1);
         tr2values = dvec(nbranches-1);
         for(ktrack=0; ktrack < nbranches; ktrack++)
         {  
            for(ktrack2= 0 ; ktrack2 < nbranches; ktrack2++)
            {
               invvcv[ktrack][ktrack2] = varcovar[ktrack*nbranches+ktrack2];
            }
         }   
         partialinvert(invvcv, 0,nbranches-1);
         hereyet = 1;
      }
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         tr2values[ktrack] = values[ktrack] - mean[ktrack];
      }
      for(ktrack  =0; ktrack < nbranches; ktrack++)
      {
         tempvec[ktrack] = 0.0;
         for(ktrack2 = 0; ktrack2 < nbranches; ktrack2++)
         {  
            tempvec[ktrack] += tr2values[ktrack2]* invvcv[ktrack2][ktrack];
         }  
      }   
      thesum = 0.0;
      for(ktrack=0; ktrack < nbranches; ktrack++)
      {
         thesum += tr2values[ktrack]*tempvec[ktrack];   
      }   
      thesum *= -0.5;
   }
   else
   {
      thesum= 0.0; /* NO  DATA VERSION */
   }
   return(thesum); 
}

/* this routine assumes that the log-likelihood surface is being approximated
by a multivariate normal distribution with mean equal to the
maximum likelihood values of the branch lengths and the variance
derived from the curvature of the log-likelihood surface.  the routine
assumes that branch lengths surrounding a particular node have changed
but all others are constant.  So, it calculates the density (up to a 
proportionality constant) of the branch lengths surrounding the node
conditional on all other branch lengths.  It returns the logarithm
of this conditional density. nodata = 1 causes this routine to return
the same value no matter what the data are -- this option
is for investigating the prior distribution */
double logcondmvn(values, whichnode)
double *values;
int whichnode; 
{

   static int hereyet = 0;
   static int **pos, **invpos;
   static int *maxbrpos;
   int holdindex,pivotnode, ktrack, ktrack2, ktrack3, brpos1, brpos2;
   double thesum;
   double **vcv, **dummy;
   static int maxmaxbrpos;
   static double ***pp;
   static double ***qq;
   static double *tr2values;
   static double *tr3;

   void partialinvert();
   extern int *ivec();
   extern int **imat();
   extern double **dmat();
   extern double *dvec();
   extern double ***threedmat();
   extern void freedmat();

   if(nodata !=1)
   {
      if(hereyet==0)
      {
         vcv = dmat(nbranches-1, nbranches-1);
         for(ktrack=0; ktrack < nbranches; ktrack++)
         {
            for(ktrack2= 0 ; ktrack2 < nbranches; ktrack2++)
            {
               vcv[ktrack][ktrack2] = varcovar[ktrack*nbranches+ktrack2];
            }
         }
         pos = imat(nnodes-1, nnodes-1);
         invpos = imat(nnodes-1, nnodes-1);
         maxbrpos = ivec(nnodes-1);
         maxmaxbrpos = 0;
         for(pivotnode=0; pivotnode < nnodes; pivotnode++)
         {
            if(pivotnode == consnode)
            {
               maxbrpos[pivotnode] =tonode[consnode][0] +
                  tonode[nnodes-1][0];
            }
            else
            {
               if(pivotnode == nnodes-1)
               {
                  maxbrpos[pivotnode] = tonode[nnodes-1][0];
               }
               else
               {
                  maxbrpos[pivotnode] = 1+tonode[pivotnode][0];
               }
            }        
            maxmaxbrpos = max2(maxmaxbrpos,maxbrpos[pivotnode]);
         }
         maxmaxbrpos -= 1;
         pp = threedmat(nnodes-1, maxmaxbrpos, nnodes-1);
         qq = threedmat(nnodes-1, maxmaxbrpos, maxmaxbrpos);
         tr2values = dvec(nnodes-1);
         tr3 = dvec(nnodes-1);
         dummy = dmat(nnodes-1,nnodes-1);
         for(pivotnode=0; pivotnode < nnodes; pivotnode++)
         {
            brpos1 = 0;
            brpos2 = 0;
            for(ktrack = 0; ktrack < nbranches; ktrack++)
            {
               if((ktrack== pivotnode) || (pivotnode == consnode 
                  && (nnodes-1) == tonode[ktrack][MAXOFF]) || (
                  tonode[ktrack][MAXOFF] == pivotnode))
               {      
                  invpos[pivotnode][brpos1] = ktrack;
                  pos[pivotnode][ktrack] = brpos1;
                  brpos1++;
               }
               else
               {
                  invpos[pivotnode][brpos2+maxbrpos[pivotnode]] = ktrack;
                  pos[pivotnode][ktrack] = brpos2+maxbrpos[pivotnode];
                  brpos2++;
               }
            }
            for(ktrack = 0; ktrack < nbranches; ktrack++)
            {
               for(ktrack2=0; ktrack2 < nbranches; ktrack2++)
               {
                  dummy[ktrack][ktrack2] = vcv[invpos[pivotnode][ktrack]][
                     invpos[ pivotnode][ktrack2]];
               }
            }
            partialinvert(dummy, maxbrpos[pivotnode],nbranches-1); 
            for(ktrack=0; ktrack< maxbrpos[pivotnode]; ktrack++)
            {
               for(ktrack2=maxbrpos[pivotnode]; ktrack2 < nbranches; ktrack2++)
               {
                  pp[pivotnode][ktrack][ktrack2-maxbrpos[pivotnode]] = 0.0;
                  for(ktrack3=maxbrpos[pivotnode];ktrack3 < nbranches;ktrack3++)
                  {
                     pp[pivotnode][ktrack][ktrack2-maxbrpos[pivotnode]] +=
                        dummy[ktrack][ktrack3]*dummy[ktrack3][ktrack2];
                  }
               }
            }
            for(ktrack=0; ktrack < maxbrpos[pivotnode]; ktrack++)
            {
               for(ktrack2=0; ktrack2 < maxbrpos[pivotnode]; ktrack2++)
               {
                  qq[pivotnode][ktrack][ktrack2]=vcv[invpos[pivotnode][ktrack]][
                     invpos[pivotnode][ktrack2]];
                  for(ktrack3=0; ktrack3 < nbranches-maxbrpos[pivotnode]; 
                     ktrack3++)
                  {
                     qq[pivotnode][ktrack][ktrack2] -= pp[pivotnode][ktrack][
                        ktrack3] * dummy[ktrack3+maxbrpos[pivotnode]][ktrack2];
                  }
               }
            }
            partialinvert(qq[pivotnode],0,maxbrpos[pivotnode]-1);
         }
         freedmat(vcv,nbranches-1, nbranches-1);
         freedmat(dummy,nnodes-1,nnodes-1);
         hereyet = 1;
      }
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         holdindex = invpos[whichnode][ktrack];
         tr2values[ktrack] = values[holdindex] - mean[holdindex];
      }
      for(ktrack  =0; ktrack < maxbrpos[whichnode]; ktrack++)
      {
         for(ktrack2 = 0; ktrack2 < nbranches-maxbrpos[whichnode]; ktrack2++)
         {
            tr2values[ktrack] -= pp[whichnode][ktrack][ktrack2] * 
            tr2values[ktrack2+maxbrpos[whichnode]];   
         }
      } 
      for(ktrack=0; ktrack < maxbrpos[whichnode]; ktrack++)
      {
         tr3[ktrack] = 0.0;
         for(ktrack2=0; ktrack2 < maxbrpos[whichnode]; ktrack2++)
         {
            tr3[ktrack] += tr2values[ktrack2]*qq[whichnode][ktrack2][ktrack];
         }
      }
      thesum = 0.0;
      for(ktrack=0; ktrack < maxbrpos[whichnode]; ktrack++)
      {
         thesum += tr3[ktrack]*tr2values[ktrack];
      }
      thesum *= -0.5; /* logarithm of the conditional density */
   }
   else
   {
      thesum = 0.0; /* NO  DATA VERSION */
   }
   return(thesum); 
}


void partialinvert(a,lower, upper)
int lower,upper;
double **a;
{
  int i, j, k ;
  double temp;

  /*Gauss-Jordan reduction -- invert matrix a in place,
         overwriting previous contents of a.  On exit, matrix a
         contains the inverse.*/
  for (i = lower; i <= upper; i++) {
    temp = 1.0 / a[i][i];
    a[i][i] = 1.0;
    for (j = lower; j <= upper; j++) {
      a[i][j] *= temp;
    }
    for (j = lower; j <= upper; j++) {
      if (j != i) {
	temp = a[j][i];
	a[j][i] = 0.0;
	for (k = lower; k <= upper; k++) {
	  a[j][k] -= temp * a[i][k];
	}
      }
    }
  }
}  /*invert*/


void rddata(treefilename)
char *treefilename;
{
   void rdinfo();
   void rdcntrl();

   int stopflag, thenode, ktrack, ktrack2, ktrack3, ktrack4;
   int numchildren, checkthem, keepgoing, child[MAXNODE], *order, place;
   FILE *treefile;
   char thischar, *tempname;
 
   treefile = gfopen(treefilename,"r");
/* read rooted topology */
   printf("Before rdinfo\n");

   rdinfo(treefile);
   printf("After rdinfo\n");

   while(getc(treefile) != '\n');
   while(getc(treefile) != '\n');

   order = ivec(nnodes);
   hightime = dvec(nnodes);
   lowtime = dvec(nnodes);
   tipcons = dvec(ntips);
   varcovar = dvec((nbranches+1) * (nbranches+1));
   tempname = cvec(NAMELEN);

   place = 0;
   for(ktrack=0; ktrack < ntips; ktrack++)
   {
      thischar = getc(treefile);
      while(thischar != ' ')
      {
         tempname[place] = thischar;
         thischar = getc(treefile);
         place++;
      } 
      tempname[place] = '\n';
      place = 0;
      fscanf(treefile,"%d\n", &thenode);
      stopflag = 0;
      for(ktrack2=0; ktrack2 < ntips; ktrack2++)
      {
         for(ktrack3=0; ktrack3 <= NAMELEN; ktrack3++)
         {
            if( tempname[ktrack3] != name[ktrack2][ktrack3])
            {
               break;
            }
            else
            {
               if(tempname[ktrack3] == '\n')
               {
                  order[thenode] = ktrack2;
                  stopflag = 1;
                  break;
               }
            }
         }
         if( stopflag ==1)
         {
            break;
         }
      }
   }
   while(getc(treefile) != '\n');
   for(ktrack = ntips; ktrack < nnodes; ktrack++)
   {
      numchildren = 2;
      fscanf(treefile,"%d%d%d", &child[1], &child[2], &thenode);
      if(getc(treefile) != '\n')
      {
         keepgoing = 1;
         numchildren++;
         child[numchildren] = thenode;
         while(keepgoing ==1)
         {
            fscanf(treefile, "%d", &thenode);
            if(getc(treefile) != '\n' )
            {
               numchildren++;
               child[numchildren] = thenode;
            }
            else
            {
               keepgoing = 0;
            }
         }
      }
      for(ktrack2= ntips; ktrack2 <= nnodes; ktrack2++)
      { 
         if(numchildren == tonode[ktrack2][0])
         {
            for(ktrack4 = 1; ktrack4 <= numchildren; ktrack4++)
            {
               for(ktrack3=1; ktrack3 <= numchildren; ktrack3++)
               {
                  if(order[child[ktrack3]] != tonode[ktrack2][ktrack4])
                  {
                     checkthem =1;
                  }
                  else
                  {
                     checkthem = 0;
                     break;
                  }
               }   
               if(checkthem == 1)
               {
                  break;
               }
            }
         }
         else
         {
            checkthem =1;
         }
         if(checkthem ==0)
         {
            order[thenode] = ktrack2;
            break;
         }
         if(ktrack2 == nnodes)
         {
            printf("trouble mapping names and nodes to tree\n");
            break;
         }
      }
   }

   rdcntrl();
   for(ktrack =0; ktrack < ntips; ktrack++)
   {
      lowtime[ktrack] = tipcons[ktrack];
      hightime[ktrack] = tipcons[ktrack];
   }
   for(ktrack= 0; ktrack < nnodes-1; ktrack++)
   {
      lowtime[tonode[ktrack][MAXOFF]] = max2( 
         lowtime[tonode[ktrack][MAXOFF]], lowtime[ktrack]);
   }
   for(ktrack= nnodes-2; ktrack  >= 0; ktrack--)
   {
      hightime[ktrack] = min2(hightime[ktrack],
         hightime[tonode[ktrack][MAXOFF]]);
   }
   while(getc(treefile) != '\n');
   
/* read covariances */
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      for(ktrack2=0; ktrack2 < nbranches; ktrack2++)
      {
         fscanf(treefile, "%lf", &varcovar[order[ktrack]*nbranches +
            order[ktrack2]]);
      }
   }

   freeivec(order);
   freecvec(tempname);
   fclose(treefile);
}

void rdcntrl()
{
   int ktrack, ncons, whichnode;
   char constype;
   double thecons;
   FILE *datfile;

   datfile = gfopen("divcntrl.dat", "r");

   while ( getc(datfile) != '\n');
   while ( getc(datfile) != '\n');
   while ( getc(datfile) != '\n'); 

   fscanf(datfile, "%ld", &numsamps);
   while ( getc(datfile) != '\n'); 
   fscanf(datfile, "%ld", &sampfreq);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%ld", &burnin);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%d", &starttype);
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &rttm);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &rttmsd);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &rtrate);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &rtratesd);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &brownmean);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &brownsd);
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &minab);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &newk);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &othk);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &thek);
   while ( getc(datfile) != '\n');  
   fscanf(datfile, "%lf", &bigtime);
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   while ( getc(datfile) != '\n');  
   fscanf(datfile,"%d", &ncons);
   while ( getc(datfile) != '\n');  

   for(ktrack = 0; ktrack < nnodes; ktrack++)
   {
      lowtime[ktrack] = 0.0;
      hightime[ktrack] = bigtime;
   }

   for(ktrack=1; ktrack <= ncons; ktrack++)
   {
      fscanf(datfile, "%1c%d%lf", &constype, &whichnode, &thecons);
      if(constype == 'L')
      {
         lowtime[whichnode] = thecons;
      }
      else
      {
         if(constype == 'U')
         {
            hightime[whichnode] = thecons;
         }
         else
         {
            printf("Problem: constraint type is %1c ... What does this mean?\n", 
               constype);
         }
      }
      while ( getc(datfile) != '\n');  
   }
   for(ktrack=0; ktrack < ntips; ktrack++)
   {
      tipcons[ktrack] = 0.0;
   }
   fscanf(datfile, "%d", &ntipcons);
   while ( getc(datfile) != '\n');  
   for(ktrack=1; ktrack <= ntipcons; ktrack++)
   {
      fscanf(datfile, "%1c%d%lf", &constype, &whichnode, &thecons);
      if(constype == 'T')
      {
         tipcons[whichnode] = thecons;
      }
      else
      {
         printf("Problem tipcons: constype is %1c ... What does this mean?\n",
            constype);
      }
      while ( getc(datfile) != '\n');
   }
   fscanf(datfile,"%d", &nodata);
   fclose(datfile);
}


void initialize(time,rate,rndstart)
int rndstart;
double *time, *rate;
{

   double rnd();
   double exptial();
   
   int ktrack, ktrack2, highnode;
   int maxdepth, thisparent,depth[MAXNODE];
   double *mfac, *tblen, holdptime;
   double temptime,maxdis, *dist;


   mfac = dvec(nnodes);
   tblen = dvec(nbranches);


   for(ktrack=0; ktrack < ntips; ktrack++)
   {
      mfac[ktrack] = 1.0;
      depth[ktrack] = 0;
   }
   for(ktrack=ntips; ktrack < nnodes; ktrack++)
   {
      mfac[ktrack] = 1.0;
      maxdepth = depth[tonode[ktrack][1]];
      for(ktrack2 = 2; ktrack2 <= tonode[ktrack][0]; ktrack2++)
      {
         maxdepth = max2(maxdepth, depth[tonode[ktrack][ktrack2]]);
      }
      depth[ktrack] = 1+ maxdepth;
   }
   for(ktrack = ntips; ktrack < nnodes-1; ktrack++)
   {
      a[ktrack] = minab*depth[ktrack];
      b[ktrack] = minab;
      a[ktrack] -= 1.0;
      b[ktrack] -= 1.0;
   }        
   viral = 0;
   for(ktrack = 0; ktrack < ntips; ktrack++)
   {
      time[ktrack] = tipcons[ktrack];
      if(time[ktrack] > SMALLNUM)
      {
         viral = 1;
      }
   }
   for(ktrack=1; ktrack <= tonode[nnodes-1][0]; ktrack++)
   {
      if(tonode[nnodes-1][ktrack] >= ntips || 
         time[tonode[nnodes-1][ktrack]] == 0.0)
      {
         consnode = tonode[nnodes-1][ktrack];
         break;
      }
   }
   printf("starttype = %1d ... ", rndstart);
   if( rndstart != 1)
   {
      printf("starting from a hopefully good guess\n");
      maxdis = 0.0;
      dist = dvec(ntips-1);
      for(ktrack = 0; ktrack < ntips; ktrack++)
      {
         highnode = tonode[ktrack][MAXOFF];
         dist[ktrack] = max2(0.01,mean[ktrack]);
         while(tonode[highnode][MAXOFF] != -1)
         {
            dist[ktrack] += max2(0.01,mean[highnode]);
            highnode = tonode[highnode][MAXOFF];
         }
         if(maxdis < dist[ktrack])
         {
            maxdis = dist[ktrack];
         }
      }

      rate[nnodes-1] = rtrate;
      rate[consnode] = rtrate;
      time[nnodes-1] = maxdis/rtrate;
      for(ktrack = nnodes-2; ktrack >= ntips; ktrack--)
      {
         time[ktrack] = time[tonode[ktrack][MAXOFF]] - 
            max2(mean[ktrack]/rtrate, 0.01);
      }
      if(brownmean > SMALLNUM)
      {
         for(ktrack = nnodes-2; ktrack >= 0; ktrack--)
         {
            if(ktrack != consnode)
            {
               rate[ktrack] = 2.0*max2(0.01,mean[ktrack])/(
                  time[tonode[ktrack][MAXOFF]]- time[ktrack]) - 
                  rate[tonode[ktrack][MAXOFF]];
               rate[ktrack] = max2(rate[ktrack], SMALLRATE);
            }
         }
      }
      else
      {
          for(ktrack=0; ktrack < nnodes-1; ktrack++)
          {
             rate[ktrack] = rate[nnodes-1];
          }
      }
      rate[nnodes] = brownmean; 
         /* initialize BROWNIAN motion constant */
   }
   else
   {
      printf("starting at semi-random guess\n");
      time[nnodes-1] = exptial(1.0/rttm);
      if(brownmean > SMALLNUM)
      {
         rate[nnodes] = exptial(1.0/brownmean);
      }
      else
      {
         rate[nnodes] = 0.0;
      }
      if(brownmean > SMALLNUM)
      {
         for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
         {
            rate[ktrack] = 0.5+rnd(seed)*1.5;
         }
         rate[consnode] = rate[nnodes-1];
      }
      else
      {
         rate[nnodes-1] = 0.1*rtrate + 1.9*rtrate*rnd(seed);
         for(ktrack=0; ktrack < nnodes-1; ktrack++)
         {
            rate[ktrack] = rate[nnodes-1];
         }
      }

      for(ktrack=nnodes-2; ktrack >= ntips; ktrack--)
      {
         if(mean[ktrack] > TINYLEN)
         {
            time[ktrack] = time[tonode[ktrack][MAXOFF]] - 
            (mean[ktrack]/(0.5*(rate[ktrack]+rate[tonode[ktrack][MAXOFF]])));
         }
         else 
         {
            time[ktrack] = (1.0-TINYLEN)*time[tonode[ktrack][MAXOFF]];
         }
         if(time[ktrack] < TINYTIME)
         {
            time[ktrack] = (1.0-TINYTIME)*time[tonode[ktrack][MAXOFF]];
         }
      }  

   }

   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      tblen[ktrack] = (time[tonode[ktrack][MAXOFF]] - time[ktrack])*
         (rate[tonode[ktrack][MAXOFF]] + rate[ktrack])* 0.5;
   }

   for(ktrack= nnodes-1; ktrack >= ntips; ktrack--)
   {
      thisparent = tonode[ktrack][MAXOFF];
      if(ktrack != nnodes-1)
      {
         mfac[ktrack] = mfac[thisparent];
         time[ktrack] *= mfac[ktrack];
         holdptime = time[thisparent];
      }
      else
      {
         holdptime = hightime[ktrack];
      }
      temptime = time[ktrack];
      if(temptime > min2(holdptime, hightime[ktrack])
         || time[ktrack] < lowtime[ktrack])
      {
         if(rndstart == 1)
         {
            temptime = lowtime[ktrack] + rnd(seed)*(min2(holdptime,
               hightime[ktrack])- lowtime[ktrack]);
         }
         else
         {
            temptime = lowtime[ktrack] + 0.5*(min2(holdptime
               ,hightime[ktrack])- lowtime[ktrack]);
         }
         mfac[ktrack] *= temptime/time[ktrack];
      }
      time[ktrack] = temptime;
   }      
   rate[nnodes-1] = tblen[consnode]/(time[nnodes-1] - time[
      consnode]);
   if(brownmean > SMALLNUM)
   {
      rate[consnode] = rate[nnodes-1];
      for(ktrack = nnodes -2; ktrack >=0; ktrack--)
      {
         if(ktrack != consnode)
         {
            thisparent = tonode[ktrack][MAXOFF];
            rate[ktrack] = 2*tblen[ktrack]/(time[thisparent] -
               time[ktrack]) - rate[thisparent];
            if(rate[ktrack] < SMALLRATE)
            {
               rate[ktrack] = SMALLRATE;
            }
         }
      }
   }
   else
   {
      for(ktrack = 0; ktrack < nnodes-1; ktrack++)
      {
         rate[ktrack] = rate[nnodes-1];
      }
   }

   for(ktrack = 0 ; ktrack < nnodes; ktrack++)
   {
      printf("%3d:t=%9.4lf hight=%9.4lf lowt=%9.4lf rt=%9.4lf pnode=%3d\n",
         ktrack,time[ktrack],hightime[ktrack], lowtime[ktrack], rate[ktrack],
         tonode[ktrack][MAXOFF]);
   
   }

   freedvec(mfac);
   freedvec(tblen);
}

void dothemcmc(double *time, double *rate, double **samptime, double **samprate)
{
   int whichsamp, ktrack, ktrack2;
   FILE *sampfile;
   double *logr; /* save time by precalculating log of rates */

   void timechange();
   void ratechange();
   void changebrown();
   void consflate();
   void rateflate();
   void timeflate();
   void bothflate();
   void partrateflate();
   void parttimeflate();
   void partbothflate();
   void partconsflate();

   logr = dvec(nnodes);
   for(ktrack = 0; ktrack < nnodes; ktrack++)
      logr[ktrack] = log(rate[ktrack]);

   sampfile = gfopen("sampfile","w");
   for(ktrack = 0; ktrack <= burnin+ (numsamps-1)*sampfreq; ktrack++)
   {
      if( (ktrack >= burnin) && ((ktrack - burnin)%sampfreq== 0 ))
      {
         for(ktrack2=0; ktrack2<nnodes; ktrack2++)
         {
            fprintf(sampfile, "%10.5lf ", time[ktrack2]);
            whichsamp = (ktrack-burnin)/sampfreq;
            samptime[ktrack2][whichsamp] = time[ktrack2];
            samprate[ktrack2][whichsamp] = rate[ktrack2];
         }
         samprate[nnodes][whichsamp] = rate[nnodes];
         for(ktrack2=0; ktrack2<=nnodes; ktrack2++)
         {
            fprintf(sampfile, "%10.5lf ", rate[ktrack2]);
         }
         fprintf(sampfile,"\n");
      }
      timechange(time,rate,logr);
      if(brownmean > SMALLNUM)
      {
         ratechange(time,rate,logr);
         changebrown(time,rate,logr);
         consflate(time,rate,logr);
      }
      rateflate(time,rate,logr);
      timeflate(time,rate,logr);
      bothflate(time,rate,logr);
      /* below is necessary to insure that there is a subtree below
         root on which to inflate/deflate -- if nintnodes =1 then there
         is only 1 internal non-root node and this might be constrained
         to have same rate as root node */
      if(nintnodes > 1)
      {
         if(brownmean > SMALLNUM)
         {
            partconsflate(time,rate,logr);
            partrateflate(time,rate,logr);
            partbothflate(time,rate,logr);
         }
         parttimeflate(time,rate,logr);
      }

     if((ktrack+1)%10000==0)
        fprintf(stderr, "\r%10d  %4.1f%% finished", ktrack-burnin+1, (ktrack-burnin+1)/((numsamps-1.)*sampfreq)*100.);

   }
   putchar('\n');
   fclose(sampfile);
   freedvec(logr);
}



void parttimeflate(double *time, double *rate, double *logr)
{
   double probratio();
   double logmvn();
   double rnd();
   double ratio, randnum, invfac, flatefac, holdthis;
   double jterm, maxcurratio, maxnewratio;
   double lbound,ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   double lmin, lmax;
   static int **belownode;
   static double halfk, neghalfk, *brlen, *propbrlen;
   static double invothk;
   static double *newtime, *newrate, *newlogr;
   int ktrack, ktrack2;
   int thisnode, parentnode, thisparent;
   int chooseparent, choosenode;
   static int hereyet=0;
 
   if(hereyet ==0)
   {
      brlen = dvec(nbranches);
      propbrlen = dvec(nbranches);
 
      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      belownode = imat(nnodes,nnodes);
      for(ktrack = 0; ktrack <= nnodes; ktrack++)
      {
         for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
         {
            belownode[ktrack][ktrack2] = 0;
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         belownode[ktrack][ktrack] = 1;
/* belownode[k][nnodes] counts the number of nodes that are tips below node k
   belownode[nnodes][k] counts total number of node below and including node k
*/
         for(ktrack2= ktrack+1; ktrack2 < nnodes; ktrack2++)
         {
            if(belownode[ktrack2][tonode[ktrack][MAXOFF]] == 1)
            {
               belownode[ktrack2][ktrack] = 1;
            }
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         for(ktrack2 = 0; ktrack2 <= ktrack ; ktrack2++)
         {
            if(belownode[ktrack][ktrack2] == 1 )
            {
               if(ktrack2 < ntips)
               {
                  belownode[ktrack][nnodes]++;
               }
               belownode[nnodes][ktrack]++;
            }
         }
      }   


      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);

   }
   choosenode = consnode;
   while(choosenode == consnode)
   {
      choosenode = (int) ((nnodes-ntips-1)*rnd(seed)+ntips);
   }
   chooseparent = tonode[choosenode][MAXOFF];


   mincurratio = min2(time[chooseparent]/time[choosenode],
      hightime[choosenode]/time[choosenode]);
   maxcurratio = lowtime[choosenode]/time[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {  
         maxcurratio = max2(maxcurratio, lowtime[ktrack]/time[ktrack]);
         mincurratio = min2(mincurratio, hightime[ktrack]/time[ktrack]);
      }  
   }   

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(maxcurratio == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {   
      lmax = log(maxcurratio);
   }
   if(othk <  -2.0*lmax)
   {
      lownum = 0.0;
   }
   else
   {   
      lownum = 0.5+ lmax*invothk;
   }
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      newrate[ktrack] = rate[ktrack];
      newlogr[ktrack] = logr[ktrack];
      if(ktrack >= ntips && belownode[choosenode][ktrack] == 1)
      {  
         newtime[ktrack] = flatefac*time[ktrack];
      }   
      else
      {   
         newtime[ktrack] = time[ktrack];
      }  
   }   
   newrate[nnodes] = rate[nnodes];
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
         thisparent = tonode[ktrack][MAXOFF];
         holdthis = 0.5*(rate[thisparent]+rate[ktrack]);
         brlen[ktrack] = holdthis* (time[thisparent] - time[ktrack]);
         propbrlen[ktrack] = holdthis* (newtime[thisparent] - newtime[ktrack]);
   }

   minnewratio = min2(newtime[chooseparent]/newtime[choosenode],
      hightime[choosenode]/newtime[choosenode]);
   maxnewratio = lowtime[choosenode]/newtime[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {  
         maxnewratio = max2(maxnewratio, lowtime[ktrack]/newtime[ktrack]);
         minnewratio = min2(minnewratio, hightime[ktrack]/newtime[ktrack]);
      }  
   }   

   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(lmax == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {   
      lmax = log(maxnewratio);
   }
   if(othk > -2.0*lmax)
   {
      lownum = 0.5+ lmax*invothk;
   }
   else
   {   
      lownum = 0.0;
   }
   ratio = (jterm /(highnum-lownum))*
      exp((belownode[ktrack][nnodes] - belownode[nnodes][ktrack])*loginvfac
      + logmvn(propbrlen) -logmvn(brlen)) *
      probratio(newtime,newrate,newlogr,time, rate, logr);
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack <= choosenode; ktrack++)
      {  
         time[ktrack] = newtime[ktrack];
      }  
      sucparttime++;
   }
   else
   {   
      failparttime++;
   }
}

void partbothflate(time,rate,logr)
double *time, *rate, *logr;
{
   double probratio();
   double logmvn();
   double rnd();
   double ratio, randnum, invfac, flatefac ;
   double jterm, maxcurratio, maxnewratio;
   double lbound,ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   double lmin, lmax;
   static int **belownode;
   static double halfk, neghalfk, *brlen, *propbrlen;
   static double invothk;
   static double *newtime, *newrate, *newlogr;
   int ktrack, ktrack2;
   int thisnode, parentnode, thisparent;
   int chooseparent, choosenode;
   static int hereyet=0;
 
   if(hereyet ==0)
   {
      brlen = dvec(nbranches);
      propbrlen = dvec(nbranches);
 
      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      belownode = imat(nnodes,nnodes);
      for(ktrack = 0; ktrack <= nnodes; ktrack++)
      {
         for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
         {
            belownode[ktrack][ktrack2] = 0;
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         belownode[ktrack][ktrack] = 1;
/* belownode[k][nnodes] counts the number of nodes that are tips below node k
   belownode[nnodes][k] counts total number of node below and including node k
*/
         for(ktrack2= ktrack+1; ktrack2 < nnodes; ktrack2++)
         {
            if(belownode[ktrack2][tonode[ktrack][MAXOFF]] == 1)
            {
               belownode[ktrack2][ktrack] = 1;
            }
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         for(ktrack2 = 0; ktrack2 <= ktrack ; ktrack2++)
         {
            if(belownode[ktrack][ktrack2] == 1 )
            {
               if(ktrack2 < ntips)
               {
                  belownode[ktrack][nnodes]++;
               }
               belownode[nnodes][ktrack]++;
            }
         }
      }   

      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);

   }
   choosenode = consnode;
   while(choosenode == consnode)
   {
      choosenode = (int) ((nnodes-ntips-1)*rnd(seed)+ntips);
   }
   chooseparent = tonode[choosenode][MAXOFF];


   mincurratio = min2(time[chooseparent]/time[choosenode],
      hightime[choosenode]/time[choosenode]);
   maxcurratio = lowtime[choosenode]/time[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {  
         maxcurratio = max2(maxcurratio, lowtime[ktrack]/time[ktrack]);
         mincurratio = min2(mincurratio, hightime[ktrack]/time[ktrack]);
      }  
   }   

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(maxcurratio == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {   
      lmax = log(maxcurratio);
   }
   if(othk <  -2.0*lmax)
   {
      lownum = 0.0;
   }
   else
   {   
      lownum = 0.5+ lmax*invothk;
   }
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(belownode[choosenode][ktrack] == 1)
      {  
         newrate[ktrack] = invfac*rate[ktrack];
         newlogr[ktrack] = logr[ktrack] + log(invfac);
         if(ktrack >= ntips)
         {
            newtime[ktrack] = flatefac*time[ktrack];
         }
         else
         {
            newtime[ktrack] = time[ktrack];
         }
      }   
      else
      {   
         newtime[ktrack] = time[ktrack];
         newrate[ktrack] = rate[ktrack];
         newlogr[ktrack] = logr[ktrack];
      }  
   }   
   newrate[nnodes] = rate[nnodes];
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
         thisparent = tonode[ktrack][MAXOFF];
         brlen[ktrack] = 0.5*(rate[thisparent]+rate[ktrack])*
            (time[thisparent] - time[ktrack]);
         propbrlen[ktrack] = 0.5*(newrate[thisparent] + newrate[ktrack])*
            (newtime[thisparent] - newtime[ktrack]);
   }

   minnewratio = min2(newtime[chooseparent]/newtime[choosenode],
      hightime[choosenode]/newtime[choosenode]);
   maxnewratio = lowtime[choosenode]/newtime[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {  
         maxnewratio = max2(maxnewratio, lowtime[ktrack]/newtime[ktrack]);
         minnewratio = min2(minnewratio, hightime[ktrack]/newtime[ktrack]);
      }  
   }   

   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(lmax == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {   
      lmax = log(maxnewratio);
   }
   if(othk > -2.0*lmax)
   {
      lownum = 0.5+ lmax*invothk;
   }
   else
   {   
      lownum = 0.0;
   }
   ratio = (jterm /(highnum-lownum))*
      exp(belownode[choosenode][nnodes]*loginvfac
      + logmvn(propbrlen) -logmvn(brlen)) *
      probratio(newtime,newrate,newlogr,time, rate, logr);
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      rate[nnodes] = newrate[nnodes];
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {  
         rate[ktrack] = newrate[ktrack];
         time[ktrack] = newtime[ktrack];
         logr[ktrack] = newlogr[ktrack];
      }  
      sucpartboth++;
   }
   else
   {   
      failpartboth++;
   }
}

void partrateflate(time,rate,logr)
double *time, *rate, *logr;
{
   double probratio();
   double logmvn();
   double rnd();
   double ratio, randnum, invfac, flatefac, holdthis;
   double loginvfac;
   double runlog;
   static int **belownode;
   static double halfk, neghalfk, *brlen, *propbrlen;
   static double invothk;
   static double *newtime, *newrate, *newlogr;
   int ktrack, ktrack2;
   int thisnode, parentnode, thisparent;
   int chooseparent, choosenode;
   static int hereyet=0;
 
   if(hereyet ==0)
   {
      brlen = dvec(nbranches);
      propbrlen = dvec(nbranches);
 
      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      belownode = imat(nnodes,nnodes);
      for(ktrack = 0; ktrack <= nnodes; ktrack++)
      {
         for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
         {
            belownode[ktrack][ktrack2] = 0;
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         belownode[ktrack][ktrack] = 1;
/* belownode[k][nnodes] counts the number of nodes that are tips below node k
   belownode[nnodes][k] counts total number of node below and including node k
*/
         for(ktrack2= ktrack+1; ktrack2 < nnodes; ktrack2++)
         {
            if(belownode[ktrack2][tonode[ktrack][MAXOFF]] == 1)
            {
               belownode[ktrack2][ktrack] = 1;
            }
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         for(ktrack2 = 0; ktrack2 <= ktrack ; ktrack2++)
         {
            if(belownode[ktrack][ktrack2] == 1 )
            {
               if(ktrack2 < ntips)
               {
                  belownode[ktrack][nnodes]++;
               }
               belownode[nnodes][ktrack]++;
            }
         }
      }   

      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);

   }
   choosenode = consnode;
   while(choosenode == consnode)
   {
      choosenode = (int) ((nnodes-ntips-1)*rnd(seed)+ntips);
   }
   runlog = 0.0;
   chooseparent = tonode[choosenode][MAXOFF];


   loginvfac = halfk - othk*rnd(seed);
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      newtime[ktrack] = time[ktrack];
      if(belownode[choosenode][ktrack] == 1)
      {  
         newrate[ktrack] = flatefac*rate[ktrack];
         newlogr[ktrack] = logr[ktrack] - loginvfac;
      }   
      else
      {   
         newrate[ktrack] = rate[ktrack];
         newlogr[ktrack] = logr[ktrack];
      }  
   }   
   newrate[nnodes] = rate[nnodes];
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
         thisparent = tonode[ktrack][MAXOFF];
         holdthis = 0.5*(time[thisparent] - time[ktrack]);
         brlen[ktrack] = holdthis*(rate[thisparent]+rate[ktrack]);
         propbrlen[ktrack] = holdthis*(newrate[thisparent] + newrate[ktrack]);
   }

   ratio = exp(logmvn(propbrlen) -logmvn(brlen)- loginvfac*
      belownode[nnodes][choosenode]) *
      probratio(newtime,newrate,newlogr,time, rate, logr);
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack <= choosenode; ktrack++)
      {  
         rate[ktrack] = newrate[ktrack];
         logr[ktrack] = newlogr[ktrack];
      }  
      sucpartrate++;
   }
   else
   {   
      failpartrate++;
   }
}

void partconsflate(time,rate,logr)
double *time, *rate, *logr;
{
   double probratio();
   double logcondmvn();
   double rnd();
   double ratio, randnum, invfac, flatefac ;
   double jterm, maxcurratio, maxnewratio;
   double lbound,ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   double runlog, lmin, lmax, blen;
   static int **belownode;
   static double halfk, neghalfk, *brlen, *propbrlen;
   static double invothk;
   static double *newtime, *newrate, *newlogr;
   int ktrack, ktrack2;
   int thisnode, parentnode, thisparent;
   int chooseparent, choosenode;
   static int hereyet=0;

   if(hereyet ==0)
   {
      brlen = dvec(nbranches);
      propbrlen = dvec(nbranches);

      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      belownode = imat(nnodes,nnodes);
      for(ktrack = 0; ktrack <= nnodes; ktrack++)
      {
         for(ktrack2=0; ktrack2 <= nnodes; ktrack2++)
         {
            belownode[ktrack][ktrack2] = 0;
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         belownode[ktrack][ktrack] = 1;
/* belownode[k][nnodes] counts the number of nodes that are tips below node k
   belownode[nnodes][k] counts total number of node below and including node k
*/
         for(ktrack2= ktrack+1; ktrack2 < nnodes; ktrack2++)
         {
            if(belownode[ktrack2][tonode[ktrack][MAXOFF]] == 1)
            {
               belownode[ktrack2][ktrack] = 1;
            }
         }
      }   
      for(ktrack = nnodes-1; ktrack >=0; ktrack--)
      {  
         for(ktrack2 = 0; ktrack2 <= ktrack ; ktrack2++)
         {
            if(belownode[ktrack][ktrack2] == 1 )
            {
               if(ktrack2 < ntips)
               {
                  belownode[ktrack][nnodes]++;
               }
               belownode[nnodes][ktrack]++;
            }
         }
      }   

   
      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);

   }
   choosenode = consnode;
   while(choosenode == consnode)
   {
      choosenode = (int) ((nnodes-ntips-1)*rnd(seed)+ntips);
   }
   runlog = 0.0;
   chooseparent = tonode[choosenode][MAXOFF];
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
         thisparent = tonode[ktrack][MAXOFF];
         blen = time[thisparent] - time[ktrack];
         brlen[ktrack] = 0.5*(rate[thisparent]+rate[ktrack])*blen;         
         propbrlen[ktrack] = brlen[ktrack];
   }

   mincurratio = min2(time[chooseparent]/time[choosenode],
      hightime[choosenode]/time[choosenode]);
   maxcurratio = lowtime[choosenode]/time[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {
         maxcurratio = max2(maxcurratio, lowtime[ktrack]/time[ktrack]);
         mincurratio = min2(mincurratio, hightime[ktrack]/time[ktrack]);
      }                        
   }      
   for(ktrack =0 ; ktrack < ntips; ktrack++)
   {
      if( time[ktrack] > 0.0 &&  belownode[choosenode][ktrack] ==1)
      {
         parentnode = tonode[ktrack][MAXOFF];
         if( time[parentnode] * rate[ktrack] < time[ktrack]*(
            rate[ktrack]+rate[parentnode]) )
         {
            mincurratio = min2(mincurratio, (time[ktrack]*rate[parentnode])/
               (time[ktrack]*(rate[ktrack]+rate[parentnode]) -
                time[parentnode]*rate[ktrack]) );
         }
      }
   }

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   } 
   else
   {
      highnum = 0.5+ lmin*invothk;
   }
   if(maxcurratio == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {
      lmax = log(maxcurratio);
   }
   if(othk <  -2.0*lmax)
   {
      lownum = 0.0;
   }
   else
   {
      lownum = 0.5+ lmax*invothk;
   }
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;
    
   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(belownode[choosenode][ktrack] == 1)
      {
         if(ktrack < ntips && time[ktrack] > 0.0)
         {
            parentnode = tonode[ktrack][MAXOFF];
            newtime[ktrack] = time[ktrack];
            newrate[ktrack] = (time[parentnode] - time[ktrack])*(rate[
               parentnode] + rate[ktrack])/(newtime[parentnode] - time[ktrack])
               -newrate[parentnode];
            newlogr[ktrack] = log(newrate[ktrack]);
         }
         else
         {
            newrate[ktrack] = invfac*rate[ktrack];
            newlogr[ktrack] = loginvfac + logr[ktrack];
            newtime[ktrack] = flatefac*time[ktrack];
         } 
      }
      else
      {
         newlogr[ktrack] = logr[ktrack];
         newrate[ktrack] = rate[ktrack];
         newtime[ktrack] = time[ktrack];
      }
   }
   newrate[nnodes] = rate[nnodes];
   propbrlen[choosenode] = (newtime[chooseparent] -newtime[choosenode])*
      0.5*(newrate[chooseparent] + newrate[choosenode]);

   minnewratio = min2(newtime[chooseparent]/newtime[choosenode], 
      hightime[choosenode]/newtime[choosenode]);
   maxnewratio = lowtime[choosenode]/newtime[choosenode];
   for(ktrack = ntips; ktrack < choosenode; ktrack++)
   {
      if(belownode[choosenode][ktrack] ==1)
      {
         maxnewratio = max2(maxnewratio, lowtime[ktrack]/newtime[ktrack]);
         minnewratio = min2(minnewratio, hightime[ktrack]/newtime[ktrack]);
      }                        
   }      
   for(ktrack =0 ; ktrack < ntips; ktrack++)
   {
      if( belownode[choosenode][ktrack] ==1)
      {
         parentnode = tonode[ktrack][MAXOFF];
         if(time[ktrack] > 0.0)
         {
            runlog += log((time[parentnode] - time[ktrack])/
               (newtime[parentnode] - newtime[ktrack]));   
         }
         else
         {
            runlog += loginvfac;
         }
         if( newtime[ktrack] > 0.0 )
         {  
            parentnode = tonode[ktrack][MAXOFF];
            if( newtime[parentnode] * newrate[ktrack] < newtime[ktrack]*(
               newrate[ktrack]+newrate[parentnode]) )
            {
               minnewratio = min2(minnewratio, (newtime[ktrack]*newrate[
                  parentnode])/ (newtime[ktrack]*(newrate[ktrack]+newrate[
                  parentnode]) - newtime[parentnode]*newrate[ktrack]) );
            }
         }
      }   
   }   

   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {  
      highnum = 1.0;
   }  
   else  
   {  
      highnum = 0.5+ lmin*invothk;
   }  
   if(lmax == 0.0)
   {
      lmax = BIGNEGLOG;
   }
   else
   {
      lmax = log(maxnewratio);
   }
   if(othk > -2.0*lmax)
   {
      lownum = 0.5+ lmax*invothk;
   }
   else
   {
      lownum = 0.0;
   }
   ratio = (jterm /(highnum-lownum))*
      exp(runlog + 
      logcondmvn(propbrlen,choosenode) -logcondmvn(brlen,choosenode)) *
      probratio(newtime,newrate,newlogr,time, rate, logr); 
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack <= choosenode; ktrack++)
      {
         rate[ktrack] = newrate[ktrack];
         time[ktrack] = newtime[ktrack];
         logr[ktrack] = newlogr[ktrack];
      }
      sucpartcons++;
   }   
   else
   {
      failpartcons++;
   }
}

void timeflate(time,rate,logr)
double *time, *rate, *logr;
{
   double probratio();
   double rnd();
   double ratio, randnum, invfac, flatefac, holdthis ;
   double logmvn();
   double jterm, maxcurratio, maxnewratio;
   double lmin, lmax,lbound, ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   static double halfk, neghalfk, *thelowcon, invothk;
   static double *thehighcon;
   static double *newtime, *newrate, *newlogr;
   static double *brlen, *propbrlen;
   int ktrack, ktrack2, breakthis;
   int thisnode, parentnode;
   static int hereyet=0, numlowcons, *nodelowcon, numhighcons, *nodehighcon;
   if(hereyet ==0)
   { 
      brlen = dvec(nnodes);
      propbrlen = dvec(nnodes);
      thelowcon = dvec(nnodes);
      thehighcon = dvec(nnodes);
      nodelowcon = ivec(nnodes);
      nodehighcon = ivec(nnodes);
      hereyet = 1;
      numhighcons = 0;
      numlowcons = 0;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      for(ktrack = ntips; ktrack < nnodes; ktrack++)
      {
         breakthis = 0;
         for(ktrack2 = 1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thisnode = tonode[ktrack][ktrack2];
            if(lowtime[ktrack] == lowtime[thisnode] &&
               (thisnode >= ntips ||
               lowtime[thisnode] == 0.0) )
            {
               breakthis = 1;
               break;
            }
         }
         if( breakthis == 0)
         {
            numlowcons++;
            thelowcon[numlowcons] = lowtime[ktrack];
            nodelowcon[numlowcons] = ktrack;
         }
         if(ktrack == nnodes-1)
         {
            numhighcons++;
            thehighcon[numhighcons] = hightime[ktrack];
            nodehighcon[numhighcons] = ktrack;
         }
         else
         {
            if(hightime[ktrack] != hightime[tonode[ktrack][MAXOFF]])
            {
               numhighcons++;
               thehighcon[numhighcons] = hightime[ktrack];
               nodehighcon[numhighcons] = ktrack;
            }
         }
      }   
      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);
   }
   mincurratio = thehighcon[1]/time[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      mincurratio = min2( mincurratio,
         thehighcon[ktrack]/time[nodehighcon[ktrack]]);
   }

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons == 0)
   {
      lownum = 0.0;
      lmax = BIGNEGLOG;
   }
   else
   {   
      maxcurratio = thelowcon[1]/time[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxcurratio = max2(maxcurratio,
            thelowcon[ktrack]/time[nodelowcon[ktrack]]);
      }  
      if(maxcurratio == 0.0)
      {  
         lmax = BIGNEGLOG;
      }  
      else
      {   
         lmax = log(maxcurratio);
      }  
      if(othk < -2.0*lmax)
      {  
         lownum = 0.0;
      }  
      else
      {   
         lownum = 0.5+ lmax*invothk;
      }  
   }   
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(ktrack >= ntips)
      {
         newtime[ktrack] = flatefac*time[ktrack];
      }
      else
      {
         newtime[ktrack] = time[ktrack];
      }
      newrate[ktrack] = rate[ktrack];
      newlogr[ktrack] = logr[ktrack];
   }   

   newrate[nnodes] = rate[nnodes];
   minnewratio = thehighcon[1]/newtime[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      minnewratio = min2(minnewratio, thehighcon[ktrack]/
         newtime[nodehighcon[ktrack]]);
   }
   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons != 0)
   {
      maxnewratio = thelowcon[1]/newtime[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxnewratio = max2(maxnewratio, thelowcon[ktrack]/
            newtime[nodelowcon[ktrack]]);
      }  
      if(maxnewratio == 0.0)
      {  
         lmax = BIGNEGLOG;
      }  
      else
      {   
         lmax = log(maxnewratio);
      }  
      if(othk > -2.0*lmax)
      {  
         lownum = 0.5 + lmax*invothk;
      }  
      else
      {   
         lownum = 0.0;
      }  
   }   
   else
   {   
      lownum = 0.0;
   }
   for(ktrack=0;ktrack < nbranches; ktrack++)
   {
      parentnode = tonode[ktrack][MAXOFF];
      holdthis = 0.5*(rate[ktrack]+rate[parentnode]);
      brlen[ktrack] = holdthis*( time[parentnode] - time[ktrack]);
      propbrlen[ktrack] = holdthis* (newtime[parentnode] - newtime[ktrack]);
   }
   jterm /= (highnum - lownum);
   ratio = exp(-1.0*(nnodes-ntips)*loginvfac + logmvn(propbrlen) - 
      logmvn(brlen) ) * jterm
      * probratio(newtime,newrate,newlogr,time,rate, logr);
      
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {  
         time[ktrack] = newtime[ktrack];
      }  
      suctimeflate++;
   }
   else
   {   
      failtimeflate++;
   }
}

void bothflate(time,rate,logr)
double *time, *rate, *logr;
{
   double probratio();
   double rnd();
   double ratio, randnum, invfac, flatefac ;
   double logmvn();
   double jterm, maxcurratio, maxnewratio;
   double lmin, lmax,lbound, ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   static double halfk, neghalfk, *thelowcon, invothk;
   static double *thehighcon;
   static double *newtime, *newrate, *newlogr;
   static double *brlen, *propbrlen;
   int ktrack, ktrack2, breakthis;
   int thisnode, parentnode;
   static int hereyet=0, numlowcons, *nodelowcon, numhighcons, *nodehighcon;
   if(hereyet ==0)
   { 
      brlen = dvec(nnodes);
      propbrlen = dvec(nnodes);
      thelowcon = dvec(nnodes);
      thehighcon = dvec(nnodes);
      nodelowcon = ivec(nnodes);
      nodehighcon = ivec(nnodes);
      hereyet = 1;
      numhighcons = 0;
      numlowcons = 0;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      for(ktrack = ntips; ktrack < nnodes; ktrack++)
      {
         breakthis = 0;
         for(ktrack2 = 1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thisnode = tonode[ktrack][ktrack2];
            if(lowtime[ktrack] == lowtime[thisnode] &&
               (thisnode >= ntips ||
               lowtime[thisnode] == 0.0) )
            {
               breakthis = 1;
               break;
            }
         }
         if( breakthis == 0)
         {
            numlowcons++;
            thelowcon[numlowcons] = lowtime[ktrack];
            nodelowcon[numlowcons] = ktrack;
         }
         if(ktrack == nnodes-1)
         {
            numhighcons++;
            thehighcon[numhighcons] = hightime[ktrack];
            nodehighcon[numhighcons] = ktrack;
         }
         else
         {
            if(hightime[ktrack] != hightime[tonode[ktrack][MAXOFF]])
            {
               numhighcons++;
               thehighcon[numhighcons] = hightime[ktrack];
               nodehighcon[numhighcons] = ktrack;
            }
         }
      }   
      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);
   }
   mincurratio = thehighcon[1]/time[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      mincurratio = min2( mincurratio,
         thehighcon[ktrack]/time[nodehighcon[ktrack]]);
   }

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons == 0)
   {
      lownum = 0.0;
      lmax = BIGNEGLOG;
   }
   else
   {   
      maxcurratio = thelowcon[1]/time[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxcurratio = max2(maxcurratio,
            thelowcon[ktrack]/time[nodelowcon[ktrack]]);
      }  
      if(maxcurratio == 0.0)
      {  
         lmax = BIGNEGLOG;
      }  
      else
      {   
         lmax = log(maxcurratio);
      }  
      if(othk < -2.0*lmax)
      {  
         lownum = 0.0;
      }  
      else
      {   
         lownum = 0.5+ lmax*invothk;
      }  
   }   
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(ktrack >= ntips)
      {
         newtime[ktrack] = flatefac*time[ktrack];
      }
      else
      {
         newtime[ktrack] = time[ktrack];
      }
      newrate[ktrack] = invfac*rate[ktrack];
      newlogr[ktrack] = logr[ktrack] +loginvfac;
   }   

   newrate[nnodes] = rate[nnodes];
   minnewratio = thehighcon[1]/newtime[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      minnewratio = min2(minnewratio, thehighcon[ktrack]/
         newtime[nodehighcon[ktrack]]);
   }
   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   }
   else
   {   
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons != 0)
   {
      maxnewratio = thelowcon[1]/newtime[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxnewratio = max2(maxnewratio, thelowcon[ktrack]/
            newtime[nodelowcon[ktrack]]);
      }  
      if(maxnewratio == 0.0)
      {  
         lmax = BIGNEGLOG;
      }  
      else
      {   
         lmax = log(maxnewratio);
      }  
      if(othk > -2.0*lmax)
      {  
         lownum = 0.5 + lmax*invothk;
      }  
      else
      {   
         lownum = 0.0;
      }  
   }   
   else
   {   
      lownum = 0.0;
   }
   for(ktrack=0;ktrack < nbranches; ktrack++)
   {
      parentnode = tonode[ktrack][MAXOFF];
      brlen[ktrack] = 0.5*(rate[ktrack]+rate[parentnode])*(
         time[parentnode] - time[ktrack]);
      propbrlen[ktrack] = 0.5*(newrate[ktrack]+newrate[parentnode])*
         (newtime[parentnode] - newtime[ktrack]);
   }
   jterm /= (highnum - lownum);
   if(brownmean > SMALLNUM)
   {
      ratio = exp((ntips-1.0)*loginvfac + logmvn(propbrlen) - 
         logmvn(brlen) ) * jterm
         * probratio(newtime,newrate,newlogr,time,rate, logr);
   }
   else
   {
      ratio = exp(-1.0*(nnodes-ntips-1.0)*loginvfac + logmvn(propbrlen) -
         logmvn(brlen) ) * jterm
         * probratio(newtime,newrate,newlogr,time,rate, logr);
   }   
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {  
         time[ktrack] = newtime[ktrack];
         rate[ktrack] = newrate[ktrack];
         logr[ktrack] = newlogr[ktrack];
      }  
      sucboth++;
   }
   else
   {   
      failboth++;
   }
}

void rateflate(time,rate,logr)
double *time, *rate, *logr;
{
   double probratio();
   double rnd();
   double ratio, randnum, invfac, flatefac ;
   double logmvn();
   double loginvfac;
   static double halfk, invothk;
   static double *newtime, *newrate, *newlogr;
   static double *brlen, *propbrlen;
   int ktrack;
   int thisnode, parentnode;
   static int hereyet=0;
   if(hereyet ==0)
   { 
      brlen = dvec(nnodes);
      propbrlen = dvec(nnodes);
      hereyet = 1;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);
   }
   loginvfac = halfk - othk*rnd(seed);
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      newtime[ktrack] = time[ktrack];
      newrate[ktrack] = flatefac*rate[ktrack];
      newlogr[ktrack] = logr[ktrack]-loginvfac;
   }   

   newrate[nnodes] = rate[nnodes];
   for(ktrack=0;ktrack < nbranches; ktrack++)
   {
      parentnode = tonode[ktrack][MAXOFF];
      brlen[ktrack] = 0.5*(rate[ktrack]+rate[parentnode])*(
         time[parentnode] - time[ktrack]);
      propbrlen[ktrack] = 0.5*(newrate[ktrack]+newrate[parentnode])*
         (newtime[parentnode] - newtime[ktrack]);
   }
   if(brownmean > SMALLNUM)
   {
      ratio = exp(-1.0*(nnodes-1)*loginvfac + logmvn(propbrlen) - 
         logmvn(brlen) ) 
         * probratio(newtime,newrate,newlogr,time,rate, logr);
   }
   else
   {
      ratio = exp(-1.0*loginvfac + logmvn(propbrlen) -
         logmvn(brlen) )
         * probratio(newtime,newrate,newlogr,time,rate, logr);
   }   
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {  
         rate[ktrack] = newrate[ktrack];
         logr[ktrack] = newlogr[ktrack];
      }  
      sucrateflate++;
   }
   else
   {   
      failrateflate++;
   }
}



void consflate(time,rate,logr)
double *time, *rate, *logr;
{
   double probratio();
   double rnd();
   double ratio, runlog, randnum, invfac, flatefac ;
   double jterm, maxcurratio, maxnewratio;
   double lmin, lmax,lbound, ubound, loginvfac;
   double highnum, lownum,mincurratio, minnewratio;
   static double halfk, neghalfk, *thelowcon, invothk;
   static double *thehighcon;
   static double *newtime, *newrate, *newlogr;
   int ktrack, ktrack2, breakthis;
   int thisnode, parentnode;
   static int hereyet=0, numlowcons, *nodelowcon, numhighcons, *nodehighcon;
   if(hereyet ==0)
   {
      thelowcon = dvec(nnodes);
      thehighcon = dvec(nnodes);
      nodelowcon = ivec(nnodes);
      nodehighcon = ivec(nnodes);
      hereyet = 1;
      numhighcons = 0;
      numlowcons = 0;
      invothk = 1.0/othk;
      halfk = 0.5*othk;
      neghalfk = -1.0*halfk;
      for(ktrack = ntips; ktrack < nnodes; ktrack++)
      {
         breakthis = 0;
         for(ktrack2 = 1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thisnode = tonode[ktrack][ktrack2];
            if(lowtime[ktrack] == lowtime[thisnode] &&
               (thisnode >= ntips || 
               lowtime[thisnode] == 0.0) )
            {
               breakthis = 1;
               break;
            }
         }
         if( breakthis == 0) 
         {
            numlowcons++;
            thelowcon[numlowcons] = lowtime[ktrack];
            nodelowcon[numlowcons] = ktrack;
         }
         if(ktrack == nnodes-1)
         {
            numhighcons++;
            thehighcon[numhighcons] = hightime[ktrack];
            nodehighcon[numhighcons] = ktrack;
         }
         else
         {
            if(hightime[ktrack] != hightime[tonode[ktrack][MAXOFF]])
            {
               numhighcons++; 
               thehighcon[numhighcons] = hightime[ktrack];   
               nodehighcon[numhighcons] = ktrack; 
            }
         }
      }
      newtime = dvec(nnodes);
      newrate = dvec(nnodes);
      newlogr = dvec(nnodes);
   }
   runlog = 0.0;
   mincurratio = thehighcon[1]/time[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {
      mincurratio = min2( mincurratio,
         thehighcon[ktrack]/time[nodehighcon[ktrack]]);
   }
   for(ktrack =0 ; ktrack < ntips; ktrack++)
   {
      if(time[ktrack] > 0.0)
      {
         parentnode = tonode[ktrack][MAXOFF];
         if( time[parentnode] * rate[ktrack] < time[ktrack]*(
            rate[ktrack]+rate[parentnode]) )
         {
            mincurratio = min2(mincurratio, (time[ktrack]*rate[parentnode])/
               (time[ktrack]*(rate[ktrack]+rate[parentnode]) -
                time[parentnode]*rate[ktrack]) );
         }
      }
   }   

   lmin = log(mincurratio);
   if(othk < 2.0*lmin )
   {
      highnum = 1.0;
   } 
   else
   {
      highnum = 0.5+ lmin*invothk;
   }
   if(numlowcons == 0)
   {
      lownum = 0.0;
      lmax = BIGNEGLOG;
   }
   else
   {
      maxcurratio = thelowcon[1]/time[nodelowcon[1]];   
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {
         maxcurratio = max2(maxcurratio, 
            thelowcon[ktrack]/time[nodelowcon[ktrack]]);
      }
      if(maxcurratio == 0.0)
      {
         lmax = BIGNEGLOG;
      }
      else
      {
         lmax = log(maxcurratio);
      }
      if(othk < -2.0*lmax)
      {
         lownum = 0.0;
      }
      else
      {
         lownum = 0.5+ lmax*invothk;
      }
   }   
   jterm = highnum - lownum;
   lbound = max2(lmax,neghalfk);
   ubound = min2(lmin,halfk);
   loginvfac = (lbound-ubound)*rnd(seed) - lbound;
   invfac = exp(loginvfac);
   flatefac = 1.0/invfac;

   for(ktrack=nnodes-1; ktrack >= 0; ktrack--)
   {
      if(ktrack < ntips && time[ktrack] > 0.0)
      {
         parentnode = tonode[ktrack][MAXOFF];
         newtime[ktrack] = time[ktrack];
         newrate[ktrack] = (time[parentnode] - time[ktrack])*(rate[   
            parentnode] + rate[ktrack])/(newtime[parentnode] - time[ktrack]) 
            -newrate[parentnode]; 
         newlogr[ktrack] = log(newrate[ktrack]);
      } 
      else
      {  
         newrate[ktrack] = invfac*rate[ktrack];
         newlogr[ktrack] = loginvfac + logr[ktrack];
         newtime[ktrack] = flatefac*time[ktrack];
      }
   }

   newrate[nnodes] = rate[nnodes];
   minnewratio = thehighcon[1]/newtime[nodehighcon[1]];
   for(ktrack=2; ktrack <= numhighcons; ktrack++)
   {  
      minnewratio = min2(minnewratio, thehighcon[ktrack]/
         newtime[nodehighcon[ktrack]]);
   }  
   for(ktrack =0 ; ktrack < ntips; ktrack++)
   {
      parentnode = tonode[ktrack][MAXOFF];
      if(time[ktrack] > 0.0)
      {
         runlog += log((time[parentnode] - time[ktrack])/ 
            (newtime[parentnode] - newtime[ktrack]));
      }
      else
      {
         runlog += loginvfac;
      }
      if(newtime[ktrack] > 0.0)
      {
         if( newtime[parentnode] * newrate[ktrack] < newtime[ktrack]*(
            newrate[ktrack]+newrate[parentnode]) )
         {
            minnewratio = min2(minnewratio, (newtime[ktrack]*newrate[
               parentnode])/ (newtime[ktrack]*(newrate[ktrack]+ newrate[
               parentnode]) - newtime[parentnode]*newrate[ktrack]) );
         }
      }
   }   
   lmin = log(minnewratio);
   if(othk < 2.0*lmin )
   {  
      highnum = 1.0;
   }  
   else  
   {  
      highnum = 0.5+ lmin*invothk;
   }  
   if(numlowcons != 0)
   {
      maxnewratio = thelowcon[1]/newtime[nodelowcon[1]];
      for(ktrack = 2; ktrack <= numlowcons; ktrack++)
      {  
         maxnewratio = max2(maxnewratio, thelowcon[ktrack]/
            newtime[nodelowcon[ktrack]]);
      }  
      if(maxnewratio == 0.0)
      {
         lmax = BIGNEGLOG;
      }
      else
      {
         lmax = log(maxnewratio);
      }
      if(othk > -2.0*lmax)
      {
         lownum = 0.5 + lmax*invothk;
      }
      else
      {
         lownum = 0.0;
      }
   }
   else
   {
      lownum = 0.0;
   }
   jterm /= (highnum - lownum);
   ratio = exp(runlog-loginvfac) * jterm
      * probratio(newtime,newrate,newlogr,time,rate, logr);
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      for(ktrack=0; ktrack < nnodes; ktrack++)
      {
         rate[ktrack] = newrate[ktrack];
         time[ktrack] = newtime[ktrack];
         logr[ktrack] = newlogr[ktrack];
      }
      succons++;
   }   
   else
   {
      failcons++;
   }
}

void changebrown(time,rate, logr)
double *time, *rate, *logr;
{
   double probratio();
   double rnd();
   
   double ratio, randnum;
   static double  *newrate;
   int ktrack;
   static int hereyet=0;

   if(hereyet ==0)
   {
      hereyet = 1;
      newrate = dvec(nnodes);
   }
   for(ktrack=0; ktrack < nnodes; ktrack++)
   {
      newrate[ktrack] = rate[ktrack];
   }
   newrate[nnodes] = rate[nnodes]*exp(thek *(rnd(seed)-0.5));
   ratio = (newrate[nnodes]/rate[nnodes])*
   probratio(time,newrate,logr,time,rate,logr);
   randnum = rnd(seed);
   if(randnum < ratio)
   {
      rate[nnodes] = newrate[nnodes];
      sucrate[nnodes]++;
   }   
   else
   {
      failrate[nnodes]++;
   }
}


void timechange(time,rate,logr)
double *time,*rate,*logr;
{
   void freedvec();
   double *dvec();
   double logcondmvn();
   double rnd();
   double blen, randnum, ratio; 
   double proptime;
   double minposs,maxposs, maxtime, maxchild, numratio, denomratio, oldfrac;
   double newfrac, junk, oldwin,newwin;
   double denregfac, numregfac,rtratio, expratio, regratio;
   double lhold,lbound, ubound,newrdif, nowrdif, newvar, nowvar, newlen, nowlen;
   static double halfk, neghalfk, *brlen, *propbrlen, invthek;
   int thisbranch, thisparent, thischild, ktrack, ktrack2;
   static int hereyet = 0;


   if(hereyet ==0)
   {
      brlen = dvec(nbranches);
      propbrlen = dvec(nbranches);
      invthek = 1.0/thek;
      halfk = 0.5*thek;
      neghalfk = -1.0*halfk;
      hereyet=1;
   }
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      blen = time[thisparent] - time[ktrack];
      brlen[ktrack] = 0.5*(rate[thisparent]+rate[ktrack])*blen;
      propbrlen[ktrack] = brlen[ktrack];
   }
   for(ktrack = ntips; ktrack < nnodes; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      maxchild = time[tonode[ktrack][1]];
      for(ktrack2=2; ktrack2 <= tonode[ktrack][0]; ktrack2++)
      {
         maxchild = max2(maxchild, time[tonode[ktrack][ktrack2]]);
      }
      minposs = max2(maxchild, lowtime[ktrack]);
      if(ktrack < nnodes -1)
      {
         maxposs = min2(hightime[ktrack], time[thisparent]);
         maxtime = maxposs - minposs;
         oldfrac = (time[ktrack]-minposs)/maxtime;
         oldwin = min2(oldfrac,1.0-oldfrac);
         junk = rnd(seed);
         if(junk < newk)
         {
            newfrac = rnd(seed);
         }
         else
         {
            newfrac = oldfrac + (2.0*rnd(seed) - 1.0)*oldwin;
         }
         newwin = min2(newfrac, 1.0-newfrac);
         proptime = minposs + maxtime*newfrac;
         if((oldfrac > (newfrac-newwin)) && (oldfrac < (newfrac+newwin)))
         {
            numratio = newk + (1.0-newk )*0.5/newwin;
         }
         else
         {
            numratio = newk;
         }
         if((newfrac > (oldfrac-oldwin)) && (newfrac < (oldfrac+oldwin)))
         {
            denomratio = newk+ (1.0-newk)*0.5/oldwin;
         }
         else
         {
            denomratio = newk;
         }
         propbrlen[ktrack] = (time[thisparent] - proptime)*0.5*(
            rate[thisparent]+rate[ktrack]);
         rtratio = 1.0;
         if(brownmean > SMALLNUM)
         {
            newlen = time[thisparent] - proptime;
            nowlen = time[thisparent] - time[ktrack];
            newvar = newlen*rate[nnodes];
            nowvar = nowlen*rate[nnodes];
            if(nowvar > 0.0)
            {
               newrdif = logr[ktrack] - logr[thisparent] + 0.5*newvar;
               nowrdif = logr[ktrack] - logr[thisparent] + 0.5*nowvar;
               regratio = numratio/ denomratio;
               expratio = a[ktrack]*log(proptime/time[ktrack])+ b[ktrack]*
                  log((time[thisparent] - proptime)/(time[thisparent] - 
                  time[ktrack])) ;
               if( ktrack != consnode)
               {
                  expratio += 0.5*(nowrdif*nowrdif/nowvar - newrdif*newrdif/
                     newvar);
                  rtratio *= nowvar/newvar;
               }
            }
            else
            {
               rtratio = 0.0;
               expratio = 0.0;
               regratio = 0.0;
            }
         }
         else
         {
            regratio = numratio/ denomratio;
            expratio = a[ktrack]*log(proptime/time[ktrack])+ b[ktrack]*
               log((time[thisparent] - proptime)/(time[thisparent] -
               time[ktrack])) ;
         }
      }
      else 
      {
/* ktrack = nnodes -1 ... try to change root depth */  
         maxposs = hightime[nnodes-1];
         lbound = neghalfk;
         lhold = log((maxposs-minposs)/(time[ktrack] - minposs)); 
         ubound = min2(halfk,lhold);
         proptime = minposs+(time[ktrack]-minposs)*exp(lbound+
            (ubound-lbound)*rnd(seed) );
         denregfac = min2(1.0, invthek*lhold+0.5);
         numregfac = min2(1.0, invthek*log((maxposs-minposs)/(proptime -
            minposs)) + 0.5); 

         regratio = ((proptime - minposs)*numregfac)/(denregfac*(
            time[ktrack] - minposs));
         expratio = (timegamr - 1.0)*log(proptime/time[nnodes-1])
         + timegamlam*(time[nnodes-1] - proptime);
         rtratio = 1.0;
      }
      for(ktrack2=1; ktrack2 <=tonode[ktrack][0]; ktrack2++)
      {
         thischild = tonode[ktrack][ktrack2];
         newlen = proptime - time[thischild];
         nowlen = time[ktrack] - time[thischild];
         propbrlen[thischild] = newlen*0.5*(rate[thischild] + rate[ktrack]);
         if(thischild >= ntips)
         {
            expratio += a[thischild]*log(time[ktrack]/proptime) +
               b[thischild]*log(newlen*time[ktrack]/(nowlen*proptime));
            regratio *= time[ktrack]/proptime;
         }
         if(brownmean > SMALLNUM)
         {
            newvar = newlen*rate[nnodes];
            if(newvar > 0.0)
            {
               if(thischild != consnode)
               {
                  nowvar = nowlen*rate[nnodes];
                  newrdif = logr[thischild] - logr[ktrack] + 0.5*newvar;
                  nowrdif = logr[thischild] - logr[ktrack] + 0.5*nowvar;
                  expratio += 0.5*((nowrdif*nowrdif)/nowvar -(newrdif*newrdif)/
                     newvar);
                  rtratio *= nowvar/newvar;
               }
            }
            else
            {
               rtratio = 0.0;
               expratio = 0.0;
               regratio = 0.0;
            }  
         }
      }
      ratio = exp(logcondmvn(propbrlen,ktrack)-logcondmvn(brlen,ktrack) +
         expratio)*sqrt(rtratio)* regratio;
      randnum = rnd(seed);
      if(randnum < ratio)
      { 
         suctime[ktrack]++;
         time[ktrack] = proptime;
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thisbranch = tonode[ktrack][ktrack2];
            brlen[thisbranch] =propbrlen[thisbranch];
         }
         if(ktrack < nnodes-1)
         {
            brlen[ktrack] = propbrlen[ktrack];
         }
      }
      else
      {
         failtime[ktrack]++;
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++) 
         { 
            thisbranch = tonode[ktrack][ktrack2];
            propbrlen[thisbranch] = brlen[thisbranch];         
         } 
         if(ktrack < nnodes-1) 
         { 
            propbrlen[ktrack] = brlen[ktrack];
         } 

      }
   }
}
 
void ratechange(time,rate,logr)
double *time,*rate, *logr;
{
   void freedvec();
   double *dvec();
   double logcondmvn();
   double rnd();
   double thelen, var, holdlen;
   double rfac, logrfac, proprate, proplogr, randnum, ratio;
   static double *brlen, *propbrlen;
   double expratio, regratio, newrdif, nowrdif;
   int thischild, thisparent, ktrack, ktrack2;
   static int hereyet = 0;


   if(hereyet == 0)
   {
      brlen = dvec(nbranches);
      propbrlen = dvec(nbranches);
      hereyet = 1;
   }
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      thisparent = tonode[ktrack][MAXOFF];
      holdlen =  time[thisparent] - time[ktrack];
      brlen[ktrack] = holdlen*0.5*(rate[ktrack]+rate[thisparent]);
      propbrlen[ktrack] = brlen[ktrack];
   }
   for(ktrack = 0; ktrack < nnodes-1; ktrack++)
   {
      logrfac = thek*(rnd(seed)-0.5);
      rfac = exp(logrfac);
      proprate = rfac*rate[ktrack];
      proplogr = logr[ktrack]+logrfac;
      expratio = 0.0;
      regratio = 1.0;
      thisparent = tonode[ktrack][MAXOFF];
      thelen = time[thisparent] - time[ktrack];
      if(ktrack != consnode)
      {
         var = thelen*rate[nnodes];
         propbrlen[ktrack] = thelen*0.5*(rate[thisparent]+proprate);
         if(var > 0.0)
         {
            newrdif = proplogr - logr[thisparent] + 0.5*var;
            nowrdif = logr[ktrack] - logr[thisparent] + 0.5*var;
            regratio *= rate[ktrack]/ proprate;
            expratio += 0.5*(nowrdif*nowrdif - newrdif*newrdif)/var;
         }
         else
         {
            expratio = 0.0;
            regratio = 0.0;
         }
      }          
      else
      {
         propbrlen[ktrack] = thelen*proprate;
         expratio += (rategamr - 1.0)*logrfac+
            rategamlam*(rate[ktrack] - proprate);   
         for(ktrack2 = 1; ktrack2 <= tonode[nnodes-1][0]; ktrack2++)
         {
            thischild = tonode[nnodes-1][ktrack2];
            if(thischild != consnode)
            {
               thelen = time[nnodes-1] - time[thischild];
               var = thelen*rate[nnodes];
               if( var > 0.0)
               {
                  newrdif = logr[thischild] - proplogr + 0.5*var;
                  nowrdif = logr[thischild] - logr[ktrack] + 0.5*var;
                  propbrlen[thischild ] = 0.5*(proprate+rate[thischild])*thelen;
                  expratio += 0.5*(nowrdif*nowrdif - newrdif*newrdif)/var;
               }
               else
               {
                  expratio = 0.0;
                  regratio = 0.0;
               }              
            }
         }
      }
      if(ktrack >= ntips)
      {
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thischild = tonode[ktrack][ktrack2];
            thelen = time[ktrack] - time[thischild];
            var = thelen*rate[nnodes];
            if( var > 0.0)
            {
               newrdif = logr[thischild] - proplogr + 0.5*var;
               nowrdif = logr[thischild] - logr[ktrack] + 0.5*var;
               propbrlen[thischild ] = 0.5*(proprate+ rate[thischild])*thelen;
               expratio += 0.5*(nowrdif*nowrdif - newrdif*newrdif)/var;
            }
            else
            {
               expratio = 0.0;
               regratio = 0.0;
            }
         }
      }
      expratio += logcondmvn(propbrlen, ktrack) - logcondmvn(brlen, ktrack);
      ratio = regratio*exp(expratio)* rfac;
      randnum = rnd(seed);

      if(randnum < ratio)
      { 
         sucrate[ktrack]++;
         rate[ktrack] = proprate;
         logr[ktrack] = proplogr;
         brlen[ktrack] = propbrlen[ktrack];
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++)
         {
            thischild = tonode[ktrack][ktrack2];
            brlen[thischild] =propbrlen[thischild];
         }         
         if(ktrack == consnode)
         {
            rate[nnodes-1] = proprate;
            logr[nnodes-1] = proplogr;
            for(ktrack2 = 1; ktrack2 <= tonode[nnodes-1][0]; ktrack2++)
            {
               thischild = tonode[nnodes-1][ktrack2];
               brlen[thischild] = propbrlen[thischild];
            }
         }
      }
      else
      {
         failrate[ktrack]++;
         propbrlen[ktrack] = brlen[ktrack];
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++)  
         {         
            thischild = tonode[ktrack][ktrack2];
            propbrlen[thischild] =brlen[thischild];
         }        
         if(ktrack == consnode)
         {
            for(ktrack2=1; ktrack2 <= tonode[nnodes-1][0]; ktrack2++)
            {
               thischild = tonode[nnodes-1][ktrack2];
               propbrlen[thischild] = brlen[thischild];
            }
         }
      }
   }
}
 
void summarize(double **samptime, double **samprate,char*treefilename,char*nodefilename)
{
   int jtrack,ktrack, ktrack2;
   int lowval, highval;
   double thelowval, thehighval;
   double *sumsq,*sum, *sampmean, *sampmeansq, *sumtime, *sumsqtime;
   double holdrt, holdtemp;
   double *ntime, *nrate, holdparent, holdbrlen;
   double holdtime;
   double *oldtime,*newtime,*oldrate,*newrate,theratio;
   double *actualtime, *oldlogr, *newlogr;
   double **prtime, **hbrlen;
   double *dumvec;
   FILE *ratiofile, *treefile, *nodefile;

   void fprinttree();
   double probratio();
   void sort();

   lowval = ((int) floor(numsamps*LOWFRAC +0.5)) -1;
   highval = ((int) floor(numsamps*HIGHFRAC + 0.5))-1;
   printf("POSTERIOR INTERVAL from probability %9.5lf to %9.5lf\n",
      LOWFRAC, HIGHFRAC);
   printf("(i.e. %5d -th lowest value to %5d -th lowest value)\n",
      lowval+1, highval+1);
   printf("Consflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      succons, failcons, (1.0*succons)/ (succons+ failcons) ); 
   printf("Bothflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucboth, failboth, (1.0*sucboth)/ (sucboth+ failboth) );
   printf("Rateflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucrateflate, failrateflate, (1.0*sucrateflate)/ 
      (sucrateflate+ failrateflate) );
   printf("Timeflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      suctimeflate, failtimeflate, (1.0*suctimeflate)/ (suctimeflate+ 
      failtimeflate) );
   printf("Partconsflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucpartcons, failpartcons, (1.0*sucpartcons)/ (sucpartcons+
      failpartcons) );   
   printf("Partbothflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucpartboth, failpartboth, (1.0*sucpartboth)/ (sucpartboth+
      failpartboth) );   
   printf("Parttimeflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucparttime, failparttime, (1.0*sucparttime)/ (sucparttime+
      failparttime) );   
   printf("Partrateflate: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
      sucpartrate, failpartrate, (1.0*sucpartrate)/ (sucpartrate+
      failpartrate) );   
   if(brownmean > SMALLNUM)
   {
      for(ktrack=0; ktrack <= nnodes; ktrack++)
      {
         if(ktrack < nnodes)
         {
            printf("RATE %2d: ", ktrack);
         } 
         else 
         {
            printf("Brownian: ");
         }
         if(ktrack != nnodes-1)
         {
            printf("nsuc=%10ld nfail=%10ld psuc= %7.5lf\n", 
               sucrate[ktrack], failrate[ktrack], (1.0*sucrate[ktrack])/
               (sucrate[ktrack]+failrate[ktrack]) );
         }
         else
         {
            printf("ROOT rate constrained to equal rate of node %2d\n",
               consnode);
         } 
      }
   }

   for(ktrack=ntips; ktrack <= nnodes-1; ktrack++) 
   { 
      printf("TIME %2d: nsuc=%10ld nfail=%10ld psuc= %7.5lf\n",
         ktrack, suctime[ktrack], failtime[ktrack], (1.0*suctime[ktrack])/ 
         (suctime[ktrack]+failtime[ktrack]) ); 
   }

   putchar('\n');
   oldlogr = dvec(nnodes);
   newlogr = dvec(nnodes);
   oldtime = dvec(nnodes);
   newtime = dvec(nnodes);
   oldrate = dvec(nnodes);
   newrate = dvec(nnodes);

   nodefile = gfopen(nodefilename, "w");
   ratiofile = gfopen("ratiofile","w");
   for(ktrack=0; ktrack<= nnodes; ktrack++)
   {
      if(ktrack < nnodes)
      {
         oldtime[ktrack] = samptime[ktrack][0];
      }
      oldrate[ktrack] = samprate[ktrack][0];
      oldlogr[ktrack] = log(oldrate[ktrack]);
   }
   for(ktrack2=0; ktrack2 < numsamps; ktrack2++)
   {
      for(ktrack=0; ktrack<= nnodes; ktrack++) 
      { 
         if(ktrack < nnodes)
         {
            newtime[ktrack] = samptime[ktrack][ktrack2]; 
         }
         newrate[ktrack] = samprate[ktrack][ktrack2]; 
         newlogr[ktrack] = log(newrate[ktrack]);
      }
      theratio = probratio(newtime,newrate, newlogr,oldtime,oldrate,oldlogr);
      fprintf(ratiofile, "%12.6lf\n", theratio);
   } 
   fclose(ratiofile);
   freedvec(oldtime);
   freedvec(newtime);
   freedvec(oldrate);
   freedvec(newrate);
   freedvec(oldlogr);
   freedvec(newlogr);
   ntime = dvec(nnodes);
   nrate = dvec(nnodes);
   sumtime = dvec(nnodes);
   sumsqtime = dvec(nnodes);
   sumsq = dvec(nnodes+nnodes);
   sum = dvec(nnodes+nnodes);
   sampmean = dvec(nnodes);
   sampmeansq = dvec(nnodes);
   prtime = dmat(nnodes,numsamps);
   hbrlen = dmat(nnodes,numsamps);
   dumvec = dvec(numsamps);

/* CHANGE BELOW */
   for(ktrack=0; ktrack <= nnodes+nnodes; ktrack++)
   {
      sum[ktrack] = 0.0;
      sumsq[ktrack] = 0.0;
      if(ktrack < nnodes)
      {
         sumtime[ktrack] = 0.0;
         sumsqtime[ktrack] = 0.0;
      }

      if(ktrack < nnodes-1)
      {
         sampmean[ktrack] = 0.0;
         sampmeansq[ktrack] = 0.0;
      }
      for(ktrack2=0; ktrack2 < numsamps; ktrack2++)
      {
         if(ktrack < nnodes)
         {
            holdtime = samptime[ktrack][ktrack2];
            sumtime[ktrack] += holdtime;
            sumsqtime[ktrack] += holdtime*holdtime;
         }

         if(ktrack < nnodes-1)
         {
            holdparent = samptime[tonode[ktrack][MAXOFF]][ktrack2];
            holdrt = 0.5*(samprate[tonode[ktrack][MAXOFF]][ktrack2] +
               samprate[ktrack][ktrack2]);
            holdbrlen = holdrt*(holdparent-holdtime);
            hbrlen[ktrack][ktrack2] = holdbrlen;
            sampmean[ktrack] += holdbrlen;
            sampmeansq[ktrack] += holdbrlen*holdbrlen; 
            holdrt = samptime[nnodes-1][ktrack2];
            holdtemp =  holdtime/holdrt;
            prtime[ktrack][ktrack2] = holdtemp;
            sum[ktrack] += holdtemp;
            sumsq[ktrack] += holdtemp*holdtemp;
         }
         else
         {
            if(ktrack == nnodes -1)
            {
               holdrt = samptime[ktrack][ktrack2];
               sum[ktrack] += holdrt;
               sumsq[ktrack] += holdrt*holdrt; 
            }
            else
            {
               holdrt = samprate[ktrack-nnodes][ktrack2];
               sum[ktrack] += holdrt;
               sumsq[ktrack] += holdrt*holdrt; 
            } 
         }      
      }
      if(ktrack < nnodes-1)
      {
         sort(prtime[ktrack],numsamps,dumvec);
         thelowval = dumvec[lowval];
         thehighval = dumvec[highval];
         fprintf(nodefile,"p %2d", ktrack);
         printf("Prop. Node %2d to Root= ", ktrack);
      }
      else
      {
         if(ktrack==nnodes-1)
         {
/* DEBUG HAVE TO CREATE DUMMY VECTORS TO AVOID DESTRUCTION */
            sort(samptime[nnodes-1],numsamps,dumvec);
            thelowval = dumvec[lowval];
            thehighval = dumvec[highval];
            fprintf(nodefile, "p %2d", ktrack);
            printf("Root Depth = ");
         }
         else
         {
            sort(samprate[ktrack-nnodes],numsamps,dumvec);
            thelowval = dumvec[lowval];
            thehighval = dumvec[highval];
            fprintf(nodefile,"R %2d", ktrack-nnodes);
            if(ktrack < nnodes+nnodes)
               printf("Rate of Branch %2d = ", ktrack-nnodes);
            else
               printf("Brownian Constant = ");
         }
      }

if(sumsq[ktrack] - sum[ktrack]*sum[ktrack]/numsamps< 0.0)
{
   printf("prob in summary, can't take sqr. root of %18.12lf\n",
      (sumsq[ktrack] - (sum[ktrack]*sum[ktrack]/numsamps))/(numsamps-1.0));
}
      fprintf(nodefile, " %9.5lf %9.5lf %9.5lf %9.5lf\n", sum[ktrack]/numsamps
         , sqrt((sumsq[ktrack] - sum[ktrack]*sum[ktrack]/numsamps
         )/(numsamps-1.0)), thelowval, thehighval);
      printf(" %9.5lf (Std. Dev. = %9.5lf) (%9.5lf,%9.5lf)\n", 
sum[ktrack]/numsamps
         , sqrt((sumsq[ktrack] - sum[ktrack]*sum[ktrack]/numsamps
         )/(numsamps-1.0)), dumvec[lowval],dumvec[highval]);
   }
   putchar('\n');

   for(ktrack=0 ; ktrack < nnodes-1; ktrack++)
   { 
      sort(hbrlen[ktrack],numsamps,dumvec);     
      thelowval = dumvec[lowval];
      thehighval = dumvec[highval];
printf("br %3d samp mean= %8.5lf (%8.5lf) est. = %8.5lf (%9.5lf, %9.5lf)\n", 
         ktrack, sampmean[ktrack]/numsamps, sqrt((sampmeansq[ktrack] -
         sampmean[ktrack]*sampmean[ktrack]/numsamps)/(numsamps-1.0)), 
         mean[ktrack], dumvec[lowval], dumvec[highval]);
   }
   putchar('\n');
   for(ktrack = 0; ktrack < nnodes; ktrack++)
   {
      sort(samptime[ktrack],numsamps,dumvec);
      thelowval = dumvec[lowval];
      thehighval = dumvec[highval];
      fprintf(nodefile,"T %2d %9.5lf %9.5lf %9.5lf %9.5lf\n", ktrack, 
         sumtime[ktrack]/numsamps,  sqrt((sumsqtime[ktrack] - 
         sumtime[ktrack]* sumtime[ktrack]/numsamps)/(numsamps-1.0)),
         thelowval, thehighval);
      printf("Actual time node %2d = %9.5lf (Std. Dev. = ", 
         ktrack, sumtime[ktrack]/numsamps);
      if(sumsqtime[ktrack] - sumtime[ktrack]*sumtime[ktrack]/numsamps< 
         NEGSMALLNUM)
      {
         printf(" UNDEFINED)\n");
         printf("Can't take square root of ... %25.20lf\n", 
            sumsqtime[ktrack] - sumtime[ktrack]*sumtime[ktrack]/numsamps);
      }
      else
      {
         if(sumsqtime[ktrack] - sumtime[ktrack]*sumtime[ktrack]/numsamps<=
            0.0)
         {
            printf(" %9.5lf) (%9.5lf, %9.5lf)\n", 0.0, dumvec[lowval], 
               dumvec[highval]);
         }
         else
         {
            printf(" %9.5lf) (%9.5lf, %9.5lf)\n", 
               sqrt((sumsqtime[ktrack] - sumtime[ktrack]* sumtime[ktrack]/
               numsamps)/(numsamps-1.0)), dumvec[lowval], dumvec[highval]);
         }
      }
   }
   actualtime = dvec(nnodes);
   for(ktrack = 0; ktrack < nnodes-1; ktrack++)
   {
      actualtime[ktrack] = (sumtime[tonode[ktrack][MAXOFF]] - sumtime[ktrack])
         /numsamps;
   }
   treefile = gfopen(treefilename, "w");
   fprinttree(treefile,nnodes-1,1,nnodes-1,actualtime);
   fclose(treefile);
   fclose(nodefile);

   freedvec(dumvec);
   freedmat(hbrlen,nnodes,numsamps);
   freedmat(prtime,nnodes,numsamps);
   freedvec(actualtime);
   freedvec(sumtime);
   freedvec(sumsqtime);
   freedvec(nrate);
   freedvec(ntime);
   freedvec(sampmean);
   freedvec(sampmeansq);
   freedvec(sumsq);
   freedvec(sum);

}


 
void
fprintname(outfile,whichseq)
FILE *outfile;
int whichseq;
   /* prints the name of sequence number whichseq */
{
   int ttrack;
   char c;
 
   ttrack = 0;
   c = ' ';
   while (c != '\n')
   {
      c= name[whichseq][ttrack];
      if( c != '\n') fprintf(outfile,"%1c",c);
      ttrack++;
   }
}

/* print the tree */
void fprinttree(outfile,whichnode,whichoption,highnodenum,param)
FILE *outfile;
int whichnode, whichoption,highnodenum;
double *param;
/* if whichoption=1 then print sequence names else print sequence
numbers */
{
   int ktrack;

   if(tonode[whichnode][0] ==0)
   {
      if(whichoption==1)
      {
         fprintname(outfile,whichnode);
      }
      else
      {   
         if(whichoption==0)
         {
            fprintf(outfile,"%d", whichnode);
         }
      }   
   }   
   else
   {   
      for(ktrack=1; ktrack <= tonode[whichnode][0]; ktrack++)
      {
         if(ktrack==1)
         {
            fprintf(outfile,"(");
         }

         else
         {   
            fprintf(outfile,",");
         }

         fprinttree(outfile,tonode[whichnode][ktrack],whichoption,
            highnodenum, param);
      }
      fprintf(outfile, ")");
   }
   if(whichnode==highnodenum)
   {
      fprintf(outfile, ";\n");
   }
   else
   {   
      fprintf(outfile, ":%9.5f",param[whichnode]);
   }
}



void
printname(whichseq)
int whichseq;
   /* prints the name of sequence number whichseq */
{
   int ttrack;
   char c;

   ttrack = 0;
   c = ' ';

   while (c != '\n')
   {
      c= name[whichseq][ttrack];
      if( c != '\n') printf("%1c",c);
      ttrack++;
   }
}

/*  This routine calculates proposed state probability divided by current
state probability except for the fact the branch length terms are ignored.
It is assumed that ratio of proposal densitites is handled by routine
that calls this routine */
double probratio(proptime, proprate, proplogr, oldtime, oldrate, oldlogr)
double *proptime, *proprate, *proplogr, *oldtime, *oldrate, *oldlogr;
{

   double oldvar,newvar;
   double oldprop, newprop;
   double theratio, oldlength, newlength, oldrdif, newrdif;
   double rtratio, regratio, expratio;
   int panode, ktrack;


   expratio = 0.0;
   rtratio = 1.0;
   regratio = 1.0;
   if(brownmean > SMALLNUM)
   {
      for(ktrack = 0; ktrack < nbranches; ktrack++)
      {
         if(ktrack != consnode)
         {
            panode = tonode[ktrack][MAXOFF];
            oldlength = oldtime[panode] - oldtime[ktrack];
            oldvar = oldlength*oldrate[nnodes];
            oldrdif =  oldlogr[ktrack] - oldlogr[panode]+0.5*oldvar;
            newlength = proptime[panode] - proptime[ktrack];
            newvar =  newlength*proprate[ nnodes];
            newrdif =  proplogr[ktrack] - proplogr[panode]+0.5*newvar;
            if(newvar > 0.0)
            {
               rtratio *= oldvar/newvar;
               expratio += (newrdif*newrdif/newvar)
                  - (oldrdif*oldrdif/oldvar);
               regratio *= oldrate[ktrack]/proprate[ktrack];
            }
            else
            {
               rtratio = 0.0;
               expratio = 0.0;
               regratio = 0.0;
            }
         }
      }
      expratio *= (-0.5);
   }
   expratio += (timegamr - 1.0)*log(proptime[nnodes-1]/oldtime[nnodes-1])
      + timegamlam*(oldtime[nnodes-1] - proptime[nnodes-1]) +
      (rategamr - 1.0)*log(proprate[nnodes-1]/oldrate[nnodes-1]) +
      rategamlam*(oldrate[nnodes-1] - proprate[nnodes-1]); 
   if(brownmean > SMALLNUM)
   {
      expratio +=  browngamlam*(oldrate[nnodes] - proprate[nnodes]) +
         (browngamr - 1.0)*log(proprate[nnodes]/oldrate[nnodes]);
   }
   for(ktrack = ntips; ktrack < nnodes-1; ktrack++)
   {
         panode = tonode[ktrack][MAXOFF];
         newprop = proptime[ktrack]/proptime[panode];
         oldprop = oldtime[ktrack]/oldtime[panode];
         expratio += a[ktrack]*log(newprop/oldprop) + b[ktrack]*
            log((1.0-newprop)/(1.0-oldprop));
         regratio *= oldtime[panode]/proptime[panode];
   }
   theratio = sqrt(rtratio)*exp(expratio)*regratio;
   return(theratio);
}

void rdinfo(treefile)
FILE *treefile;
{
   int ktrack;
   double *brlen;
   char thischar;
 
   void docomma();
   void doleftpar();

   mean = dvec(MAXNODE);
   name = cmat(MAXNODE,NAMELEN);
   tonode = imat(MAXNODE,MAXOFF);


   for(ktrack=0;ktrack < MAXNODE; ktrack++)
   {
      tonode[ktrack][0]=0;
   }
   tonode[0][MAXOFF]= -1;


   currentnode=0;
   ntips =1;
   nnodes=1;
   ktrack=0;
   thischar= ' ';
printf("Before While\n");
   while(thischar != ';' )
   {
      thischar=getc(treefile);
      while(thischar == '\n')
      {
        thischar=getc(treefile);
      }
      switch(thischar)
      {
         case ')':
            currentnode = tonode[currentnode][MAXOFF];
            ktrack=0;
            break;
         case '(':
            doleftpar();
            ktrack=0;
            break;
         case ',':
            docomma();
            ktrack=0;
            break;
         case ':':
            fscanf(treefile,"%lf",&mean[currentnode]);
/* Do we need to do something somewhere to make sure correct number
of changes is assigned to each branch as tree grows? */

            ktrack=0; /* docolon() */
            break;
         case ';':
            ktrack=0;
            break;
         case ' ':
            break;
         default:
            name[currentnode][ktrack] = thischar;
            ktrack++;
            name[currentnode][ktrack] = '\n';
            break;
      }
   }   

   nbranches = nnodes-1;
   brlen = dvec(nbranches);
   for(ktrack=0;ktrack < nbranches; ktrack++)
   {
      brlen[ktrack] = mean[ktrack];
   }
   freedvec(mean);
   mean = dvec(nbranches);
   for(ktrack=0;ktrack < nbranches; ktrack++)
   {
      mean[ktrack] = brlen[ktrack];
   }
   freedvec(brlen);

}

void docomma()
{
   int ktrack, ktrack2;

   currentnode = tonode[currentnode][MAXOFF];
   nnodes++;
   ntips++;
   for(ktrack=nnodes-2;ktrack >= ntips-1; ktrack--)
   {
      if(ktrack== nnodes-2)
      {
         tonode[nnodes-1][MAXOFF] = -1;
      }

      else
      {   
         mean[ktrack+1] = mean[ktrack];
         tonode[ktrack+1][MAXOFF] = tonode[ktrack][MAXOFF]+1;
      }
      tonode[ktrack+1][0] = tonode[ktrack][0];
      for(ktrack2=1; ktrack2 <= tonode[ktrack][0];ktrack2++)
      {
         if(tonode[ktrack][ktrack2] >= ntips-1)
         {
            tonode[ktrack+1][ktrack2] = tonode[ktrack][ktrack2] +1;
         }
         else
         {   
            tonode[ktrack+1][ktrack2] = tonode[ktrack][ktrack2];
         }
      }   
   }   
   tonode[ntips-1][MAXOFF] = currentnode;
   tonode[ntips-1][0]=0;
   tonode[currentnode+1][0]++;
   tonode[currentnode+1][tonode[currentnode+1][0]] = ntips-1;
   currentnode = ntips-1;
   for(ktrack=ntips-1; ktrack >=0;ktrack--)
   {
      tonode[ktrack][MAXOFF]++;
   }
}

 
void doleftpar()
{
   int ktrack, ktrack2;

   nnodes++;
   for(ktrack=nnodes-2;ktrack >= ntips; ktrack--)
   {
      if(ktrack== nnodes-2)
      {
         tonode[nnodes-1][MAXOFF] = -1;
      }
      else
      {   
         mean[ktrack+1] = mean[ktrack];
         tonode[ktrack+1][MAXOFF] = tonode[ktrack][MAXOFF]+1;
      }
      tonode[ktrack+1][0] = tonode[ktrack][0];
      for(ktrack2=1;ktrack2 <= tonode[ktrack][0];ktrack2++)
      {
         if(tonode[ktrack][ktrack2] >= ntips)
         {
            tonode[ktrack+1][ktrack2] = tonode[ktrack][ktrack2] +1;
         }
         else
         {   
            if(tonode[ktrack][ktrack2] == currentnode)
            {
               tonode[ktrack+1][ktrack2] = ntips;
            }
            else
            {   
               tonode[ktrack+1][ktrack2] = tonode[ktrack][ktrack2] ;
            }
         }   
      }   
   }   
   if( tonode[currentnode][MAXOFF] == -1)
   {
      tonode[ntips][MAXOFF] = -1;
   }
   else
   {   
      tonode[ntips][MAXOFF] = tonode[currentnode][MAXOFF] + 1;
   }
   tonode[ntips][0]=1;
   tonode[ntips][1]=currentnode;
   for(ktrack=ntips-1; ktrack >=0;ktrack--)
   {
      if(ktrack == currentnode)
      {
         tonode[currentnode][MAXOFF] = ntips;
      }

      else
      {   
         tonode[ktrack][MAXOFF]++;
      }
   }   
}

double snorm()
{
   double rnd();

   static int onemore=0;
   static double rankeep;
   double u1,u2,x1,x2;

   if(onemore==0)
   {
      u1=rnd(seed);
      u2=rnd(seed);
      x1= sqrt(-2.0*log(u1))*cos(2.0*PI*u2);
      x2= sqrt(-2.0*log(u1))*sin(2.0*PI*u2);
      rankeep=x2;
   }
   else
   {
      x1=rankeep;
   }
   onemore= 1-onemore;
   return(x1);
}

double exptial(explambda)
double explambda;
{
   double expvalue;
   extern double rnd();
   expvalue = -(1/ explambda) * log( 1 - rnd(seed));
   return(expvalue);
}

  
/* From Joe Felsenstein's PHYLIP package ...
   check for "needed by rnd" for other lines of code this routine needs */
double rnd(seed)
long *seed;
{
  /* random number generator -- slow but machine independent */
  long i, j, k, sum;
  longer mult, newseed;
  double x;
  mult[0] = 13;
  mult[1] = 24;
  mult[2] = 22;
  mult[3] = 6;
  for (i = 0; i <= 5; i++)
    newseed[i] = 0;
  for (i = 0; i <= 5; i++) {
    sum = newseed[i];
    k = i;
    if (i > 3)
      k = 3;
    for (j = 0; j <= k; j++)
      sum += mult[j] * seed[i - j];
    newseed[i] = sum;
    for (j = i; j <= 4; j++) {
      newseed[j + 1] += newseed[j] / 64;
      newseed[j] &= 63;
    }
  }  
  memcpy(seed, newseed, sizeof(longer));
  seed[5] &= 3;
  x = 0.0;
  for (i = 0; i <= 5; i++)
    x = x / 64.0 + seed[i];
  x /= 4.0;
  return x;
}   /* rnd */




/* copies a vector of doubles named vec of length veclen 
   (running from vec[0] to vec[veclen-1]) into 
   orderedvec and then sorts orderedvec from lowest to highest. 
   -- notice that the vector vec DOES NOT
   run from vec[1]to vec[veclen] */
void sort(vec,veclen,orderedvec)
double *vec,*orderedvec;
int veclen;
{
   int ktrack, ktrack2;
   double dummy;
   for(ktrack=0; ktrack < veclen; ktrack++)
   {
      orderedvec[ktrack] = vec[ktrack];
   }
   for(ktrack=1; ktrack < veclen; ktrack++)
   {
      for(ktrack2=ktrack; ktrack2>=1; ktrack2--)
      {
         if(orderedvec[ktrack2] < orderedvec[ktrack2-1])
         {
            dummy = orderedvec[ktrack2];
            orderedvec[ktrack2] = orderedvec[ktrack2-1];
            orderedvec[ktrack2-1] = dummy;
         }
         else
         {
            break;
         }
      }
   }   
}


FILE *gfopen(char *filename, char *mode)
{
   FILE *fp=(FILE*)fopen(filename, mode);
   if(fp==NULL) {
      printf("\nerror when opening file %s\n", filename);
      exit(-1);
   }
   return(fp);
}
