/* (c) Copyright 1999,2000 by Jeff Thorne
 
As this program is very crude, please do not distribute it. Instead,
people can obtain hopefully improved versions by contacting me (Jeff Thorne)
(Also, please don't delete this message) 

this file is part of a program for inferring branch lengths of
an evolutionary tree from DNA or amino acid sequences.  For
the most complicated protein model, the inference is based 
on a model that considers secondary structure and solvent accessibility.

The file hmmcntrl.dat allows you to specify the topology, the
model desired and various other things.

Aligned sequences are to be placed in a file named testseq.  The first
line of this file specifies the number of sequences and then the
length of the sequences.  Gaps and unknown nucleotides or amino acids are
represented by the "?" character.  Each sequence is terminated by the
"*" character.  Sequence specification begins with the ">" character
on a new line.  Any information that you want to specify particular
to a sequence should be follow the ">" character and be on the
same line as the ">" character.  The line after the line with ">"
should contain the name of the sequence -- it might be a good idea to
keep the name to 10 or fewer characters.

To compile for protein data in a Unix environment type...

gcc -DDNA estbranches.c -DHESSIAN arrayutl.c inv3.c -lm -O -o estbranches 
cc -o estbranches -fast estbranches.c -DHESSIAN arrayutl.c inv3.c -lm

To compile for DNA data in a Unix environment type...

gcc -DDNA estbranches.c -DHESSIAN arrayutl.c inv3.c -lm -O -o estbranches 

(if you have access to the "cc" compiler instead of the "gcc" compiler,
 then you "cc" instead of "gcc" above)

 MS VC++ 6.0: 
    cl -FeestbOld.exe -Ot -O2 -W3 -DDNA -DHESSIAN estbranches.c arrayutl.c inv3.c

Conditional compilation signals:

DNA - analyze DNA sequences
HESSIAN - calculate HESSIAN matrix and output it
*/

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "arrayutl.h"
#define STOPTHRESH  0.0001  /* Stop likelihood maximization when
                             likelihood can no longer be improved by at least
                             STOPTHRESH in a cycle */
#define INCFRAC 0.01     /* what proportion of parameter value to be
                            used when calculating approx. slope */
#define STEPFRAC 0.1     /* multiply slope by this if 2nd deriv  > 0 */
#define NAMELEN 53  /* length of sequence name */  
#define MAXNODE 200 /* Maximum number of nodes on a tree */ 
#define MAXOFF 10 /* Maximum number of offspring nodes of a node */

#ifndef DNA
#define NUMLETTERS 20  
#define DEFLATE 100.0
#else
#define NUMLETTERS 4
#define DEFLATE 1.0
#endif

FILE *gfopen(char *filename, char *mode);

FILE *modelinf;  /* This is the file that specifies the hmm model */
int **seq; /* contains the sequence information */
int currentnode,nnodes;
int *mark2rep;
int *sumcat, nsumcats; /* respectively represent
                          which categories to lump together when
                          reporting prediction probabilities and how
                          many categories remain after lumping */
int repcats, nseqs,nnodes,nbranches, ncats; /* respectively, #replcement
       categories -1, #sequences, #nodes, #branches, #hmm categories - 1 */
#ifndef DNA
char **revps, **evalfl, **evecfl, **ievecfl; /* names of files with
                                  information about replacement categories */
#endif
char **name; /* names of sequences */
char *model; /* specifies the replacement model to be used  */
char outlevel; /* How much output to screen? Q=Quiet (this option
                  should be used with parametric bootstrap), L=Loud=Lots of 
                  output (this is the default option) */
char predstrct; /* P=Predict Structure, D= Do not Predict Structure
                   (D is the default option) */
char usertree; /* N=Names of sequences specified in user tree,
                  0=numbers are in user tree and they correspond
                  to the order of the sequence in the sequence file */
#ifdef DNA
int group[NUMLETTERS+1];
double ratecons, **gfrq;
#endif

main()
{
   void printtree(); /* Prints the tree */
   void maxandprdct();  /* The guts, maximizes likelihood and predicts */
   void rdmodelinf(); /* reads HMM information */
   int **rdinfo(); /* read sequence and tree topology information, etc. */
   void beloud(); /* prints various info. about the run for verbose mode */

   extern double **dmat();
   extern int **imat();
   extern double *dvec();
   extern char **cmat();
   extern void freeivec();
   extern void freedmat();
   extern void freeimat();
   extern void freedvec();
   extern void freecmat();
   extern void freecvec();
   extern char *cvec();

   int ktrack;
   int **tonode;  /* contains the tree topology information:
                        tonode[i][0] contains the number of nodes with
                           an index less than i that are connected to node i
                           via a branch.
                        tonode[i][j] for j>0 contains a value corresponding
                           to one of the tonode[i][0] nodes that node i is
                           connected to via a branch */
   double *param; /* contains values of the model parameters -
                     param[0] to param[nbranches-1] contain branch length info 
                  */
   double *structfrq, **fromto; /* expectation of frequencies of hmm
                                   categories at position 1 of sequence */
   tonode = rdinfo();
   modelinf = gfopen(model,"r");
   while( getc(modelinf) != '\n');
   while( getc(modelinf) != '\n');
   fscanf(modelinf,"%d", &ncats);
   fromto = dmat(ncats,ncats);
   structfrq = dvec(ncats);
   while( getc(modelinf) != '\n');
   fscanf(modelinf,"%d", &repcats);
#ifndef DNA
   rdmodelinf(structfrq,fromto);
#endif
   param = dvec(nbranches-1);
   /* Initial branch length guesses */
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
#ifndef DNA
      param[ktrack] = 10.0;
#else
      param[ktrack] = 0.1;
#endif
/* USE BELOW For infinite Branches
      param[ktrack] = 100000.0;
*/
   }

   if(outlevel != 'Q')
   {
      beloud(tonode);
   }

   maxandprdct(tonode,param,structfrq,fromto);

   if(outlevel != 'Q')
   {
      printf("\nTree with sequence names\n");
      printtree(tonode,nnodes-1,1,nnodes-1,param);
/*
      printf("\nTree with Sequence Number IDs\n");
      printtree(tonode,nnodes-1,0,nnodes-1,param);
*/
   }

   freecvec(model);
#ifndef DNA
   freecmat(revps,repcats,NUMLETTERS);
   freecmat(evalfl,repcats,NUMLETTERS);
   freecmat(evecfl,repcats,NUMLETTERS);
   freecmat(ievecfl,repcats,NUMLETTERS);
#else
   freedmat(gfrq,ncats,NUMLETTERS);   
#endif
   freedmat(fromto,ncats,ncats);
   freedvec(structfrq,ncats);
   freedvec(param,nbranches-1);
   freeivec(sumcat,ncats);
   freeimat(tonode,nnodes-1,nnodes-1);
   freecmat(name,MAXNODE, NAMELEN);
   freeimat(seq,nseqs,seq[0][0]);
   freeivec(mark2rep,ncats);
}

void jprinttree(tonode,whichnode,whichoption,highnodenum,param, imask)
int **tonode, whichnode, whichoption,highnodenum, *imask;
double *param;
/* if whichoption=1 then print sequence names else print sequence
numbers */
{
   void printname(); /* prints selected sequence name */
   int ktrack;
   if(tonode[whichnode][0] ==0)
   {
      if(whichoption==1)
      {
         printname(whichnode);
      }
      else
      {
         if(whichoption==0)
         {
            printf("%d", imask[whichnode]);
         }
      }
   }
   else
   {
      for(ktrack=1; ktrack <= tonode[whichnode][0]; ktrack++)
      {
         if(ktrack==1)
         {
            printf("(");
         }
         else
         {
            printf(",");
         }
         jprinttree(tonode,tonode[whichnode][ktrack],whichoption,highnodenum,
           param, imask); 
      }
      printf(")");
   }
   if(imask[whichnode]==highnodenum)
   {
      printf(";\n");
   }
   else
   {
      printf(":%14.10lf",param[whichnode]);
   }
}


void printtree(tonode,whichnode,whichoption,highnodenum,param)
int **tonode, whichnode, whichoption,highnodenum;
double *param;
/* if whichoption=1 then print sequence names else print sequence
numbers */
{
   void printname(); /* prints selected sequence name */
   int ktrack;
   if(tonode[whichnode][0] ==0)
   {
      if(whichoption==1)
      {
         printname(whichnode);
      }
      else
      {
         if(whichoption==0)
         {
            printf("%d", whichnode);
         }
      }
   }
   else
   {
      for(ktrack=1; ktrack <= tonode[whichnode][0]; ktrack++)
      {
         if(ktrack==1)
         {
            printf("(");
         }
         else
         {
            printf(",");
         }
         printtree(tonode,tonode[whichnode][ktrack],whichoption,highnodenum,
           param); 
      }
      printf(")");
   }
   if(whichnode==highnodenum)
   {
      printf(";\n");
   }
   else
   {
      printf(":%9.5lf",param[whichnode]);
   }
}

void beloud(tonode)
int **tonode;
{
   int ktrack, ktrack2;
   void printname(); /* prints selected sequence name */

   printf("Number of Sequences %2d: length %5d: Number of Branches %2d\n",
           nseqs, seq[0][0], nbranches);
   printf("Number of Nodes %2d\n", nnodes );
   for(ktrack = nseqs;ktrack < nnodes; ktrack++)
   {
      for(ktrack2 = 1;ktrack2 <= tonode[ktrack][0];ktrack2++)
      {
         printf("Branch %2d connects node %2d ", tonode[ktrack][ktrack2],
            tonode[ktrack][ktrack2]);
         if(tonode[ktrack][ktrack2] < nseqs)
         {
            printf("(");
            printname(tonode[ktrack][ktrack2]);
            printf(") ");
         }
         printf("and %2d\n",ktrack);
      }
   }   
   printf("STOPTHRESH= %10.8lf INCFRAC= %10.8lf STEPFRAC = %10.8lf\n",
      STOPTHRESH, INCFRAC, STEPFRAC);
   printf("MAXNODE= %3d NAMELEN= %3d NUMLETTERS= %2d DEFLATE= %10.8lf\n",
      MAXNODE, NAMELEN, NUMLETTERS, DEFLATE);
   printf("MAXOFF= %3d\n", MAXOFF);
   printf("number of hmm categories is %2d\n", ncats);
   printf("number of distinct replacement process categories is %2d\n",
      repcats);
   if(predstrct=='P')
   {
      printf("Underlying Categories WILL be predicted at each site\n");
   }
   else 
   {
      printf("Underlying Categories will NOT be predicted at each site\n");
   }
}

#ifndef DNA
void rdmodelinf(structfrq,fromto)
double *structfrq, **fromto;
#else
void rdmodelinf(structfrq,fromto,bfrq, lneval)
double *structfrq, **fromto, **bfrq, **lneval;
#endif
/* reads the HMM information */
{

   extern int **imat();
   extern int *ivec();
   extern void freeimat();
   extern double *dvec();
   extern void freedvec();
#ifndef DNA
   extern char **cmat();
   extern char *cvec();
   extern void freecmat();
   extern void freecvec();

   char *string1,*string2,*string3,*string4;
#endif
   int markovcat, therepcat, ktrack,ktrack2,ktrack3;
   int ktrack4, ktrack5;

#ifndef DNA
   string1 = cvec(20);
   string2 = cvec(20);
   string3 = cvec(20);
   string4 = cvec(20);

   revps = cmat(repcats,20);
   evalfl = cmat(repcats,20);
   evecfl = cmat(repcats,20);
   ievecfl = cmat(repcats,20);
#endif

   while( getc(modelinf) != '\n');
   while( getc(modelinf) != '\n');
   while( getc(modelinf) != '\n');

   mark2rep = ivec(ncats);

   for(ktrack=0; ktrack <= ncats; ktrack++)
   {
      fscanf(modelinf,"%d%d\n", &markovcat, &therepcat); 
      mark2rep[markovcat] = therepcat;
   }
#ifndef DNA
   for(ktrack=0; ktrack <= repcats; ktrack++)
   {
      fscanf(modelinf,"%d\n%s\n%s\n%s\n%s\n", 
         &ktrack2, string1,string2,string3,string4);

      strcpy(revps[ktrack],string1);
      strcpy(evalfl[ktrack],string2);
      strcpy(evecfl[ktrack],string3);
      strcpy(ievecfl[ktrack],string4);
   }
#endif
   while( getc(modelinf) != '\n');
   for(ktrack=0;ktrack<=ncats;ktrack++)
   {
      for(ktrack2=0; ktrack2 <= ncats;ktrack2++)
      {
         fscanf(modelinf,"%lf", &fromto[ktrack][ktrack2]);
      }
   }
   while( getc(modelinf) != '\n');
   while( getc(modelinf) != '\n');
   for(ktrack=0;ktrack<=ncats;ktrack++)
   {
      fscanf(modelinf,"%lf",&structfrq[ktrack]);
   }
   while( getc(modelinf) != '\n');
   while( getc(modelinf) != '\n');
   while( getc(modelinf) != '\n');
   fscanf(modelinf,"%d",&nsumcats);
   if(outlevel != 'Q')
   {
      printf("nsumcats= %3d\n", nsumcats);
   }
   sumcat = ivec(ncats);
   for(ktrack=0;ktrack<=ncats;ktrack++)
   {
      sumcat[ktrack] = -1;
   }

   for(ktrack=0;ktrack<nsumcats;ktrack++)
   {
      fscanf(modelinf,"%d", &ktrack3);
      fscanf(modelinf,"%d", &ktrack4);
      for(ktrack2=0;ktrack2<ktrack4; ktrack2++)
      {
         fscanf(modelinf,"%d", &ktrack5);
/* DEBUG
         printf("ktr %2d ktr2 %2d ktr3 %2d ktr4 %2d ktr5 %2d\n", 
            ktrack, ktrack2, ktrack3, ktrack4, ktrack5);
*/
         sumcat[ktrack5] = ktrack3;
      }
   }
 
   for(ktrack=0;ktrack<=ncats;ktrack++)
   {
      if(sumcat[ktrack] < 0)
      {
         printf("Problem in modelinf.dat: How should HMM\n");
         printf("category %d be summarized in prediction output?\n", ktrack);
      } 
   }
#ifdef DNA
   while( getc(modelinf) != '\n');
   while( getc(modelinf) != '\n');
   while( getc(modelinf) != '\n');
   for(ktrack=0; ktrack <= ncats; ktrack++)
   {
      fscanf(modelinf, "%d", &ktrack2);
      for(ktrack3=0; ktrack3 < NUMLETTERS; ktrack3++)
      {
         fscanf(modelinf, "%lf", &bfrq[ktrack][ktrack3]);
      }
      bfrq[ktrack][NUMLETTERS]=1.0; /* This is the "Unknown"
                                           nucleotide type */

      fscanf(modelinf, "%lf%lf\n", &lneval[ktrack][0], &lneval[ktrack][1]);
   }
#endif

   fclose(modelinf);
#ifndef DNA
   freecvec(string1);
   freecvec(string2);
   freecvec(string3);
   freecvec(string4);
#endif
}    

int **rdinfo()
/* reads information from hmmcntrl.dat that specifies what is to be
done */
{
   extern int **imat();
   extern int *ivec();
   extern char **cmat();
   extern char *cvec();

   void docomma();
   void doleftpar();
   void readseq();

   FILE *hmmcntrl;
   int ktrack, ktrack2;
   int **ttonode ;
   char thischar; 


   hmmcntrl = gfopen("hmmcntrl.dat","r");
   model = cvec(20);
   while ( getc(hmmcntrl) != '\n');
   fscanf(hmmcntrl, "%s\n", model);
   fscanf(hmmcntrl, "%c", &outlevel);
   while ( getc(hmmcntrl) != '\n');
   while ( getc(hmmcntrl) != '\n');
   while ( getc(hmmcntrl) != '\n');
   fscanf(hmmcntrl, "%c", &predstrct);
   while ( getc(hmmcntrl) != '\n');
   while ( getc(hmmcntrl) != '\n');
   fscanf( hmmcntrl, "%c", &usertree);  
   while ( getc(hmmcntrl) != '\n');
   while ( getc(hmmcntrl) != '\n');
   while ( getc(hmmcntrl) != '\n');

   /* tonode (ttonode in this procedure) specifies the tree topology - 
      it is read in from a file named hmmcntrl.dat. The topology can have 
      only bifurcations - no
      multifurcations (at least I haven't tested the program with
      non-bifurcating topologies).  
      A tree with n tips has nodes 0,1,2,..., 2n-3.
      Nodes 0,...n-1 are the tips of the tree and represent the sequences
      in file testseq.  The value of node[i][0]
      is the number of child nodes that connect to node i.  For j > 0,
      the value of node[i][j] is the name of the jth child node
      connected to node i.  It is required ( I think) that, for j> 0,
      node[i][j] < i.  */

   ttonode = imat(MAXNODE,MAXOFF);
   name = cmat(MAXNODE,NAMELEN);

   for(ktrack=0;ktrack < MAXNODE; ktrack++)
   {
      ttonode[ktrack][0]=0;
   }
   ttonode[0][MAXOFF]= -1;

   nseqs=1;
   currentnode=0;
   nnodes=1;
   ktrack=0;
   thischar= ' ';

   while(thischar != ';' )
   {
      thischar=getc(hmmcntrl);
      while(thischar == '\n')
      {
        thischar=getc(hmmcntrl);
      }
      switch(thischar)
      {
         case ')': 
            currentnode = ttonode[currentnode][MAXOFF];
            ktrack=0;
            break;
         case '(': 
            doleftpar(ttonode);
            ktrack=0;
            break;
         case ',': 
            docomma(ttonode);
            ktrack=0;
            break;
         case ':': 
            ktrack=0; /* docolon() */
            break;
         case ';': 
            ktrack=0;
            break;
         case ' ':
            break;
         default: 
            name[currentnode][ktrack] = thischar;
            ktrack++;
            name[currentnode][ktrack] = '\n';
            break;
      }
   }

   nbranches = nnodes-1; 
    
   fclose(hmmcntrl);

   readseq(nseqs);

   return(ttonode);
}

void docomma(ttonode)
int **ttonode;
{
   int ktrack, ktrack2;

   currentnode = ttonode[currentnode][MAXOFF];
   nnodes++;
   nseqs++;
   for(ktrack=nnodes-2;ktrack >= nseqs-1; ktrack--)
   {
      if(ktrack== nnodes-2)
      {
         ttonode[nnodes-1][MAXOFF] = -1;
      }
      else
      {   
         ttonode[ktrack+1][MAXOFF] = ttonode[ktrack][MAXOFF]+1;
      }
      ttonode[ktrack+1][0] = ttonode[ktrack][0];
      for(ktrack2=1; ktrack2 <= ttonode[ktrack][0];ktrack2++)
      {
         if(ttonode[ktrack][ktrack2] >= nseqs-1)
         {
            ttonode[ktrack+1][ktrack2] = ttonode[ktrack][ktrack2] +1;
         }  
         else
         {
            ttonode[ktrack+1][ktrack2] = ttonode[ktrack][ktrack2];
         }
      }
   }
   ttonode[nseqs-1][MAXOFF] = currentnode;
   ttonode[nseqs-1][0]=0;
   ttonode[currentnode+1][0]++;
   ttonode[currentnode+1][ttonode[currentnode+1][0]] = nseqs-1;
   currentnode = nseqs-1; 
   for(ktrack=nseqs-1; ktrack >=0;ktrack--)
   {
      ttonode[ktrack][MAXOFF]++;
   }
}

void doleftpar(ttonode)
int **ttonode;
{
   int ktrack, ktrack2;

   nnodes++;
   for(ktrack=nnodes-2;ktrack >= nseqs; ktrack--)
   {
      if(ktrack== nnodes-2)
      {
         ttonode[nnodes-1][MAXOFF] = -1;
      }
      else
      {
         ttonode[ktrack+1][MAXOFF] = ttonode[ktrack][MAXOFF]+1;
      }
      ttonode[ktrack+1][0] = ttonode[ktrack][0];
      for(ktrack2=1;ktrack2 <= ttonode[ktrack][0];ktrack2++)
      {
         if(ttonode[ktrack][ktrack2] >= nseqs)
         {
            ttonode[ktrack+1][ktrack2] = ttonode[ktrack][ktrack2] +1;
         }
         else
         {
            if(ttonode[ktrack][ktrack2] == currentnode)
            {
               ttonode[ktrack+1][ktrack2] = nseqs;
            }
            else
            {
               ttonode[ktrack+1][ktrack2] = ttonode[ktrack][ktrack2] ;
            }
         } 
      }
   }
   if( ttonode[currentnode][MAXOFF] == -1)
   {
      ttonode[nseqs][MAXOFF] = -1;
   }
   else
   {    
      ttonode[nseqs][MAXOFF] = ttonode[currentnode][MAXOFF] + 1;
   }
   ttonode[nseqs][0]=1;
   ttonode[nseqs][1]=currentnode;
   for(ktrack=nseqs-1; ktrack >=0;ktrack--)
   {
      if(ktrack == currentnode)
      {
         ttonode[currentnode][MAXOFF] = nseqs;
      }
      else
      {
         ttonode[ktrack][MAXOFF]++;
      }
   }
}

void maxandprdct(tonode,param, structfrq,fromto)
int **tonode;   
double *param, *structfrq, **fromto;
{
#ifdef HESSIAN
   void hessian();
#endif
   void predict();
   void dofrq();
#ifdef DNA 
   void rdmodelinf();
#endif
   double calclhood();
   extern double ***threedmat();
   extern void freethreedmat();
   extern double **dmat();
   extern void freedmat();
   extern double exp();
   extern double log();
   double inc, incstep, oldparam, lhood, plushood, minushood, oldhood, nowhood;
   double ***evecarray, ***ievecarray, **lnevalarray, **bfrq;
   double lasthood;
   int hereyet, ktrack,stoploop,whichparam;
   int cat;
   FILE *bootfile;

   evecarray = threedmat(repcats,NUMLETTERS,NUMLETTERS);
   ievecarray = threedmat(repcats,NUMLETTERS,NUMLETTERS);
   lnevalarray = dmat(repcats,NUMLETTERS);
   bfrq = dmat(repcats,NUMLETTERS);
#ifdef DNA
   rdmodelinf(structfrq,fromto,bfrq, lnevalarray);
#endif
   for(ktrack=0;ktrack < nbranches;ktrack++)
   {
      param[ktrack] = exp(-1.0*param[ktrack]/DEFLATE);
   }
#ifndef DNA   
   for(cat=0;cat<=repcats;cat++)
   {
      dofrq(evecarray[cat],ievecarray[cat],lnevalarray[cat],cat, bfrq[cat]);
   }
#else
   dofrq(fromto,structfrq,lnevalarray,bfrq);
#endif

   oldhood = -1.0* BIGNUM;
   lhood = oldhood/2.0;
   hereyet = 0;
   while( lhood - oldhood > STOPTHRESH)
   {  
      oldhood = lhood;
      for(whichparam = nbranches-1; whichparam >= 0 ; whichparam--)
      {
         lasthood = lhood;

         oldparam = param[whichparam ];
         inc = INCFRAC*min2(param[whichparam], 1.0 - param[whichparam]);
         while((oldparam + inc) > 1.0 || (oldparam - inc < 0.0))
         {
            inc *= 0.5;
         }
         if(hereyet ==0)
         {
            nowhood = calclhood(tonode,param,
               structfrq,fromto,evecarray,ievecarray, lnevalarray,bfrq);
            lasthood = nowhood;
            hereyet=1;
         }
         else
         {   
            nowhood = lasthood;
         }
/* DEBUG - to prevent problems at boundary of parameter space */
         if(inc <= 0.0)
         {
            stoploop = 0;
         }
         else
         {
            param[whichparam] += inc;

            plushood = calclhood(tonode,param,
	       structfrq,fromto,evecarray,ievecarray, lnevalarray,bfrq);

            param[whichparam] -= 2*inc;
            minushood =  calclhood(tonode,param,
               structfrq,fromto,evecarray,ievecarray, lnevalarray,bfrq);
            if(plushood+minushood - 2*nowhood < 0.0)
            {
               /* Newton Step */
               incstep =  (-0.5)* inc*(plushood-minushood)/(plushood +
                    minushood - 2*nowhood);
            }
            else
            {
               /* Little Slope Step */
               incstep = STEPFRAC* (plushood - nowhood)/inc;
            }
            stoploop =1;
         }
         while(stoploop > 0)
         {

            if((oldparam + incstep > 1.0) || (oldparam + incstep < 0.0) )
            {
              /* Overshot ... */
               incstep *= 0.5;
            }
            else
            {   
               param[whichparam] = oldparam + incstep;
               lhood = calclhood(tonode,param,
                  structfrq,fromto,evecarray,ievecarray, lnevalarray,bfrq);
               if(outlevel != 'Q')
               {
                  printf(
                     "param[%2d] = %8.5f loghood= %12.5f, lastloghd= %12.5f\n",
                     whichparam, param[whichparam], lhood, lasthood);
               }
               if(lhood >= lasthood ||  param[whichparam] == oldparam)
               {
                  stoploop = 0;
               }
               else incstep *= 0.5;
            }


if(outlevel == 'Q') fprintf(stderr, "\rIterating, lnL = %12.6f  ", lhood);


         }   
      }   
   }   

   if(outlevel != 'Q')
   {
      printf("FINAL LIKELIHOOD IS %12.6f\n", lhood);
   }
   else
   {
      bootfile= gfopen("bootfile","a");
      fprintf(bootfile,"O 3 ");
      fprintf(bootfile,"%12.6lf ", lhood);
   }

   for(ktrack=0;ktrack < nbranches;ktrack++)
   {
      param[ktrack] = -1.0*DEFLATE*log(param[ktrack]);
      if(outlevel != 'Q') 
      {
         printf("FINAL ESTIMATE: param[%2d] = %10.5lf\n",ktrack,param[ktrack]);
      }
      else
      {
         fprintf(bootfile,"%10.4lf ", param[ktrack]);
      }
   }

   if(outlevel == 'Q')
   {
      fprintf(bootfile,"\n");
      fclose(bootfile);
   }

   if(predstrct == 'P')
   {
      predict(tonode,param,structfrq,fromto,evecarray,ievecarray,
              lnevalarray,bfrq);
   }

#ifdef HESSIAN
   hessian(tonode,param,structfrq,fromto,evecarray,ievecarray,
              lnevalarray,bfrq);
#endif

   freethreedmat(evecarray,repcats,NUMLETTERS,NUMLETTERS);
   freethreedmat(ievecarray,repcats,NUMLETTERS,NUMLETTERS);
   freedmat(lnevalarray,repcats,NUMLETTERS);
   freedmat(bfrq,repcats,NUMLETTERS);
}

#ifdef HESSIAN

void hessian(tonode,param, structfrq,fromto,evecarray,ievecarray,
                 lnevalarray,bfrq)
int **tonode;
double *param, *structfrq, **fromto, ***evecarray;
double ***ievecarray, **lnevalarray, **bfrq;
{
   int nvary, ktrack, ktrack1, ktrack2;
   int *finmask;
   int *imask;
   int holdplace, thisone;
   int newx, newy; 
   double tcorr, mhood,**hihi,**hilo,**lolo,*hino,*lono,*inc;
   double **invmat,**derivmat;
   double lhood, *newparam;
   double littlenum;
   double dohessian();
   void printname();
   void jprinttree();
   void countonodes();

   extern double *dvec();
   extern double **dmat();
   extern int *ivec();
   extern void freedvec();
   extern void freedmat();
   extern void freeivec();
   extern int **imat();
   extern void freeimat();
 
   finmask = ivec(nbranches);
   newparam = dvec(nbranches);
   inc=dvec(nbranches);
   if(NUMLETTERS > 5)
   {
      littlenum = 300.0/seq[0][0];
   }
   else
   {
      littlenum = 3.0/seq[0][0];
   }
   nvary = -1;
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      newparam[ktrack] = param[ktrack];
      if(param[ktrack] > littlenum)
      {
         nvary++;
         finmask[ktrack] = nvary;
         inc[finmask[ktrack]]= 0.001*newparam[ktrack];
      }
      else
      {
         finmask[ktrack] = -1;
      }
   }   

   hihi=dmat(nvary,nvary);
   lolo=dmat(nvary,nvary);
   hilo=dmat(nvary,nvary);
   hino = dvec(nvary);
   lono = dvec(nvary);
   invmat=dmat(nvary,nvary);
   derivmat=dmat(nvary,nvary);

   lhood = dohessian(tonode,newparam, structfrq,fromto,evecarray,ievecarray,
                 lnevalarray,bfrq);
   mhood = lhood;
   for (ktrack1=0;ktrack1< nbranches;ktrack1++)
   {
      if (finmask[ktrack1] >=0)
      {  
         newparam[ktrack1] += inc[finmask[ktrack1]];

         hino[finmask[ktrack1]] = dohessian(tonode,newparam, structfrq,
            fromto,evecarray,ievecarray, lnevalarray,bfrq); 

         for (ktrack2=0;ktrack2< nbranches;ktrack2++)
         {
            if (finmask[ktrack2] >= 0)
            {
               newparam[ktrack2]  += inc[finmask[ktrack2]];
               if (ktrack2 > ktrack1)
               {
                  hihi[finmask[ktrack1]][finmask[ktrack2]]=
                     dohessian(tonode,newparam, structfrq,fromto,evecarray,
                     ievecarray, lnevalarray,bfrq);

               }
               newparam[ktrack2] -= 2.0 * inc[finmask[ktrack2]];
               if (ktrack1 != ktrack2)
               {
                  hilo[finmask[ktrack1]][finmask[ ktrack2]]=
                     dohessian(tonode,newparam, structfrq,fromto,evecarray,
                     ievecarray, lnevalarray,bfrq);

               }
               newparam[ktrack2] += inc[finmask[ktrack2]];
            }
         }   
         newparam[ktrack1] -= 2.0 * inc[finmask[ktrack1]];
         lono[finmask[ktrack1]]= 
                     dohessian(tonode,newparam, structfrq,fromto,evecarray,
                     ievecarray, lnevalarray,bfrq);
         for (ktrack2=0;ktrack2 < nbranches ;ktrack2++)
         {
            if (finmask[ktrack2] >= 0)
            {
               newparam[ktrack2] -= inc[finmask[ktrack2]];
               if (ktrack2 > ktrack1)
               {
                  lolo[finmask[ktrack1]][finmask[ktrack2]] =
                     dohessian(tonode,newparam, structfrq,fromto,evecarray,
                     ievecarray, lnevalarray,bfrq);
               }
               newparam[ktrack2] += inc[finmask[ktrack2]];
            }
         }   
         newparam[ktrack1] += inc[finmask[ktrack1]];
      }
   }   
 
   for (ktrack1=0;ktrack1< nbranches;ktrack1++)
   {
      if (finmask[ktrack1] >=0)
      {
         derivmat[finmask[ktrack1]][finmask[ktrack1]] =
            (double) ((hino[finmask[ktrack1]]  -2.0*mhood +
            lono[finmask[ktrack1]])/ (inc[finmask[ktrack1]]*
            inc[finmask[ktrack1]]) );

         for (ktrack2=ktrack1+1;ktrack2< nbranches;ktrack2++)
         {
            if (finmask[ktrack2] >=0)
            {
               derivmat[finmask[ktrack1]][finmask[ktrack2]] = (double)
                  ( ( hihi[finmask[ktrack1]][finmask[ktrack2]] -
                  hilo[finmask[ktrack1]][finmask[ktrack2]] -
                  hilo[finmask[ ktrack2]][ finmask[ktrack1]] +
                  lolo[finmask[ktrack1]][ finmask[ktrack2]]) /
                  (4.0*inc[finmask[ktrack1]]*inc[finmask[ ktrack2]]) );
               derivmat[finmask[ktrack2]][finmask[ktrack1]] =
                  derivmat[finmask[ktrack1]][finmask[ktrack2]];
            }
         }   
      }   
   }   

   for (ktrack1=0;ktrack1<=nvary;ktrack1++)
   {
      for (ktrack2=0;ktrack2<=nvary;ktrack2++)
         invmat[ktrack1][ktrack2] = -1.0*derivmat[ktrack1][ktrack2];
   }
   invert(invmat,nvary);
   imask = ivec(MAXNODE);
   for(ktrack=0; ktrack <= nnodes-1; ktrack++)
   {
      imask[ktrack] =0;
   }
   countonodes(imask,tonode,tonode[nnodes-1][tonode[nnodes-1][0]]);
   for(ktrack=1; ktrack <= nnodes-1; ktrack++)
   {
      imask[ktrack] += imask[ktrack-1];
   }
   for(ktrack=nnodes-1; ktrack >= 1; ktrack--)
   {
      if(imask[ktrack] > imask[ktrack-1])
      {
         imask[ktrack] = -1;
      }
      else
      {
         imask[ktrack] = ktrack - imask[ktrack];
      }
   }
   imask[0] = 0 - imask[0];
   tonode[nnodes-1][0] = tonode[nnodes-1][0]-1;
   jprinttree(tonode,nnodes-1,1,imask[nnodes-1],newparam, imask);
/* now print the info. to screen today to a file tomorrow */
   printf("list of names and tip numbers follows:\n");
   for(ktrack=0; ktrack < nseqs; ktrack++)
   {  
      if(imask[ktrack] >=0)
      {
         printname(imask[ktrack]);
         printf(" %2d\n", imask[ktrack]);
      }
   }  
   printf("list of child1, child2, ..., parent follows:\n");
   for(ktrack=nseqs; ktrack <=nnodes-1; ktrack++)
   {
      if(imask[ktrack] >=0)
      {
         for(ktrack2=1; ktrack2 <= tonode[ktrack][0]; ktrack2++) 
         { 
            printf("%2d ", imask[tonode[ktrack][ktrack2]]);
         }
         printf("%2d\n", imask[ktrack]);
      }
   }
   printf("variance-covariance matrix follows:\n");
   for (ktrack1=0;ktrack1<nbranches;ktrack1++)
   {  
      for (ktrack2=0;ktrack2<nbranches;ktrack2++)
      {
         if(imask[ktrack1] >= 0 && imask[ktrack2] >= 0)
         {
            if( finmask[ktrack1] >= 0 && finmask[ktrack2] >= 0)
            {
if(NUMLETTERS > 5)
{
               printf("%9.4lf ", invmat[finmask[ktrack1]][finmask[ktrack2]]);
} 
else
{
               printf("%16.13lf ", invmat[finmask[ktrack1]][finmask[ktrack2]]);
}
            }
            else
            {
               if(ktrack1 == ktrack2)
               {
if(NUMLETTERS > 5)
{
                  printf("%9.4lf ", 10000.0*3.0/(seq[0][0]*seq[0][0]));
}
else
{
                  printf("%16.13lf", 3.0/(seq[0][0]*seq[0][0]));
}
               }
               else
               {
                  printf("   0.0    ");
               }
            }
         }
      }
      if(imask[ktrack1] >= 0)
      {
         printf("\n");
      }
   }  
 
   freeivec(imask);
   freeivec(finmask);
   if (nvary>=0)
   {
      freedmat(invmat,nvary,nvary);
      freedmat(derivmat,nvary,nvary);
      freedvec(lono);
      freedvec(hino);
      freedmat(hihi,nvary,nvary);
      freedmat(lolo,nvary,nvary);
      freedmat(hilo,nvary,nvary);
   }
   freedvec(inc);
   freedvec(newparam);
}

void countonodes(imask,tnode,whichnode)
int *imask, **tnode, whichnode;
{
   int ktrack;
   void countonodes();

   imask[whichnode] = 1;
   if(tnode[whichnode][0] != 0)
   {
      for(ktrack=1; ktrack <= tnode[whichnode][0]; ktrack++)
      {
         countonodes(imask,tnode,tnode[whichnode][ktrack]);
      }
   }
}
 
double dohessian(tonode,param, structfrq,fromto,evecarray,ievecarray,
                 lnevalarray,bfrq)
int **tonode;
double *param, *structfrq, **fromto, ***evecarray;
double ***ievecarray, **lnevalarray, **bfrq;
{
   void sitecathoods();

   extern double **dmat();
   extern double *dvec();
   extern int *ivec();
   extern void freedmat();
   extern void freedvec();
   extern void freeivec();
   extern double log();
   extern double exp();

   int site, scat, lastcat, cat, nextcat;
   int twojumps, *magsofar;
   double **sitehood;
   double *maxprob,**partialhood,maxcat;
   double tsum,bigmagnify, lilmagnify,microfy;
   double theanswer;

   partialhood = dmat(ncats, seq[0][0]);
   sitehood = dmat(repcats,seq[0][0]);
   magsofar = ivec(seq[0][0]);
   maxprob = dvec(seq[0][0]);


   twojumps= JUMP+JUMP;
   bigmagnify= exp(1.0*twojumps);
   lilmagnify = exp(1.0*JUMP);
   microfy= exp(-1.0*JUMP);

   sitecathoods(param,tonode,evecarray, ievecarray, lnevalarray,bfrq, sitehood);
   magsofar[0] = JUMP;
   for(site=1;site <= seq[0][0]; site++)
   {
      if(site == 1)
      {
         maxcat = 0.0;
         for(cat=0; cat <= ncats; cat++)
         {
            partialhood[cat][1] = structfrq[cat]*
               sitehood[mark2rep[cat]][1] *lilmagnify;
            maxcat = max2(maxcat, partialhood[cat][site]);
         }
      }   
      else
      {   
         maxcat = 0.0;
         for(cat=0; cat <= ncats; cat++)
         {
            partialhood[cat][site] = 0.0;
            for(lastcat=0; lastcat <= ncats;lastcat++)
            {
               partialhood[cat][site] += partialhood[lastcat][site-1] *
                  fromto[lastcat][cat];
            }
            partialhood[cat][site] *= sitehood[mark2rep[cat]][site];
            maxcat = max2(maxcat, partialhood[cat][site]);
         }
      }   
      magsofar[site] = magsofar[site-1];
      if(maxcat < microfy)
      {
         for(cat=0; cat <= ncats; cat++)
         {
            partialhood[cat][site] *= bigmagnify;
         }
         magsofar[site] += twojumps ;
      }
   }   

   tsum = 0.0;
   for(cat=0;cat<= ncats;cat++)
   {
      tsum += partialhood[cat][seq[0][0]];
   }
   theanswer = log(tsum) - magsofar[seq[0][0]];

   freedmat(partialhood,ncats,seq[0][0]);
   freedmat(sitehood,repcats,seq[0][0]);
   freedvec(maxprob); 
   freeivec(magsofar);
   return(theanswer);


}
#endif

void predict(tonode,param, structfrq,fromto,evecarray,ievecarray,
                 lnevalarray,bfrq)
int **tonode;
double *param, *structfrq, **fromto, ***evecarray;
double ***ievecarray, **lnevalarray, **bfrq;
{
   void sitecathoods();

   extern double **dmat();
   extern double *dvec();
   extern int *ivec();
   extern void freedmat();
   extern void freedvec();
   extern void freeivec();
   extern double log();
   extern double exp();

   int site, scat, lastcat, cat, nextcat;
   int twojumps, *magsofar;
   int *mxcat;
   double **sitehood, **posterior, *postpartial, *denom;
   double *maxprob,**partialhood,maxcat;
   double tsum,bigmagnify, lilmagnify,microfy;
   double *sumcatprob;
   FILE *siteguess, *siteprobs;

   partialhood = dmat(ncats, seq[0][0]);
   sitehood = dmat(repcats,seq[0][0]);
   magsofar = ivec(seq[0][0]);
   maxprob = dvec(seq[0][0]);
   mxcat = ivec(seq[0][0]);

   twojumps= JUMP+JUMP;
   bigmagnify= exp(1.0*twojumps);
   lilmagnify = exp(1.0*JUMP);
   microfy= exp(-1.0*JUMP);

   sitecathoods(param,tonode,evecarray, ievecarray, lnevalarray,bfrq, sitehood);
   magsofar[0] = JUMP;
   for(site=1;site <= seq[0][0]; site++)
   {
      if(site == 1)
      {
         maxcat = 0.0;
         for(cat=0; cat <= ncats; cat++)
         {
            partialhood[cat][1] = structfrq[cat]*
               sitehood[mark2rep[cat]][1] *lilmagnify;
            maxcat = max2(maxcat, partialhood[cat][site]);
         }
      }   
      else
      {   
         maxcat = 0.0;
         for(cat=0; cat <= ncats; cat++)
         {
            partialhood[cat][site] = 0.0;
            for(lastcat=0; lastcat <= ncats;lastcat++)
            {
               partialhood[cat][site] += partialhood[lastcat][site-1] *
                  fromto[lastcat][cat];
            }
            partialhood[cat][site] *= sitehood[mark2rep[cat]][site];
            maxcat = max2(maxcat, partialhood[cat][site]);
         }
      }   
      magsofar[site] = magsofar[site-1];
      if(maxcat < microfy)
      {
         for(cat=0; cat <= ncats; cat++)
         {
            partialhood[cat][site] *= bigmagnify;
         }
         magsofar[site] += twojumps ;
      }
   }   

   tsum = 0.0;
   for(cat=0;cat<= ncats;cat++)
   {
      tsum += partialhood[cat][seq[0][0]];
   }
   tsum = log(tsum) - magsofar[seq[0][0]];
   if(outlevel != 'Q') 
   {
      printf("The log-likelihood of the sequences is %10.6E\n", tsum);
   }
   freedmat(sitehood,repcats,seq[0][0]);

   postpartial = dvec(ncats);
   denom = dvec(ncats);
   posterior= dmat(ncats, seq[0][0]);
#ifndef DNA
   if(outlevel != 'Q') 
   {
      printf("\nCategory 0: Coil (i.e. other)\n");
      printf("Category 1: Helix\n");
      printf("Category 2: Sheet\n\n");
   }
#endif
   siteprobs = gfopen("siteprbs","w");
   fprintf(siteprobs,"%2d %4d %2d\n", nseqs, seq[0][0], nsumcats-1);

   sumcatprob = dvec(nsumcats-1);
   siteguess = gfopen("guess.nck","w");
   fprintf(siteguess,"%2d %4d %2d\n", nseqs, seq[0][0], nsumcats);
   for(site = seq[0][0]; site >=1; site--)
   {
      tsum = 0.0;
      for(scat=0;scat< nsumcats;scat++)
      {
         sumcatprob[scat]=0.0;
      }
      for(cat=0;cat <= ncats; cat++)
      {
         tsum += partialhood[cat][site];
      }
      fprintf(siteprobs,"%4d ", site);
      for(cat=0;cat <= ncats; cat++)
      {
         postpartial[cat] = partialhood[cat][site]/tsum;
      }
      
      if(site != seq[0][0])
      {
         for(nextcat= 0; nextcat <= ncats; nextcat++)
         {
            denom[nextcat] = 0.0;
            for(cat=0;cat <= ncats; cat++)
            {
               denom[nextcat] += postpartial[cat]*fromto[cat][nextcat];
            }
         }
      }
      if(outlevel != 'Q') 
      {
         printf("Site %4d Category probabilities: ", site); 
      }
      for(cat= 0; cat <= ncats; cat++)
      {
         if( site != seq[0][0])
         {

/* Comment : posterior does not actually need to be a matrix here - it could be
2 arrays of length ncats+1 of doubles */
            posterior[cat][site] = 0.0;
            for(nextcat = 0; nextcat <= ncats; nextcat++)
            {
/* BEGIN CHANGE */
               if(denom[nextcat] > microfy)
               {
                  posterior[cat][site] += (posterior[nextcat][site+1] *
                     fromto[cat][nextcat])/ denom[nextcat];
               }
/* above is CHANGED - I think it's OK but CHECK !! */
/* END CHANGE */
            }
            posterior[cat][site] *= postpartial[cat];
         }
         else
         {
            posterior[cat][site] = postpartial[cat];
         }
         sumcatprob[sumcat[cat]] += posterior[cat][site];

      }
      for(scat=0; scat<nsumcats;scat++)
      {
         
         if(outlevel != 'Q') 
         {
            printf("%7.4lf ", sumcatprob[scat]);
         }
         if(scat == 0)
         {
            mxcat[site] = 0;
            maxprob[site] = sumcatprob[0];
         }
         else
         {
            if(sumcatprob[scat] > maxprob[site])
            {
               maxprob[site] = sumcatprob[scat];
               mxcat[site] = scat;
            }
         }
         fprintf(siteprobs,"%8.6lf ", sumcatprob[scat]);
      }
      fprintf(siteprobs,"\n");

      if(outlevel != 'Q') 
      {
         printf("\n");
      }
   }

/* DEBUG- NOTE - CHANGE ... the following should be changed -- it
   really only works when only interest in predicting helix, sheet, coil 
   categories */
   for(site = 1; site <= seq[0][0]; site++)
   {
#ifndef DNA
      if(mxcat[site] == 0)
      {
         fprintf(siteguess, "3");
      }
      else
      {
         fprintf(siteguess, "%1d", mxcat[site]);
      }
#else
      fprintf(siteguess, "%1d", mxcat[site]);
#endif
      if(site % 60 == 0)
      {
         fprintf(siteguess,"\n");
      }
   }
   if( seq[0][0] % 60 != 0) 
   {
      fprintf(siteguess,"\n");
   }
#ifndef DNA
   fprintf(siteprobs,"most likely: Cat 0 coil, Cat 1 helix, Cat 2 sheet\n");
   fprintf(siteguess, "Cat 1 Helix - Cat 2 Sheet - Cat 3 Loop\n");
#endif
   fprintf(siteprobs,"Number of summary categories %d\n", nsumcats);
   fprintf(siteprobs,"Number of categories in HMM %d\n", ncats+1);
   fprintf(siteprobs,"Which HMM Category: Which Summary Category\n");

   for(cat=0;cat<=ncats;cat++)
   {
      fprintf(siteprobs, "%d %d\n", cat, sumcat[cat]);
   }
   fclose(siteprobs);
   fclose(siteguess);

   freedvec(sumcatprob,nsumcats-1);
   freedvec(denom,ncats);
   freedvec(postpartial,ncats);
   freedmat(posterior,ncats,seq[0][0]);
   freedmat(partialhood,ncats,seq[0][0]);
   freedvec(maxprob,seq[0][0]);
   freeivec(mxcat,seq[0][0]);
   freeivec(magsofar,seq[0][0]);
}



double calclhood(tonode,param, structfrq,fromto,evecarray,ievecarray,
                 lnevalarray,bfrq)
int **tonode;
double *param, *structfrq, **fromto, ***evecarray;
double ***ievecarray, **lnevalarray, **bfrq;
{
   void sitecathoods();
   
   extern double **dmat();
   extern int *ivec();
   extern void freedmat();
   extern void freeivec();
   extern double log();
   extern double exp();

   int ktrack,site, lastcat, cat;
   int twojumps, *magsofar;
   double **sitehood, temphood;
   double **partialhood,maxcat;
   double tsum,bigmagnify, lilmagnify,microfy;

   partialhood = dmat(ncats, seq[0][0]);
   sitehood = dmat(repcats,seq[0][0]);
   magsofar = ivec(seq[0][0]);
 
   twojumps= JUMP+JUMP;
   bigmagnify= exp(1.0*twojumps);
   lilmagnify = exp(1.0*JUMP);
   microfy= exp(-1.0*JUMP);

   for(ktrack=0;ktrack < nbranches;ktrack++)
   {
      param[ktrack] = -1.0*DEFLATE*log(param[ktrack]);
   }

   sitecathoods(param,tonode,evecarray, ievecarray, lnevalarray,bfrq, sitehood);

   magsofar[0] = JUMP;
   for(site=1;site <= seq[0][0]; site++)
   {
      if(site == 1)
      {
         maxcat = 0.0;
         for(cat=0; cat <= ncats; cat++)
         {
            partialhood[cat][1] = structfrq[cat]*
               sitehood[mark2rep[cat]][1] *lilmagnify;
            maxcat = max2(maxcat, partialhood[cat][site]);
         }
      }
      else
      {
         maxcat = 0.0;
         for(cat=0; cat <= ncats; cat++)
         {
            temphood = 0.0;
            for(lastcat=0; lastcat <= ncats;lastcat++)
            {
               temphood += partialhood[lastcat][site-1] * 
                  fromto[lastcat][cat];
            }
            temphood *= sitehood[mark2rep[cat]][site];
            maxcat = max2(maxcat, temphood);
            partialhood[cat][site] = temphood;
         }
      }  
      magsofar[site] = magsofar[site-1];
      if(maxcat < microfy)
      {
         for(cat=0; cat <= ncats; cat++)
         {
            partialhood[cat][site] *= bigmagnify;
         }
         magsofar[site] += twojumps ;
      }
   }

   tsum = 0.0;
   for(cat=0;cat<= ncats;cat++)
   {
      tsum += partialhood[cat][seq[0][0]];
   }
   tsum = log(tsum) - magsofar[seq[0][0]];

   freedmat(sitehood,repcats,seq[0][0]);
   freedmat(partialhood,ncats,seq[0][0]);
   freeivec(magsofar,seq[0][0]);

   for(ktrack=0;ktrack < nbranches;ktrack++)
   {
      param[ktrack] = exp(-1.0*param[ktrack]/DEFLATE);
   }

   return(tsum);
}


double calcsitehood(sitenum,tonode,fij,pi)
int sitenum, **tonode;
double ***fij,*pi;
{
   extern double **dmat();

   static int hereyet=0;
   int highernode,ktrack,ktrack2,ktrack3,ktrack4;
   double temp,gtemp, *fijpt, *shpt;
   static double **tsitehood;

   if(hereyet == 0)
   {   
      tsitehood = dmat(nnodes,NUMLETTERS);
      hereyet =1;
   }   

   for(ktrack=nseqs;ktrack< nnodes;ktrack++)
   {
      for(ktrack2=0;ktrack2< NUMLETTERS;ktrack2++)
      {
         gtemp = 1.0;
         for(ktrack3=1;ktrack3<= tonode[ktrack][0];ktrack3++)
         {
            temp = 0.0;
            highernode = tonode[ktrack][ktrack3];
            if(highernode < nseqs)
            {
               temp += fij[highernode][ktrack2][seq[highernode][sitenum]];
            }
            else
            {
               fijpt = &fij[highernode][ktrack2][0];
               shpt = &tsitehood[highernode][0];
               for(ktrack4=0;ktrack4< NUMLETTERS;ktrack4++)
               {
                  temp += (*shpt)*(*fijpt);
                  shpt++;
                  fijpt++;
               }
            }
            gtemp *= temp;
         }
         tsitehood[ktrack][ktrack2] = gtemp;
      }
   }   
      
   gtemp = 0.0;
   shpt = &tsitehood[nnodes-1][0];
   for(ktrack=0;ktrack< NUMLETTERS;ktrack++)
   {   
      gtemp += pi[ktrack]*(*shpt);
      shpt++;
   }   
/* SHOULD BE DONE BUT IS NOT FOR SAKE OF TIME ???
   ...  freedmat(tsitehood,nnodes,NUMLETTERS); */
   return(gtemp);
}
 

void sitecathoods(param,tonode,evecarray, ievecarray, lnevalarray, bfrq,
                    sitehood)
double *param;
double ***evecarray, ***ievecarray, **lnevalarray, **bfrq,**sitehood;
int **tonode;
{
   double calcsitehood();
   extern double ****fourdmat();
   extern double *dvec();
   void dobd();

   static int hereyet=0;
   int site,cat,branch,ktrack;
   static double ****ffourd;
   static double *lastparam; 

   if(hereyet ==0)
   {
      ffourd = fourdmat(repcats,nbranches-1,NUMLETTERS,NUMLETTERS);
      lastparam = dvec(nbranches-1);
      for(ktrack=0; ktrack <= nbranches-1; ktrack++)
      {
         lastparam[ktrack] = -10.0;
      }
   }

   for(cat=0;cat<=repcats;cat++)
   {
      for(branch=0;branch < nbranches;branch++)
      {
         if(lastparam[branch] != param[branch] || hereyet==0)
         {
#ifndef DNA
            dobd(param[branch],bfrq[cat],ffourd[cat][branch],
               evecarray[cat], ievecarray[cat], lnevalarray[cat]);  
#else
            dobd(param[branch],bfrq[cat],ffourd[cat][branch],
               gfrq[cat], lnevalarray[cat]);  
#endif
         }
      }
      for(site=1; site <= seq[0][0]; site++)
      {
         sitehood[cat][site] = calcsitehood(site,tonode,ffourd[cat], 
            bfrq[cat]);
      }
   }
   for(ktrack=0; ktrack < nbranches; ktrack++)
   {
      lastparam[ktrack]= param[ktrack];
   }
   hereyet=1;
}

#ifndef DNA
void
dofrq(evec,ievec,lneval,whichcat, bfrq) /* Read and Set-up the PAM or 
                                     PAM-like matrix */
int whichcat;
double **evec, **ievec, *lneval, *bfrq;
{
   extern int frombase();
   extern int *ivec();
   extern void freeivec();
   FILE *reveqps,*evalfile,*evecfile,*ievecfile;
   int ktrack1,ktrack2,*order;
   char thisamino;
   if(outlevel != 'Q')
   {
     printf("Model being used is specified in %s\n", model);
   }
   reveqps=gfopen(revps[whichcat],"r");
   evalfile=gfopen(evalfl[whichcat],"r");
   ievecfile= gfopen(ievecfl[whichcat],"r");
   evecfile= gfopen(evecfl[whichcat],"r");
   ktrack1=0;
   while (ktrack1<NUMLETTERS)
   {
      thisamino=getc(reveqps);
      if (frombase(thisamino) < NUMLETTERS)
      {
         fscanf(reveqps,"%lf",&bfrq[frombase(thisamino)]);
         ktrack1++;
      }
   }
   bfrq[NUMLETTERS]=1.0; /* This is the "Unknown" amino acid */
   fclose(reveqps);
   ktrack1=0;
   while (ktrack1<NUMLETTERS)
   {
      thisamino=getc(evalfile);
      if (frombase(thisamino) < NUMLETTERS)
      {
         fscanf(evalfile,"%lf",&lneval[frombase(thisamino)]);
         lneval[frombase(thisamino)]=
log(lneval[frombase(thisamino)]);
         ktrack1++;
      }
   }   
   fclose(evalfile);
   order=ivec(NUMLETTERS-1);
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {  
      thisamino= getc(evecfile);
      while (thisamino== ' ')
      {
         thisamino=getc(evecfile);
      }
      order[ktrack1]= frombase(thisamino);
   }  
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {  
      for (ktrack2=0;ktrack2 < NUMLETTERS;ktrack2++)
      {
         fscanf(evecfile,"%lf", &evec[order[ktrack1]][order[ktrack2]]);
      }
   }
   fclose(evecfile);
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {  
      thisamino= getc(ievecfile);
      while (thisamino== ' ')
      {
         thisamino=getc(ievecfile);
      }
      order[ktrack1]= frombase(thisamino);
   }
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      for (ktrack2=0;ktrack2 < NUMLETTERS;ktrack2++)
      {
         fscanf(ievecfile,"%lf", &ievec[order[ktrack1]][order[ktrack2]]);
      }
   }   
   fclose(ievecfile);
   freeivec(order,NUMLETTERS-1);
}

/* calculate Dayhoff replacement matrix and put in ftemp for amount of
   PAMs equal to grate */
void  
dobd(grate,bfrq,ftemp,evec,ievec,lneval,cat)
double grate, *bfrq,**ftemp, **evec, **ievec, *lneval;
int cat;
{
   char tobase();

   double eval[NUMLETTERS],tempmat[NUMLETTERS][NUMLETTERS];
   double tempftemp;
  
   int ktrack1,ktrack2,ktrack3;

   for (ktrack1=0;ktrack1<NUMLETTERS;ktrack1++)
   {  
      eval[ktrack1]=exp(grate*lneval[ktrack1]);
   }
   for (ktrack1=0;ktrack1<NUMLETTERS;ktrack1++)
   {
      for (ktrack2=0;ktrack2<NUMLETTERS;ktrack2++)
      {
         tempmat[ktrack1][ktrack2]=
evec[ktrack1][ktrack2]*eval[ktrack2];
      }
   }
      
 
   for (ktrack1=0;ktrack1<NUMLETTERS;ktrack1++)
   {
      tempftemp = 0.0;
      for (ktrack3=0;ktrack3<NUMLETTERS;ktrack3++)
      {
         tempftemp += tempmat[ktrack1][ktrack3]*
         ievec[ktrack3][ktrack1];
      }
      ftemp[ktrack1][ktrack1] = tempftemp;
      for (ktrack2=ktrack1+1;ktrack2<NUMLETTERS;ktrack2++)
      {
         tempftemp =0.0;
         for (ktrack3=0;ktrack3<NUMLETTERS;ktrack3++)
         {
            tempftemp += tempmat[ktrack1][ktrack3]*
               ievec[ktrack3][ktrack2];
         }
         if (tempftemp < SMALLNUM)
         {
            tempftemp =SMALLNUM;
         }
         ftemp[ktrack2][ktrack1]=tempftemp*bfrq[ktrack1]/bfrq[ktrack2];
         ftemp[ktrack1][ktrack2] = tempftemp;
      }
      ftemp[ktrack1][NUMLETTERS]= 1.0;
      ftemp[NUMLETTERS][ktrack1]=bfrq[ktrack1];
   }
   ftemp[NUMLETTERS][NUMLETTERS]=1.0;

}
 

char 
tobase( thisbase) 
/* Take integer code for an amino acid and return character code */
int thisbase;
{
  switch (thisbase){
     case 0: return('G');
        break;
     case 1: return('A');
        break;
     case 2: return('V');
        break;
     case 3: return('L');
        break;
     case 4: return('I');
        break;
     case 5: return('M');
        break;
     case 6: return('F');
        break;
     case 7: return('P');
        break;
     case 8: return('S');
        break;
     case 9: return('T');
        break;
     case 10: return('C');
        break;
     case 11: return('N');
        break;
     case 12: return('Q');
        break;
     case 13: return('Y');
        break;
     case 14: return('W');
        break;
     case 15: return('D');
        break;
     case 16: return('E');
        break;
     case 17: return('H');
        break;
     case 18: return('K');
        break;
     case 19: return('R');
        break;
     case 20: return('?');
        break;
     default: printf("ERROR2 !!\n");
        return('!');
        break;
  }
}

int 
frombase(basechar) 
/* take character code for an amino acid and return integer code */
char basechar;
{
  int whichamino;

  if(basechar > 96 && basechar < 123)
  {
     basechar -= 32;
  }

  switch ( basechar )
  {
      case 'G' :  whichamino=0;
         break;
      case 'A' :  whichamino=1;
         break;
      case 'V' :  whichamino=2;
         break;
      case 'L' :  whichamino=3;
         break;
      case 'I' :  whichamino=4;
         break;
      case 'M' :  whichamino=5;
         break;
      case 'F' :  whichamino=6;
         break;
      case 'P' :  whichamino=7;
         break;
      case 'S' :  whichamino=8;
         break;
      case 'T' :  whichamino=9;
         break;
      case 'C' :  whichamino= 10;
         break;
      case 'N' :  whichamino= 11;
         break;
      case 'Q' :  whichamino= 12;
         break;
      case 'Y' :  whichamino= 13;
         break;
      case 'W' :  whichamino= 14;
         break;
      case 'D' :  whichamino= 15;
         break;
      case 'E' :  whichamino= 16;
         break;
      case 'H' :  whichamino= 17;
         break;
      case 'K' :  whichamino= 18;
         break;
      case 'R' :  whichamino= 19;
         break;
      case '?' :  whichamino= 20;
         break;
      default :   whichamino= 99;
         break;
   }
   return(whichamino);
}

#else
/* replace preceding routines with DNA equivalents  -need DNA equivalent
of dobd() routine and dofrq() routine */

char
tobase( thisbase)
/* take integer code for base, return character code */
int thisbase;
{
  switch (thisbase){
     case 0: return('A');
        break;
     case 1: return('G');
        break;
     case 2: return('T');
        break;
     case 3: return('C');
        break;
     case 4: return('?');
        break;
     default: printf("ERROR2 !!\n");
        return('!');
        break;
  }
}

 
int
frombase(basechar)
/* take character code for base, return integer code */
char basechar;
{
  int whichbase;
 
  if(basechar > 96 && basechar < 123)
  {
     basechar -= 32;
  }
  switch ( basechar )
  {
      case 'A' :  whichbase=0;
         break;
      case 'G' :  whichbase=1;
         break;
      case 'T' :  whichbase=2;
         break;
      case 'U': whichbase =2;
         break;
      case 'C' :  whichbase=3;
         break;
      case '?' :  whichbase=4;
         break;
      case '-': whichbase=4;
         break;
      case 'R': whichbase=4;
         break;
      case 'Y': whichbase=4;
         break;
      case 'K': whichbase=4;
         break;
      case 'M': whichbase=4;
         break;
      case 'S': whichbase=4;
         break;
      case 'W': whichbase=4;
         break;
      case 'B': whichbase=4;
         break;
      case 'D': whichbase=4;
         break;
      case 'H': whichbase=4;
         break;
      case 'V': whichbase=4;
         break;
      case 'N': whichbase=4;
         break;
      default:    whichbase=99;
         break;
   }
   return(whichbase);
}

void
dofrq(fromto, structfrq, lneval, bfrq) /* Read and Set-up the  Felsenstein 84 
                                          model */ 
double **fromto, *structfrq, **lneval, **bfrq;
{
   extern int frombase();
   extern int *ivec();
   extern double **dmat();
   extern void freeivec();
   int ktrack;
   int thiscat;
   double thiscons;
   if(outlevel != 'Q')
   {
     printf("Model being used is specified in %s\n", model);
   }
 
/* normalize now so that rates are 1 */
/* DEFINE group and DEFINE gfrq for each residue type */
   ratecons = 0.0;
   group[frombase('A')] = 0;  
   group[frombase('G')] = 0;
   group[frombase('T')] = 1;
   group[frombase('C')] = 1;

   gfrq = dmat(ncats,NUMLETTERS);

   for(thiscat=0; thiscat <= ncats; thiscat++)
   {  
      gfrq[thiscat][0] = bfrq[thiscat][frombase('A')]  +
         bfrq[thiscat][frombase('G')];
      gfrq[thiscat][1] = bfrq[thiscat][frombase('T')]  +
         bfrq[thiscat][frombase('C')];
   }  
   for(thiscat=0; thiscat <= ncats; thiscat++)
   {  
      thiscons = 0.0; 
      for(ktrack=0; ktrack < NUMLETTERS; ktrack++)
      {
         thiscons += bfrq[thiscat][ktrack]*((1.0-bfrq[thiscat][ktrack]) +
            lneval[thiscat][0]* ((gfrq[thiscat][group[ktrack]] -
            bfrq[thiscat][ktrack])/gfrq[thiscat][group[ktrack]]));
      }
      ratecons += structfrq[thiscat]*lneval[thiscat][1]*thiscons;
   }
   ratecons = 1.0/ratecons;
    
}

/* calculate nucleotide substitution matrix and put in ftemp for Felsenstein
   1984 model.  grate is the general substitution rate multiplied by time.
   bfrq is the base frequency for this site category.  evec and ievec
   are ignored by this routine and are fossil evidence of bad programming.
   lneval[0] contains the ratio of the within-group rate to the between-
   group substitution rate.  lneval[1] contains the relative rate of
   the category being considered. 
*/
void  dobd(double grate, double *bfrq, double **ftemp, double *gf, double *lneval)
{
   int ktrack1,ktrack2;
   double expgrate, expwrate, thiscons;
   int frombase();

   expgrate= exp(-grate*lneval[1]*ratecons);
   expwrate= exp(-grate*lneval[1]*lneval[0]*ratecons);
   for (ktrack1=0;ktrack1< NUMLETTERS;ktrack1++)
   {
      ftemp[ktrack1][ktrack1]= (1.0-expgrate)*bfrq[ktrack1] +
         expgrate*(1.0-expwrate)*bfrq[ktrack1]/gf[group[
         ktrack1]] + expgrate*expwrate;
/*
      ftemp[ktrack1][ktrack1] *= bfrq[ktrack1];
*/
      for (ktrack2=ktrack1+1;ktrack2< NUMLETTERS;ktrack2++)
      {  
         if (group[ktrack1]==group[ktrack2])
         {
            ftemp[ktrack1][ktrack2]= (1.0-expgrate)*bfrq[ktrack2] +
               expgrate*(1.0-expwrate)*bfrq[ktrack2]/gf[group[ktrack2]];
         }
         else
         {
            ftemp[ktrack1][ktrack2]= (1.0-expgrate)*bfrq[ktrack2];
         }
         ftemp[ktrack2][ktrack1]= ftemp[ktrack1][ktrack2]*
            (bfrq[ktrack1]/bfrq[ktrack2]);
/*
         ftemp[ktrack1][ktrack2] *= bfrq[ktrack1];
*/
      }  
      ftemp[ktrack1][NUMLETTERS]=1.0;
/*
      ftemp[ktrack1][NUMLETTERS]=bfrq[ktrack1];
      ftemp[NUMLETTERS][ktrack1]=bfrq[ktrack1];
      ftemp[NUMLETTERS][NUMLETTERS]=1.0;
*/
   }
}


#endif

void
printname(whichseq)
int whichseq;
   /* prints the name of sequence number whichseq */
{
   int ttrack;
   char c;

   ttrack = 0;
   c = ' ';
   while (c != '\n')
   {
      c= name[whichseq][ttrack];
      if( c != '\n') printf("%1c",c);
      ttrack++;
   }
}

void
readseq(ntseqs)
int ntseqs;
                  /* this procedure reads in the data from a formatted file
                     It reads the number of sequences in the file
                     ("nseqs") first "nseqs" in the file.  It
                     waits for a '>' character and then 2 carriage returns
                     (i.e. '\n') before reading in each sequence.  The sequence
                     is terminated with an asterisk */
{
   int frombase();
   int **imat();
   char **cmat();
   int *ivec();
   void freeivec();
   void freecmat();
   void printname();

   FILE *f;
   char c, **fname;
   int stopflag, totlen,thislength,whichseq;
   int thisseq;
   int ktrack,ktrack1,ktrack2,ttrack;
   int nnseqs;
   int *transname;

   f=gfopen("testseq","r");
   fscanf(f,"%d%d", &nnseqs, &totlen);
 
   seq = imat(ntseqs,totlen);

   if(usertree == 'O')
   {
      transname = ivec(ntseqs);
      for(ktrack1=0; ktrack1 < ntseqs; ktrack1++)
      {
         transname[ktrack1]=0;
         stopflag = 0;
         ktrack2 = 0;
         while(stopflag != 1)
         {
            if(name[ktrack1][ktrack2] == '\n')
            {
               stopflag=1;
            }
            else
            {
               transname[ktrack1] *= 10;
               transname[ktrack1] += ((int) name[ktrack1][ktrack2]) - 48;
               ktrack2++;   
            }
	 }
      }
   }
   else
   {
      fname = cmat(0,NAMELEN);
   }
   nnseqs=0; 
   for(ktrack1=0;ktrack1 < ntseqs;ktrack1++)
   {
      seq[ktrack1][0] = totlen;
   }
 
   for(thisseq=0;thisseq<ntseqs; thisseq++)
   {
      if(usertree == 'O')
      {
         for(ktrack=0; ktrack < ntseqs; ktrack++)
         {
            if(transname[ktrack] == thisseq)
            {
               whichseq = ktrack;
               nnseqs++;
               break;
            }
         }
      }    
      while (getc(f) != '>');
      ttrack=0;

      thislength=0;
      while (getc(f) != '\n');
      c=getc(f);
      while ( c != '\n' && ttrack <80)
      {
         if( c != ' ')
         {
            if(ttrack < NAMELEN)
            {
               if(usertree == 'O')
               {
                  name[whichseq][ttrack] =c;
               }
               else
               {
                  fname[0][ttrack]=c;
               }
            }
            ttrack++;
         }
         c=getc(f);
      }

      if(usertree == 'O')
      {
         name[whichseq][min2(NAMELEN, ttrack)] = '\n';
      }
      else 
      {
         fname[0][min2(NAMELEN, ttrack)] = '\n';
         /* FIND the tip of the tree with name fname */
         whichseq = 0;
         stopflag = 0;
         while(whichseq < ntseqs)
         {
            ttrack=0;
            while( ttrack <= NAMELEN)
            {
               if(fname[0][ttrack] != name[whichseq][ttrack])
               {
                  whichseq++;
                  break;
               }
               else
               {
                  if(fname[0][ttrack] == '\n')
                  {
                     stopflag=1;
                     break;
                  }
                  else
                  {
                     ttrack++;
                  }
               }
            }
            if(stopflag ==1)
            {
               nnseqs++;
/* DEBUG 
printf("nnseqs = %d thisseq = %d thislength = %d\n",nnseqs,thisseq,thislength);
*/
               break;
            } 
         }
      }


      if (ttrack==80 && c != '\n')
      {
         while ( getc(f) != '\n');
      }

      while ( (c=getc(f)) != '*')
      {
         if ( frombase(c) != 99)
         {
            thislength++;
            seq[whichseq][thislength]= frombase(c);

         }
      }   
   }   

   fclose(f);

   if(outlevel != 'Q')
   {
      for(ktrack=0;ktrack<nnseqs;ktrack++)
      {
         printf("The name of sequence %2d is ", ktrack);
         printname(ktrack);
         printf("\n");
      }   
      printf("\n\n");
   }

   if(nnseqs != nseqs)
   {
      printf("Trouble in hmmcntrl.dat or testseq! Number of sequences\n");
      printf("specified disagree! nnseqs= %3d nseqs= %3d\n", nnseqs, nseqs);
   }

   if(usertree == 'O')
   {
      freeivec(transname,ntseqs); 
   }
   else
   {
      freecmat(fname,0,NAMELEN);
   }
}


FILE *gfopen(char *filename, char *mode)
{
   FILE *fp=(FILE*)fopen(filename, mode);
   if(fp==NULL) {
      printf("\nerror when opening file %s\n", filename);
      exit(-1);
   }
   return(fp);
}
